/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#include "globals.h"

#include "b_conn.h"
#include "b_conn_accept.h"
#include "b_conn_connect.h"
#include "b_conn_io.h"
#include "b_conn_handler.h"
#include "b_conn_ping.h"
#include "b_util.h"
#include "b_packet.h"


/* ************************************************************ */


BConn*   
b_conn_new (gchar* hostname, gint port, guint protocol, gchar* object_name,
	    JMGTOS tos, BConnFuncs* funcs, gpointer user_data)
{
  BConn* conn;

  g_return_val_if_fail (hostname, 	NULL);
  g_return_val_if_fail (port != 0, 	NULL);
  g_return_val_if_fail (object_name, 	NULL);

  DEBUGP ("b_conn_new\n");

  conn = g_new0 (BConn, 1);
  conn->conn = jm_conn_new (hostname, port, NULL, 0);	/* FIX */
  conn->protocol = protocol;
  conn->object_name = g_strdup (object_name);
  conn->tos = tos;
  conn->funcs = funcs;
  conn->user_data = user_data;
  conn->ref_count = 1;
  conn->status = B_CONN_STATUS_UNCONNECTED;

  return conn;
}


void
b_conn_ref (BConn* conn)
{
  g_return_if_fail (conn);

  ++conn->ref_count;
}


void
b_conn_unref (BConn* conn)
{
  g_return_if_fail (conn);

  --conn->ref_count;
  if (conn->ref_count == 0)
    b_conn_delete (conn);
}



void
b_conn_delete (BConn* conn)
{
  g_return_if_fail (conn != NULL);

  DEBUGP ("b_conn_delete\n");

  g_free (conn->object_name);
  b_conn_reset (conn);

  memset (conn, 0, sizeof(*conn));
  g_free (conn);
}



void
b_conn_reset (BConn* conn)
{
  GList* i;

  g_return_if_fail (conn);

  DEBUGP ("b_conn_reset\n");

  conn->status = B_CONN_STATUS_UNCONNECTED;

  /* Stop pinging it */
  b_conn_ping_stop (conn);
  /* keep the distance estimate */

  /* Reset connection */
  jm_conn_disconnect (conn->conn, TRUE);

  /* Cancel any scheduled writes/reads */
  scheduler_remove (b_conn_scheduler_read, conn);
  scheduler_remove (b_conn_scheduler_write, conn);

  /* Remove all handlers */
  b_conn_handler_remove_all (conn);

  conn->id_in = 0;
  conn->id_out = 0;

  for (i = conn->packet_queue; i != NULL; i = i->next)
    b_packet_delete ((BPacket*) i->data);

  g_list_free (conn->packet_queue);
  conn->packet_queue = NULL;

  conn->packet_queue_size = 0;

  conn->buffer_length = 0;
  conn->total_read = 0;
  if (conn->current_packet)
    {
      b_packet_delete (conn->current_packet);
      conn->current_packet = NULL;
    }
}



void
b_conn_print (FILE* file, BConn* conn)
{
  if (conn)
    {
      gchar* status;
  
      switch (conn->status)
	{
	case B_CONN_STATUS_UNCONNECTED:	status = "unconnected";	break;
	case B_CONN_STATUS_CLOSING:	status = "closing";	break;
	case B_CONN_STATUS_CONNECTED:	status = "connected";	break;
	case B_CONN_STATUS_CONNECTING:	status = "connecting";	break;
	case B_CONN_STATUS_ACCEPTED:	status = "accepted";	break;
	}

      fprintf (file,
	       "%s:%d [status = %s, distance = %d, "
	       "queue length = %d, queue size = %d]\n",
	       conn->conn->hostname, conn->conn->port, status, conn->distance,
	       g_list_length (conn->packet_queue), conn->packet_queue_size);
    }
  else
    {
      fprintf (file, "(null)");
    }
}



/* **************************************** */


void
b_conn_func_connect (BConn* conn)
{
  g_return_if_fail (conn);

  DEBUGP ("b_conn_func_connect\n");

  if (conn->funcs && conn->funcs->connect)
    conn->funcs->connect (conn, conn->user_data);
}


void
b_conn_func_ping (BConn* conn)
{
  g_return_if_fail (conn);

  DEBUGP ("b_conn_func_ping\n");

  if (conn->funcs && conn->funcs->ping)
    conn->funcs->ping (conn, conn->user_data);
}


void
b_conn_func_writeable (BConn* conn)
{
  g_return_if_fail (conn);

  DEBUGP ("b_conn_func_writeable\n");

  if (conn->funcs && conn->funcs->writeable)
    conn->funcs->writeable (conn, conn->user_data);
}


void
b_conn_func_close (BConn* conn)
{
  g_return_if_fail (conn);

  DEBUGP ("b_conn_func_close\n");

  b_conn_reset (conn);
  
  if (conn->funcs && conn->funcs->close)
    conn->funcs->close (conn, conn->user_data);
}


void
b_conn_func_fail (BConn* conn)
{
  g_return_if_fail (conn);

  DEBUGP ("b_conn_func_fail\n");

  b_conn_reset (conn);

  if (conn->funcs && conn->funcs->fail)
    conn->funcs->fail (conn, conn->user_data);
}




/* **************************************** */


gboolean
b_conn_has_packet (BConn* conn, guint8 type)
{
  GList* i;

  g_return_val_if_fail (conn != NULL, FALSE);

  for (i = conn->packet_queue; i != NULL; i = i->next)
    {
      BPacket* packet = (BPacket*) i->data;

      if (packet->type == type)
	return TRUE;
    }

  return FALSE;
}


gboolean
b_conn_has_packet_by_subtype (BConn* conn, guint8 type, guint8 subtype)
{
  GList* i;

  g_return_val_if_fail (conn != NULL, FALSE);

  for (i = conn->packet_queue; i != NULL; i = i->next)
    {
      BPacket* packet = (BPacket*) i->data;

      if (packet->type == type && packet->subtype == subtype)
	return TRUE;
    }

  return FALSE;
}


/* **************************************** */



void 
b_conn_send_packet (BConn* bconn, BPacket* packet)
{
  g_return_if_fail (bconn);
  g_return_if_fail (packet);


  DEBUGP ("b_conn_send_packet: proto = %d, version = %d, type = %d, %d, length = %d, id = %d\n",
	  packet->protocol, packet->version, packet->type, 
	  packet->subtype, g_ntohs(packet->length), packet->id);


  /* Connect if unconnected */
  if (b_conn_is_unconnected(bconn))
    {
      b_conn_connect (bconn);
    }

  /* Make sure we're watching for a write */
  else if (jm_conn_is_connected(bconn->conn))
    {
      b_conn_watch_write (bconn);
    }


  /* If the packet is a CONNECT, put it at the front of the packet list */
  if (packet->type == B_PACKET_TYPE_CONNECT)
    {
      /* Remove any CONNECTs that are in the queue now */
      b_conn_remove_packets (bconn, B_PACKET_TYPE_CONNECT);

      /* Put at the front of the queue */
      b_conn_prepend_packet (bconn, packet);
    }

  /* Likewise, put a CONNECT_OK at the front of the packet list */
  else if (packet->type == B_PACKET_TYPE_OK &&
	   packet->subtype == B_PACKET_OK_SUBTYPE_CONNECT)
    {
      /* Remove any CONNECT_OKs that are in the queue now */
      b_conn_remove_packets_by_subtype (bconn, B_PACKET_TYPE_OK,
					B_PACKET_OK_SUBTYPE_CONNECT);

      /* Put at the front of the queue */
      b_conn_prepend_packet (bconn, packet);
    }

  /* If the packet is MCAST data, make sure we have enough room. */
  else if (packet->type == B_PACKET_TYPE_MCAST)
    {
      if ((bconn->packet_queue_size + g_ntohs(packet->length)) <
	  B_CONN_MAX_MCAST_SIZE)
	{
	  b_conn_append_packet (bconn, packet);
	  bconn->packet_queue_size += g_ntohs(packet->length);
	}
      else
	{
	  DEBUGP("Dropping MCAST packet due to space\n");
	}

    }

  /* Otherwise, its just a regular packet, so add it to the back of
     the list */
  else
    b_conn_append_packet (bconn, packet);

}


void
b_conn_prepend_packet (BConn* conn, BPacket* packet)
{
  g_return_if_fail (conn);
  g_return_if_fail (packet);

  conn->packet_queue = g_list_prepend (conn->packet_queue, packet);
}


void	 
b_conn_append_packet (BConn* conn, BPacket* packet)
{
  g_return_if_fail (conn);
  g_return_if_fail (packet);

  conn->packet_queue = g_list_append (conn->packet_queue, packet);
}




/* **************************************** */


BPacket* 
b_conn_pop_packet (BConn* conn)
{
  BPacket* packet = NULL;
  BPacket* data_packet = NULL;
  GList* i;

  g_return_val_if_fail (conn != NULL, NULL);

  DEBUGP ("b_conn_pop_packet\n");

  if (conn->packet_queue == NULL)
    return NULL;

  /* Search for a control packet (other than BYE).  Also keep
     track of the first data packet we see.  */
  for (i = conn->packet_queue; i != NULL; i = g_list_next(i))
    {
      BPacket* p = (BPacket*) i->data;

      if (p->type == B_PACKET_TYPE_MCAST ||
	  p->type == B_PACKET_TYPE_UCAST)
	{
	  if (data_packet == NULL)
	    data_packet = p;
	}
      else if (p->type != B_PACKET_TYPE_BYE)
	{
	  packet = p;
	  break;
	}
      /* otherwise it's a BYE, ignore it for now.  This effectively
	 pushes the BYE to the back of the queue. */
    }


  /* If we didn't find a control packet, use a data packet.  If we
     didn't find a data packet, use the top packet, which should be a
     BYE. */
  if (packet == NULL)
    {
      if (data_packet != NULL)
	packet = data_packet;
      else
	{
	  packet = (BPacket*) conn->packet_queue->data;

	  g_return_val_if_fail (packet->type == B_PACKET_TYPE_BYE, NULL);
	}
    }

  g_return_val_if_fail (packet != NULL, NULL);

  /* If we aren't connected and this is a non-connect packet, return
     NULL.  We don't want to send anything special yet. */
  if (conn->status == B_CONN_STATUS_CONNECTING && 
      packet->type != B_PACKET_TYPE_CONNECT)
    return NULL;

  /* Remove the packet from the list */
  conn->packet_queue = g_list_remove (conn->packet_queue, packet);


  /* If it's a ping, set the time now.  (If we set it earlier, it
     might sit around in our buffers before we actually sent it,
     which could cause a bad distance estimate.) */
  if (packet->type == B_PACKET_TYPE_INFO_REQUEST &&
      packet->subtype == B_PACKET_INFO_SUBTYPE_PING)
    {
      g_assert (gettimeofday((struct timeval*) packet->data, NULL) == 0);
    }

  /* Otherwise, if it's a MCAST packet, deduct it's length from
     packet_queue_size. */
  else if (packet->type == B_PACKET_TYPE_MCAST)
    {
      conn->packet_queue_size -= g_ntohs(packet->length);
    }

  return packet;
}


void
b_conn_remove_packets (BConn* conn, guint8 type)
{
  GList* i;

  DEBUGP ("b_conn_remove_packets\n");

  for (i = conn->packet_queue; i != NULL; )
    {
      GList* old_i = i;
      BPacket* packet = (BPacket*) i->data;
      i = i->next;

      if (packet->type == type)
	{
	  if (packet->type == B_PACKET_TYPE_MCAST)
	    conn->packet_queue_size -= g_ntohs (packet->length);

	  b_packet_delete (packet);
	  conn->packet_queue = g_list_remove_link (conn->packet_queue, old_i);
	}
    }
}


void
b_conn_remove_packets_by_subtype (BConn* conn, guint8 type, guint8 subtype)
{
  GList* i;

  DEBUGP ("b_conn_remove_packets_by_subtype\n");

  for (i = conn->packet_queue; i != NULL; )
    {
      GList* old_i = i;
      BPacket* packet = (BPacket*) i->data;
      i = i->next;

      if (packet->type == type && packet->subtype == subtype)
	{
	  if (packet->type == B_PACKET_TYPE_MCAST)
	    conn->packet_queue_size -= g_ntohs (packet->length);

	  b_packet_delete (packet);
	  conn->packet_queue = g_list_remove_link (conn->packet_queue, old_i);
	}
    }
}
