/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>

#include "globals.h"
#include "jmgnet.h"

#include "b_conn_connect.h"
#include "b_conn_io.h"
#include "b_conn_handler.h"
#include "b_conn_ping.h"
#include "b_util.h"

#include "b_conn_accept.h"


static void	b_server_func (JMGServer* gserver, JMGServerStatus status, 
			       JMGConn* conn, gpointer user_data);

static void	receive_connect (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_connect_timeout (BConn* conn, gint type, gint subtype, 
					 gpointer user_data);


Scheduler* 		b_conn_scheduler_read = NULL;
Scheduler* 		b_conn_scheduler_write = NULL;

JMGServer*		b_server = NULL;
gchar* 			b_hostname = NULL;

static BConnAcceptFunc	b_accept_funcs[256] = {NULL, };

static gboolean		b_is_initialized = FALSE;


gboolean b_conn_init (gchar* hostname, GInetAddr* interface, gint port, gboolean force_port);
void	 b_conn_shutdown (void);

void	 b_conn_accept_protocol (guint8 protocol, BConnAcceptFunc func); 


/* ************************************************************ */


gboolean
b_conn_init (gchar* hostname, GInetAddr* interface, gint port, gboolean force_port)
{
  g_return_val_if_fail(hostname, TRUE);
  g_return_val_if_fail(interface, TRUE);

  DEBUGP ("b_conn_init\n");

  /* Ignore if already initialized */
  if (b_is_initialized)
    return FALSE;

  b_server = jm_server_new (interface, port, force_port, b_server_func, NULL);
  if (!b_server)
    return TRUE;

  /* Copy the hostname */
  b_hostname = g_strdup(hostname);

  /* Create bandwidth schedulers */
  b_conn_scheduler_read = scheduler_new();
  b_conn_scheduler_write = scheduler_new();

  scheduler_set_max_ups (b_conn_scheduler_write, 8000);

  b_is_initialized = TRUE;

  return FALSE;
}



void
b_conn_shutdown (void)
{
  DEBUGP ("b_conn_shutdown\n");

  g_free (b_hostname);		b_hostname = NULL;
  jm_server_delete (b_server);	b_server = NULL;

  /* Don't delete the schedulers since we need them to send things... */

/*    scheduler_delete (b_conn_scheduler_read);	b_conn_scheduler_read = NULL; */
/*    scheduler_delete (b_conn_scheduler_write);	b_conn_scheduler_write = NULL; */

}


void
b_conn_set_accept_func (guint8 protocol, BConnAcceptFunc func)
{
  DEBUGP ("b_conn_set_accept_func\n");

  b_accept_funcs[protocol] = func;
}



/* ************************************************************ */


/**

   Called when someone connects to us.

   We get the conn, create a new b_conn, add watches for read and
   error, then set a timer for receiving the CONNECT.

   Last checked: 	2000-6-18 DAH

*/
static void
b_server_func (JMGServer* gserver, JMGServerStatus status, JMGConn* conn, gpointer user_data)
{
  BConn* bconn = NULL;

  DEBUGP ("b_server_func %d\n", status); 

  g_assert (status != GNET_SERVER_STATUS_ERROR);

  /* Create a new conn.  */
  bconn = g_new0 (BConn, 1);
  bconn->conn = conn;
  bconn->ref_count = 1;
  bconn->status = B_CONN_STATUS_ACCEPTED;

  /* Watch read and err */
  b_conn_watch_read (bconn);
  b_conn_watch_error (bconn);

  /* Handle a CONNECT packet */
  b_conn_handler_add_full (bconn, B_PACKET_TYPE_CONNECT, -1, 
			   receive_connect, receive_connect_timeout, 
			   RECEIVE_CONNECT_TIME, NULL);

  /* We should now receive a CONNECT and will send back an OK or
     ERROR.  If we don't receive a CONNECT, the timer will expire and
     the connection will be closed. */
}




/* ************************************************************ */



/**

   Handle a CONNECT packet.

   Last checked:	2000-4-19 DAH

 */
static void
receive_connect (BConn* bconn, BPacket* packet, gpointer user_data)
{
  BUtilAddress* me_address = NULL;
  BUtilAddress* them_address = NULL;
  gchar* name = NULL;
  BPacket* new_packet;
  GList* i;

  DEBUGP ("receive_connect\n");

  /* Warn if we don't have handler for this protocol */
  if (b_accept_funcs[packet->protocol] == NULL)
    {
      g_warning ("Received CONNECT for unknown protocol %d\n", packet->protocol);
      goto drop;
    }

  /* Parse the packet */
  if (b_packet_parse_ok (packet, &me_address, &them_address, &name))
    {
      g_warning ("Recieved bad CONNECT\n");
      goto drop;
    }

  /* Make sure the sender isn't us */
  if (!strcmp(them_address->hostname, b_hostname) && 
      them_address->port == b_server->port)
    {
      g_warning ("Received bad CONNECT packet (from myself?)\n");
      goto drop;
    }

  /* Fix the hostname, port, and address of the connection */
  g_free (bconn->conn->hostname);
  bconn->conn->hostname = g_strdup (them_address->hostname);
  bconn->conn->port	 = them_address->port;

  gnet_inetaddr_delete (bconn->conn->inetaddr);
  bconn->conn->inetaddr = NULL;


  /* Save the protocol, object name, and ID. */
  bconn->protocol = packet->protocol;
  bconn->object_name = g_strdup (name);
  bconn->id_out = packet->id;

  /* Fix the ID of any packets in our packet queue */
  for (i = bconn->packet_queue; i != NULL; i = i->next)
    ((BPacket*) i->data)->id = bconn->id_out;

  DEBUGP ("bconn->id_out = %d\n", bconn->id_out);

  /* Dub it connected */
  bconn->status = B_CONN_STATUS_CONNECTED;

  /* Remove CONNECT handler */
  b_conn_handler_remove (bconn, B_PACKET_TYPE_CONNECT, -1, NULL);

  /* Get the tree.  This fails when we reset and another client tried
     to reconnect to us.  This fails when someone mistypes a URL. */
  switch (b_accept_funcs[packet->protocol](bconn, name))
    {
    case B_CONN_ACCEPT_OK:
      DEBUGP("B_CONN_ACCEPT_OK\n"); 		break;
    case B_CONN_ACCEPT_NO_OBJECT:
      DEBUGP("B_CONN_ACCEPT_NO_OBJECT\n");     	goto send_error;	break;
    case B_CONN_ACCEPT_ERROR:
      DEBUGP("B_CONN_ACCEPT_ERROR\n");		goto drop;		break;
    }

  /* Set the TOS */
  jmgnet_tcp_socket_set_tos (bconn->conn->socket, bconn->tos);

  /* Choose an ID for us */
  bconn->id_in = rand();

  DEBUGP ("conn->id_in = %d\n", bconn->id_in);

  /* Send an OK packet */
  new_packet = b_packet_new_ok_connect (bconn);
  b_conn_send_packet (bconn, new_packet);

  /* Start pinging the host */
  b_conn_ping_start (bconn);

  /* Call connect */
  b_conn_func_connect (bconn);

 done:
  /* Delete everything */
  b_util_address_delete (me_address);
  b_util_address_delete (them_address);
  g_free (name);
  return;

 send_error:

  /* Send an ERROR packet */
  new_packet = b_packet_new_error (bconn, B_PACKET_ERROR_SUBTYPE_CONNECT);
  b_conn_send_packet (bconn, new_packet);

  /* Close the connection */
  b_conn_close (bconn);

  goto done;

 drop:
  b_conn_unref (bconn);

  goto done;
}



static void
receive_connect_timeout (BConn* bconn, gint type, gint subtype, gpointer user_data)
{
  g_return_if_fail (type == B_PACKET_TYPE_CONNECT);

  DEBUGP ("receive_connect_timeout\n");

  b_conn_unref (bconn);
}
