/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdlib.h>

#include "globals.h"

#include "b_conn_connect.h"

#include "b_conn.h"
#include "b_conn_io.h"
#include "b_conn_handler.h"
#include "b_conn_ping.h"
#include "b_packet.h"


static gboolean connect_gconn_func (JMGConn* conn, JMGConnStatus status, 
				    gchar* buffer, gint length, gpointer user_data);

static void	receive_ok_connect (BConn* conn, BPacket* packet, gpointer user_data);
static void 	receive_error_connect (BConn* conn, BPacket* packet, gpointer user_data);
static void 	receive_ok_timeout (BConn* conn, gint type, gint subtype, gpointer user_data);



/* ************************************************************ */



/* 

   Connect to a host.

   Last checked: 2000-4-21 DAH

*/
void
b_conn_connect (BConn* bconn)
{
  g_return_if_fail (bconn);
  g_return_if_fail (bconn->conn);

  DEBUGP ("b_conn_connect %s, %d\n", 
	  bconn->conn->hostname, bconn->conn->port);

  /* Ignore if we aren't unconnected */
  if (!b_conn_is_unconnected(bconn))
    return;

  /* Double check: we shouldn't have a low-level connection */
  g_return_if_fail (!jm_conn_is_connected (bconn->conn));

  /* Dub is connecting */
  bconn->status = B_CONN_STATUS_CONNECTING;

  /* Set callbacks for low level connection */
  bconn->conn->func = connect_gconn_func;
  bconn->conn->user_data = bconn;

  /* Initiate low-level connection */
  jm_conn_connect (bconn->conn, 60000);
}



static gboolean
connect_gconn_func (JMGConn* conn, JMGConnStatus status, 
		    gchar* buffer, gint length, gpointer user_data)
{
  BConn* bconn = (BConn*) user_data;

  g_return_val_if_fail (bconn, FALSE);

  /* Don't need to call this again */
  conn->func = NULL;
  conn->user_data = NULL;

  switch (status)
    {
    case GNET_CONN_STATUS_CONNECT:
      {
	BPacket* packet;

	/* Set the TOS */
	jmgnet_tcp_socket_set_tos (conn->socket, bconn->tos);

	/* Set up watches */
	b_conn_watch_read (bconn);
	b_conn_watch_error (bconn);
	/* We will add the write watch when we call send_packet */

	/* Assign the conn an ID.  For now we just use a random
	   number, since the TCP connection is enough to identify it
	   after fully connecting.  */
	bconn->id_in = rand();

	DEBUGP ("connect_gconn_func: id_in = %d\n", bconn->id_in);

	/* Send a CONNECT packet */
	packet = b_packet_new_connect (bconn, bconn->object_name);
	b_conn_send_packet (bconn, packet);

	/* Handle an OK packet */
	b_conn_handler_add_full (bconn, B_PACKET_TYPE_OK, B_PACKET_OK_SUBTYPE_CONNECT,
				 receive_ok_connect, receive_ok_timeout,
				 RECEIVE_CONNECT_REPLY_TIME, NULL);

	/* Handle an ERROR packet (don't need a timeout - OK timeout is enough) */
	b_conn_handler_add (bconn, B_PACKET_TYPE_ERROR, B_PACKET_ERROR_SUBTYPE_CONNECT,
			    receive_error_connect, NULL);

	break;
      }

    case GNET_CONN_STATUS_ERROR:
      {
	/* Fail the node */
	b_conn_func_fail (bconn);

	break;
      }

    case GNET_CONN_STATUS_READ:
    case GNET_CONN_STATUS_WRITE:
    case GNET_CONN_STATUS_TIMEOUT:
    case GNET_CONN_STATUS_CLOSE:
      {
	g_assert_not_reached();
	break;
      }
    }

  return FALSE;
}



/* ************************************************************ */

void
b_conn_close (BConn* conn)
{
  BPacket* packet;

  g_return_if_fail (conn);

  DEBUGP ("b_conn_close\n");

  /* Close immediately if we aren't even connected */
  if (!b_conn_is_connected(conn))
    {
      b_conn_func_close (conn);
      return;
    }

  /* Stop pinging the host */
  b_conn_ping_stop (conn);

  /* TODO: We want sending the BYE to timeout */

  /* Ignore any packets we receive */
  b_conn_ignore_read (conn);

  /* Send a BYE */
  packet = b_packet_new_bye (conn);
  b_conn_send_packet (conn, packet);

  /* Set the status to closing */
  conn->status = B_CONN_STATUS_CLOSING;
}


/* ************************************************************ */


static void
receive_ok_connect (BConn* conn, BPacket* packet, gpointer user_data)
{
  GList* i;

  DEBUGP ("receive_ok_connect\n");

  /* Stop handling OK and ERROR  */
  b_conn_handler_remove (conn, B_PACKET_TYPE_OK, B_PACKET_OK_SUBTYPE_CONNECT, NULL);
  b_conn_handler_remove (conn, B_PACKET_TYPE_ERROR, B_PACKET_ERROR_SUBTYPE_CONNECT, NULL);

  /* The packet should have an ID in it */
  if (g_ntohs(packet->length) != sizeof(packet->id))
    {
      b_conn_func_fail (conn);
      return;
    }
  
  /* Save the ID of the conn */
  conn->id_out = *(guint32*) packet->data;

  /* Fix the ID of any packets in our packet queue */
  for (i = conn->packet_queue; i != NULL; i = i->next)
    ((BPacket*) i->data)->id = conn->id_out;

  /* Dub it connected */
  conn->status = B_CONN_STATUS_CONNECTED;

  /* Start sending pings */
  b_conn_ping_start (conn);

  /* Add a write watch if we have packets to send */
  if (conn->packet_queue)
    b_conn_watch_write (conn);

  /* Call up */
  b_conn_func_connect (conn);
}


static void
receive_error_connect (BConn* conn, BPacket* packet, gpointer user_data)
{
  DEBUGP ("receive_error_connect\n");

  b_conn_func_fail (conn);
}


static void
receive_ok_timeout (BConn* conn, gint type, gint subtype, gpointer user_data)
{
  DEBUGP ("receive_ok_timeout\n");

  b_conn_func_fail (conn);
}

