/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "globals.h"

#include "b_conn_handler.h"
#include "b_util.h"



typedef struct _BConnHandler
{
  BConn* conn;

  gint	type;
  gint	subtype;

  BConnHandlerReceiveFunc receive;

  BConnHandlerTimeoutFunc timeout;
  guint timer;

  gpointer user_data;

} BConnHandler;


static BConnHandler* b_conn_handler_get (BConn* conn, gint type, gint subtype, gpointer user_data);
static gboolean	handler_timeout (gpointer data);



/* ************************************************************ */



void
b_conn_handler_add (BConn* conn, gint type, gint subtype, 
		    BConnHandlerReceiveFunc receive, gpointer user_data)
{
  DEBUGP ("b_conn_handler_add %p, %d, %d, %p\n", conn, type, subtype, receive);

  b_conn_handler_add_full (conn, type, subtype, receive, NULL, 0, user_data);
}


void
b_conn_handler_add_full (BConn* conn, gint type, gint subtype, 
			 BConnHandlerReceiveFunc receive, 
			 BConnHandlerTimeoutFunc timeout, guint time,
			 gpointer user_data)
{
  BConnHandler* handler;

  g_return_if_fail (conn);
  g_return_if_fail (receive);

  DEBUGP ("b_conn_handler_add_full %p, %d, %d, %p\n", conn, type, subtype, receive);

  handler = b_conn_handler_get (conn, type, subtype, user_data);
  if (handler)
    {
      if (handler->receive != receive || handler->timeout != timeout)
	g_warning ("Attempt to replace handler for packet %d, %d\n",
		   type, subtype);
      return;
    }

  handler = g_new0 (BConnHandler, 1);
  handler->conn = conn;
  handler->type = type;
  handler->subtype = subtype;
  handler->receive = receive;
  handler->timeout = timeout;
  handler->user_data = user_data;

  if (timeout)
    b_util_timer_set (&handler->timer, time, 0, handler_timeout, handler);

  conn->handlers = g_list_prepend (conn->handlers, handler);
}



void
b_conn_handler_remove (BConn* conn, gint type, gint subtype, gpointer user_data)
{
  BConnHandler* handler;

  DEBUGP ("b_conn_handler_remove %p, %d, %d\n", conn, type, subtype);

  handler = b_conn_handler_get (conn, type, subtype, user_data);

  if (handler)
    {
      conn->handlers = g_list_remove (conn->handlers, handler);

      if (handler->timer)
	g_source_remove (handler->timer);

      g_free (handler);
    }
}



/* Return FALSE if conn is destroyed, TRUE if more packets can be
   dispatched. */
gboolean
b_conn_handler_dispatch (BConn* conn, BPacket* packet)
{
  GList* i;
  GList* handlers;
  gboolean found_handler = FALSE;

  g_return_val_if_fail (conn, FALSE);
  g_return_val_if_fail (packet, FALSE);

  DEBUGP ("b_conn_handler_dispatch %d, %d\n", packet->type, packet->subtype);

  handlers = g_list_copy (conn->handlers);

  for (i = handlers; i != NULL; i = i->next)
    {
        BConnHandler* handler = (BConnHandler*) i->data;

	g_return_val_if_fail (handler, FALSE);

	/* Skip if the handler is no long in the handler list */
	if (!g_list_find (conn->handlers, handler))
	  continue;

	/* Dispatch if this is the appropriate handler. */
	if ((handler->type == -1 || handler->type == packet->type) &&
	    (handler->subtype == -1 || handler->subtype == packet->subtype))
	  {
	    found_handler = TRUE;

	    /* Cancel the timeout */
	    b_util_timer_cancel (&handler->timer);

	    /* Call the function.  We ref the connection so that we
               can tell if the connection was deleted. */
	    b_conn_ref (conn);
	    handler->receive (conn, packet, handler->user_data);
	    if (--conn->ref_count == 0)
	      {
		b_conn_delete (conn);
		g_list_free (handlers);
		return FALSE;
	      }
	  }
    }

  g_list_free (handlers);

  if (!found_handler)
    g_warning ("Couldn't find handler for packet of type %d, %d\n", 
	       packet->type, packet->subtype);

  return TRUE;
}


void
b_conn_handler_remove_all (BConn* conn)
{
  GList* i;

  for (i = conn->handlers; i != NULL; i = i->next)
    {
        BConnHandler* handler = (BConnHandler*) i->data;

	if (handler->timer)
	  g_source_remove (handler->timer);

	g_free (handler);
    }
  
  g_list_free (conn->handlers);
  conn->handlers = NULL;
}


gboolean
b_conn_handler_have (BConn* conn, gint type, gint subtype, gpointer user_data)
{
  return b_conn_handler_get (conn, type, subtype, user_data) != NULL;
}


static BConnHandler*
b_conn_handler_get (BConn* conn, gint type, gint subtype, gpointer user_data)
{
  GList* i;

  for (i = conn->handlers; i != NULL; i = i->next)
    {
        BConnHandler* handler = (BConnHandler*) i->data;

      if (handler->type == type && 
	  handler->subtype == subtype && 
	  handler->user_data == user_data)
	return handler;
    }


  return NULL;
}


static gboolean
handler_timeout (gpointer data)
{
  BConnHandler* handler = (BConnHandler*) data;

  handler->conn->handlers = g_list_remove (handler->conn->handlers, handler);

  handler->timeout (handler->conn, handler->type, handler->subtype, handler->user_data);

  g_free (handler);

  return FALSE;
}

