/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "globals.h"

#include "b_conn_io.h"
#include "b_conn.h"
#include "b_conn_handler.h"
#include "b_conn_accept.h"
#include "b_packet.h"


static gboolean read_cb  (GIOChannel* iochannel, GIOCondition condition, gpointer data);
static gboolean write_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data);
static gboolean error_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data);

static gint	scheduler_write_cb (gpointer user_data);

static gboolean packet_read (BConn* conn, gchar* buffer, gint length, GList** packetsp, guint* bytes_read);
static gboolean	process_packets (BConn* conn, GList* packets);



/* ************************************************************ */


void
b_conn_watch_read (BConn* conn)
{
  g_return_if_fail (conn);

  DEBUGP ("b_conn_watch_read\n");

  jm_conn_watch_add_read (conn->conn, read_cb, conn);
}


void
b_conn_watch_write (BConn* conn)
{
  g_return_if_fail (conn);

  DEBUGP ("b_conn_watch_write\n");

  jm_conn_watch_add_write (conn->conn, write_cb, conn);
}


void
b_conn_watch_error (BConn* conn)
{
  g_return_if_fail (conn);

  DEBUGP ("b_conn_watch_error\n");

  jm_conn_watch_add_error (conn->conn, error_cb, conn);
}


void
b_conn_ignore_read (BConn* conn)
{
  g_return_if_fail (conn);

  DEBUGP ("b_conn_ignore_read\n");

  jm_conn_watch_remove_read (conn->conn);
}


void
b_conn_ignore_write (BConn* conn)
{
  g_return_if_fail (conn);

  DEBUGP ("b_conn_ignore_write\n");

  jm_conn_watch_remove_write (conn->conn);
}


void
b_conn_ignore_error (BConn* conn)
{
  g_return_if_fail (conn);

  DEBUGP ("b_conn_ignore_error\n");

  jm_conn_watch_remove_error (conn->conn);
}


void	
b_conn_ignore_all (BConn* conn)
{
  g_return_if_fail (conn);

  DEBUGP ("b_conn_ignore_all\n");

  b_conn_ignore_read  (conn);
  b_conn_ignore_write (conn);
  b_conn_ignore_error (conn);
}


/* ************************************************************ */


/**

   read_cb

   Called when an iochannel is readable.  Read from the iochannel,
   parse the data, then pass the packets on.

   Last checked: 	2000-4-20 DAH

 */
static gboolean
read_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data)
{
  BConn* conn = (BConn*) data;
  GIOError error;
  guint bytes_read;
  guint bytes_parsed = 0;
  GList* packets = NULL;
  
  g_return_val_if_fail (conn, FALSE);
  g_return_val_if_fail (!b_conn_is_closing (conn), FALSE);

  DEBUGP ("read_cb: buffer_length = %d\n", conn->buffer_length);

  /* Read stuff into our buffer */
  error = g_io_channel_read (iochannel, &conn->buffer[conn->buffer_length], 
			     sizeof(conn->buffer) - conn->buffer_length, 
			     &bytes_read);

  /* If we read EOF or got an error, then the connection has failed
     (we should have received a BYE) */
  if ((error == G_IO_ERROR_NONE && bytes_read == 0) ||
      (error != G_IO_ERROR_NONE && error != G_IO_ERROR_AGAIN))
    {
      DEBUGP ("read_cb: error %d, bytes_read %d\n", error, bytes_read);
      b_conn_func_fail (conn);
      return FALSE;
    }

  DEBUGP ("read_cb: bytes_read = %d\n", bytes_read);

  /* Update buffer length */
  conn->buffer_length += bytes_read;

  /* Parse it */
  if (packet_read (conn, conn->buffer, conn->buffer_length, &packets, &bytes_parsed))
    {
      /* Dump the connection if it sent us garbage */
      b_conn_func_fail (conn);
      return FALSE;
    }

  DEBUGP ("read_cb: bytes_parsed = %d\n", bytes_parsed);

  /* Move data over based on how much was read */
  if (bytes_parsed > 0)
    {
      g_memmove(conn->buffer, &conn->buffer[bytes_parsed], 
		conn->buffer_length - bytes_parsed);
      conn->buffer_length -= bytes_parsed;
    }

  /* Pass the packets to the host */
  return process_packets (conn, packets);
}



/**

   write_cb

   Called when an iochannel is writable.  If there are no packets to
   send do an upcall.  If there's a packet to send, send it.
   Otherwise, we don't want any more callbacks.

   Last checked: 	2000-4-20 DAH

 */
gboolean 
write_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data)
{
  BConn* conn = (BConn*) data;

  g_return_val_if_fail (conn != NULL, FALSE);

  DEBUGP ("write_cb %p\n", conn);


  /* If there are no data packets to send, call the write callbacks -
     they may give us a data packet. */
  if (b_conn_is_connected (conn) && 
      !(b_conn_has_packet (conn, B_PACKET_TYPE_MCAST) ||
	b_conn_has_packet (conn, B_PACKET_TYPE_UCAST)) )
    {
      b_conn_func_writeable (conn);
    }

  conn->conn->write_watch = 0;

  /* If we have a packet, schedule a write */
  if (conn->packet_queue)
    scheduler_write_cb (conn);
/*      scheduler_add (b_conn_scheduler_write, conn, scheduler_write_cb, 0); */

  /* Otherwise, if the host is being closed, close it */
  else if (b_conn_is_closing (conn))
    b_conn_func_close (conn);

  /* Otherwise, we just don't have any packets to send */

  return FALSE;
}



static gint
scheduler_write_cb (gpointer user_data)
{
  BConn* conn = (BConn*) user_data;
  BPacket* packet;

  /* Get the next packet from the conn */
  packet = b_conn_pop_packet (conn);

  /* If we have a packet, send it.  We have to check because we may
     have removed a packet since we scheduled a write. */
  if (packet != NULL)
    {
      gchar* buffer;
      guint length;
      guint bytes_writen;
      GIOError error;

      /* TODO:  Make this non-blocking */

      /* Write the top packet out */
      buffer = (gchar*) packet;
      length = B_PACKET_LENGTH(packet);

      DEBUGP ("write_cb: packet %p\n", packet);
      DEBUGP ("write_cb: write %d, %d (length %d, id %d)\n", packet->type, packet->subtype, 
	      length, packet->id);

      error = gnet_io_channel_writen (conn->conn->iochannel, buffer, length, &bytes_writen);

      g_return_val_if_fail (error == G_IO_ERROR_NONE, FALSE);
      g_return_val_if_fail (length == bytes_writen, FALSE);

      /* Delete the packet */
      b_packet_delete (packet);

      /* Add a write watch */
      b_conn_watch_write (conn);

      return (length * 8);
    }

  /* Otherwise, there's nothing to send close the host if necessary */
  else if (b_conn_is_closing (conn))
    {
      b_conn_func_close (conn);
    }

  return 0;
}




/* 

   error_cb

   Called when a socket has an ERR or HUP.

   Last checked: 	2000-4-21 DAH

 */
gboolean
error_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data)
{
  BConn* conn = (BConn*) data;

  DEBUGP ("error_cb\n");

  if (condition == G_IO_NVAL)
    DEBUGP ("error_cb G_IO_NVAL\n");

  b_conn_func_fail (conn);

  return FALSE;
}


/* ************************************************************ */


/* Return FALSE if OK, TRUE if error */
gboolean
packet_read (BConn* conn, gchar* buffer, gint length, GList** packetsp, guint* bytes_read)
{
  GList* packets = NULL;

  g_return_val_if_fail (buffer, TRUE);
  g_return_val_if_fail (length, TRUE);
  g_return_val_if_fail (packetsp, TRUE);
  g_return_val_if_fail (bytes_read, TRUE);

  /* Initialize return values */
  *packetsp = NULL;
  *bytes_read = 0;


  /* Read any remaining bytes for a previous variable length packet */
  if (conn->current_packet)
    {
      BPacket* packet = NULL;
      guint16 data_length;
      guint16 amount_to_read;

      packet = conn->current_packet;
      data_length = g_ntohs(packet->length);

      amount_to_read = MIN(data_length - conn->total_read, length);
      g_return_val_if_fail ((conn->total_read + amount_to_read) <= data_length, TRUE);

      /* Copy the data */
      memcpy (&packet->data[conn->total_read], buffer, amount_to_read);

      /* Update length and buffer */
      length -= amount_to_read;
      buffer = &buffer[amount_to_read];
      *bytes_read += amount_to_read;

      /* Update total read */
      conn->total_read += amount_to_read;

      /* Add to list of packets if done */
      if(conn->total_read == data_length)
	{
	  packets = g_list_prepend (packets, conn->current_packet);
	  conn->current_packet = NULL;
	  conn->total_read = 0;
	}
    }


  while (length >= sizeof(BPacket))
    {
      BPacket* packet;
      guint16 data_length;

      BPacket* new_packet = NULL;
      gint new_packet_length = 0;

      guint16 amount_to_copy;

      /* Get the packet and length */
      packet = (BPacket*) buffer;
      data_length = g_ntohs (packet->length);

      /* Make sure the version is right */
      if (packet->version != B_PACKET_VERSION)
	{
	  g_warning ("Received packet with bad version (%d)\n", 
		     packet->version);
	  goto error;
	}

      /* Make sure the ID is right */
      if (packet->id != conn->id_in && 
	  packet->type != B_PACKET_TYPE_CONNECT)
	{
	  g_warning ("Received packet with ID mismatch (type is %d, %d, id is %d, should be %d)\n",
		     packet->type, packet->subtype, packet->id, conn->id_in);
	  goto error;
	}


      /* Create a new packet */
      new_packet_length = sizeof(BPacket) + data_length;
      new_packet = (BPacket*) g_new0 (gchar, new_packet_length);
      DEBUGP ("packet_read: new_packet %p\n", new_packet);

      /* Copy the header */
      memcpy ((char*) new_packet, buffer, sizeof(BPacket));
	    
      /* Copy the data */
      amount_to_copy = MIN(data_length, length - sizeof(BPacket));
      memcpy (new_packet->data, packet->data, amount_to_copy);

      /* Save packet and break if we're expecting more data */
      if (amount_to_copy != data_length)
	{
	  g_return_val_if_fail ((sizeof(BPacket) + amount_to_copy) == length, TRUE);
	  g_return_val_if_fail (conn->current_packet == NULL, TRUE);

	  conn->current_packet = (BPacket*) new_packet;
	  conn->total_read = amount_to_copy;
	  *bytes_read += sizeof(BPacket) + amount_to_copy;

	  break;
	}
      /* Otherwise, we read the whole packet */

      /* Add packet to our list */
      packets = g_list_append (packets, new_packet);

      /* Update length, buffer, bytes_read */
      length -= new_packet_length;
      buffer = &buffer[new_packet_length];
      *bytes_read += new_packet_length;

      /* If it's a connect, break.  This is so we can finish
         initializing the connection */
      if (new_packet->type == B_PACKET_TYPE_CONNECT)
	break;
    }

  *packetsp = packets;
  return FALSE;

 error:
  *bytes_read += length;
  return TRUE;

}



static gboolean
process_packets (BConn* conn, GList* packets)
{
  g_return_val_if_fail (conn, FALSE);

  while (packets != NULL)
    {
      BPacket* packet;
      gboolean more;

      /* Pop off the top packet */
      packet = packets->data;      /* process_packet owns */
      packets = g_list_remove_link (packets, packets);

      /* The packet should have the right ID (assuming we didn't just
         accept it) */
      if (!b_conn_is_accepted(conn) && packet->id != conn->id_in)
	{
	  g_warning ("Packet has wrong id (is %d, should be %d)\n",
		     packet->id, conn->id_in);
	}

      /* Dispatch the packet */
      more = b_conn_handler_dispatch (conn, packet);

      /* Delete this packet */
      b_packet_delete (packet);

      /* If the connection has been destroyed, then free the packets
         and return FALSE. */
      if (!more)
	{
	  GList* l;

	  /* Delete the rest of the packets */
	  for (l = packets ; l != NULL; l = l->next)
	    b_packet_delete ((BPacket*) l->data);
	  g_list_free (packets);

	  return FALSE;
	}
    }

  return TRUE;
}

