/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#include "globals.h"

#include "b_packet.h"
#include "b_conn_accept.h"
#include "b_util.h"



BPacket* 
b_packet_new (BConn* conn, guint8 type, guint16 length)
{
  BPacket* packet;

  g_return_val_if_fail (conn, NULL);

  packet = (BPacket*) g_new0 (guchar, sizeof(BPacket) + length);

  packet->protocol = conn->protocol;
  packet->version = B_PACKET_VERSION;
  packet->id = conn->id_out;
  packet->type = type;
  packet->length = g_htons (length);

  DEBUGP ("b_packet_new %p\n", packet);

  return packet;
}


BPacket* 
b_packet_clone (BPacket* packet)
{
  BPacket* new_packet;
  gint size;

  g_return_val_if_fail (packet, NULL);

  size = sizeof(BPacket) + g_ntohs(packet->length);

  new_packet = (BPacket*) g_new0 (guchar, size);
  memcpy (new_packet, packet, size);

  return packet;
}


void
b_packet_delete (BPacket* packet)
{
  DEBUGP ("b_packet_delete %p\n", packet);

  if (packet)
    g_free (packet);
}


BPacket*
b_packet_new_ok (BConn* conn, BPacketOkSubtype subtype)
{
  BPacket* packet;

  g_return_val_if_fail (conn != NULL, NULL);

  packet = b_packet_new (conn, B_PACKET_TYPE_OK, 0);
  packet->subtype = subtype;

  return packet;
}



BPacket*
b_packet_new_ok_connect (BConn* conn)
{
  BPacket* packet;

  g_return_val_if_fail (conn != NULL, NULL);

  packet = b_packet_new (conn, B_PACKET_TYPE_OK, sizeof(guint32));
  packet->subtype = B_PACKET_OK_SUBTYPE_CONNECT;
  *(guint32*) packet->data = conn->id_in;

  return packet;
}



BPacket*
b_packet_new_error (BConn* conn, BPacketErrorSubtype subtype)
{
  BPacket* packet;

  g_return_val_if_fail (conn != NULL, NULL);

  packet = b_packet_new (conn, B_PACKET_TYPE_ERROR, 0);
  packet->subtype = subtype;

  return packet;
}



BPacket*
b_packet_new_connect (BConn* bconn, gchar* name)
{
  BPacket* packet;
  gchar* me;
  gchar* them;
  GList* strings = NULL;
  gchar* str;
  guint length;

  g_return_val_if_fail (bconn, NULL);
  g_return_val_if_fail (bconn->conn, NULL);
  g_return_val_if_fail (bconn->conn->hostname, NULL);
  g_return_val_if_fail (b_server, NULL);
  g_return_val_if_fail (b_hostname, NULL);
  g_return_val_if_fail (name, NULL);

  /* Send:
       Their hostname:port
       My hostname:port
       Name
  */
  them = g_strdup_printf ("%s:%d", bconn->conn->hostname, bconn->conn->port);
  me =   g_strdup_printf ("%s:%d", b_hostname, b_server->port);

  strings = g_list_prepend (strings, name);
  strings = g_list_prepend (strings, me);
  strings = g_list_prepend (strings, them);

  str = b_util_write_strings (strings, &length);

  packet = b_packet_new (bconn, B_PACKET_TYPE_CONNECT, length);
  packet->id = bconn->id_in;	/* We send out id in first */
  memcpy (packet->data, str, length);

  g_free (str);
  g_free (me);
  g_free (them);
  g_list_free (strings);

  return packet;
}



BPacket*
b_packet_new_bye (BConn* conn)
{
  BPacket* packet;

  g_return_val_if_fail (conn != NULL, NULL);

  packet = b_packet_new (conn, B_PACKET_TYPE_BYE, 0);

  return packet;
}



BPacket* 
b_packet_new_ucast (BConn* conn, guint8* buffer, guint16 length)
{
  BPacket* packet;

  packet = b_packet_new (conn, B_PACKET_TYPE_UCAST, length);
  memcpy (packet->data, buffer, length);

  return packet;
}



BPacket* 
b_packet_new_mcast (BConn* conn, guint8* buffer, guint16 length)
{
  BPacket* packet;

  packet = b_packet_new (conn, B_PACKET_TYPE_MCAST, length);
  memcpy (packet->data, buffer, length);

  return packet;
}



BPacket* 
b_packet_new_info_request_ping (BConn* conn)
{
  BPacket* packet;

  packet = b_packet_new (conn, B_PACKET_TYPE_INFO_REQUEST, sizeof(struct timeval));
  packet->subtype = B_PACKET_INFO_SUBTYPE_PING;
  /* We will set the timeval just before we send it */

  return packet;
}



BPacket* 
b_packet_new_info_reply_ping (BConn* conn, BPacket* ping)
{
  BPacket* packet;
  guint length;

  g_return_val_if_fail (conn != NULL, NULL);
  g_return_val_if_fail (ping != NULL, NULL);

  length = g_ntohs (ping->length);

  packet = b_packet_new (conn, B_PACKET_TYPE_INFO_REPLY, length);
  packet->subtype = B_PACKET_INFO_SUBTYPE_PING;
  memcpy (packet->data, ping->data, length);

  return packet;
}



gboolean	/* Return FALSE if OK */ 
b_packet_parse_ok (BPacket* packet, BUtilAddress** me_addressp, 
		   BUtilAddress** them_addressp, gchar** object_namep)
{
  GList* strings = NULL;
  GList* i;
  gboolean rv = TRUE;

  g_return_val_if_fail (packet, TRUE);
  g_return_val_if_fail (me_addressp, TRUE);
  g_return_val_if_fail (them_addressp, TRUE);
  g_return_val_if_fail (object_namep, TRUE);

  *me_addressp = NULL;
  *them_addressp = NULL;
  *object_namep = NULL;

  /* Read the 3 strings.  This should not fail. */
  strings = b_util_read_strings (packet->data, g_ntohs(packet->length));
  if (g_list_length(strings) != 3 || !strings->next->next->data)
    goto error;

  /* Parse the me address */
  *me_addressp = b_util_string_to_address ((gchar*) strings->data);
  if (!*me_addressp || (*me_addressp)->port == 0)
    goto error;

  /* Parse the them address */
  *them_addressp = b_util_string_to_address ((gchar*) strings->next->data);
  if (!*them_addressp || (*them_addressp)->port == 0)
    goto error;

  /* Copy the object name */
  *object_namep = strings->next->next->data;
  strings->next->next->data = NULL;

  rv = FALSE;

 done:
  for (i = strings; i != NULL; i = i->next)
    g_free ((gchar*) i->data);
  
  return rv;

 error:
  *me_addressp = NULL;
  *them_addressp = NULL;
  *object_namep = NULL;

  goto done;
}
