/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include <gnet/gnet.h>

#include "b_util.h"



/* **************************************** */

BConn*
b_util_find (GList* bconns, const gchar* hostname, const gint port)
{
  for (; bconns != NULL; bconns = bconns->next)
    {
      BConn* bconn = (BConn*) bconns->data;

      g_return_val_if_fail (bconn, NULL);
      g_return_val_if_fail (bconn->conn, NULL);
      g_return_val_if_fail (bconn->conn->hostname, NULL);

      if (port == bconn->conn->port &&
	  strcmp(hostname, bconn->conn->hostname) == 0)
	return bconn;
    }

  return NULL;
}



/* **************************************** */


/**

   Read a single NULL terminated string.

*/
gchar*
b_util_read_string (const gchar* buffer, const guint length)
{
  GList* list;
  gchar* str = NULL;

  g_return_val_if_fail (buffer != NULL, NULL);

  list = b_util_read_strings (buffer, length);

  if (list != NULL)
    {
      GList* i;

      str = (gchar*) list->data;
      list = g_list_remove_link (list, list);

      for (i = list; i != NULL; i = i->next)
	g_free ((gchar*) i->data);
      g_list_free(list);
    }

  return str;
}



/**

   Read a list of NULL terminated strings.

 */
GList* 
b_util_read_strings (const gchar* buffer, const guint length)
{
  guint i = 0;
  GList* list = NULL;

  g_return_val_if_fail (buffer != NULL, NULL);

  while (1)
    {
      const gchar* begin;
      gchar* str;

      begin = &buffer[i];

      while (i < length && buffer[i] != '\0') 
	++i;

      if (i == length)
	break;

      str = g_strndup (begin, &buffer[i] - begin);
      list = g_list_append (list, str);
      ++i;
    }

  return list;
}



/**

   Write a list of NULL terminated strings.

*/
gchar*
b_util_write_strings (GList* strings, guint* len)
{
  gchar* str = NULL;
  GList* i;
  gint length;

  if (strings == NULL)
    {
      *len = 0;
      return NULL;
    }

  length = 0;
  for (i = strings; i != NULL; i = i->next)
    length += (strlen((char*)i->data) + 1);

  str = g_new (gchar, length);

  length = 0;
  for (i = strings; i != NULL; i = i->next)
    {
      gchar* s = (gchar*) i->data;

      strcpy (&str[length], s);
      length += strlen(s);
      str[length] = '\0';
      length += 1;
    }
  
  *len = length;

  return str;
}


/* **************************************** */

void
b_util_address_delete (BUtilAddress* addr)
{
  if (addr)
    {
      g_free (addr->hostname);
      g_free (addr);
    }
}



BUtilAddress*
b_util_string_to_address (const gchar* str)
{
  gchar* server;
  const gchar* port_str = NULL;
  gint port = 0;
  BUtilAddress* address;

  if (str == NULL || str[0] == '\0')
    return NULL;

  server = (gchar*) str;
  while (*str != ':' && *str != '\0') ++str;
  if (server == str || *str == '\0') return NULL;

  server = g_strndup (server, str - server);

  port_str = ++str;
  while (*str != '\0') ++str;
  if (port_str == str) { g_free(server); return NULL; }
  port = atoi(port_str);

  address = g_new0 (BUtilAddress, 1);
  address->hostname = server;
  address->port = port;

  return address;
}


BUtilAddress*
b_util_read_address (const gchar* buffer, const guint length)
{
  gchar* str;
  BUtilAddress* address = NULL;

  g_return_val_if_fail (buffer != NULL, NULL);

  str = b_util_read_string (buffer, length);
  if (str != NULL)
    {
      address = b_util_string_to_address (str);
      g_free (str);
    }

  return address;
}


GList* 
b_util_read_addresses (const gchar* data, const guint length)
{
  GList* strings;
  GList* i;
  GList* addresses = NULL;

  strings = b_util_read_strings (data, length);

  for (i = strings; i != NULL; i = i->next)
    {
      gchar* str = (gchar*) i->data;
      BUtilAddress* address;

      address = b_util_string_to_address (str);
      if (address != NULL)
	addresses = g_list_prepend(addresses, address);

      g_free(str);
    }

  g_list_free(strings);

  return addresses;
}


gchar* 
b_util_write_address (BConn* conn, guint* len)
{
  GList list_conn;

  list_conn.data = conn;
  list_conn.next = NULL;
  list_conn.prev = NULL;

  return b_util_write_addresses (&list_conn, len);
}


gchar* 
b_util_write_addresses (GList* bconns, guint* len)
{
  GList* strings = NULL;
  GList* i;
  gchar* s = NULL;

  for (i = bconns; i != NULL; i = i->next)
    {
      BConn* bconn;
      gchar* str;

      bconn = (BConn*) i->data;
      g_return_val_if_fail (bconn, NULL);
      g_return_val_if_fail (bconn->conn, NULL);
      g_return_val_if_fail (bconn->conn->hostname, NULL);

      str = g_strdup_printf ("%s:%d", 
			     bconn->conn->hostname, bconn->conn->port);
      strings = g_list_prepend (strings, str);
    }

  strings = g_list_reverse (strings);
  s = b_util_write_strings (strings, len);

  for (i = strings; i != NULL; i = i->next)
    g_free ((gchar*) i->data);
  g_list_free (strings);

  return s;
}



/* **************************************** */

void
b_util_timer_set (guint* timer, guint time, guint delay, 
		  GSourceFunc func, gpointer user_data)
{
  g_return_if_fail (timer);
  g_return_if_fail (func);

  if (!*timer)
    {
      if (delay)
	time += rand() % delay;

      *timer = g_timeout_add (time, func, user_data);
    }
}


void
b_util_timer_cancel (guint* timer)
{
  g_return_if_fail (timer);

  if (*timer)
    {
      g_assert (g_source_remove(*timer));
      *timer = 0;
    }
}


void
b_util_timer_reset (guint* timer, guint time, guint delay, 
		    GSourceFunc func, gpointer user_data)
{
  b_util_timer_cancel (timer);
  b_util_timer_set (timer, time, delay, func, user_data);
}
