/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <string.h>
#include <errno.h>

#include "backend.h"
#include "backend_announce.h"
#include "gui.h"
#include "util.h"
#include "globals.h"
#include "jmpio.h"
#include "btp.h"
#include "b_conn_accept.h"	/* :-( */


/* Globals */
char* jm_home = NULL;
int   debug_level = 0;
int   verbose = 0;

static GHashTable*	url_to_file = NULL;	
static gboolean 	backend_initialized = FALSE;

static struct timeval 	timeofday;
static guint 		gc_timer = 0;


static void backend_shutdown_hfunc (gpointer key, gpointer value, gpointer user_data);
static void backend_delete_hfunc (gpointer key, gpointer value, gpointer user_data);

static gboolean garbage_collect_cb(gpointer user_data);
static gboolean	garbage_collect_hrfunc(gpointer key, gpointer value, gpointer user_data);

gchar*	   backend_hostname = NULL;
gint	   backend_port = 0;
GInetAddr* backend_interface = NULL;
MtpServer* backend_mtp_server = NULL;

JMGURL*	   backend_rendezvous_url = NULL;
gboolean   backend_rendezvous_locally = TRUE;
gboolean   backend_auto_mirror_on_download = TRUE;



/* ************************************************************ */


/* Start the backend.  Returns FALSE on success; TRUE otherwise. */
gboolean
backend_init (gchar* hostname, GInetAddr* interface, 
	      gint btp_port, gboolean force_btp_port,
	      gint mtp_port, gboolean force_mtp_port,
	      gboolean global_rendezvous)
{
  g_return_val_if_fail (!backend_initialized, TRUE);

  /* Initialize the RNG */
  srand(time(NULL));

  /* Create url_to_file hashtable */
  if (!url_to_file)
    url_to_file = g_hash_table_new (g_str_hash, g_str_equal);

  /* Start garbage collection */
  if (!gc_timer)
    gc_timer = g_timeout_add (GARBAGE_COLLECT_TIME, garbage_collect_cb, NULL);

  /* Return if no hostname or interface */
  if (!hostname || !interface)
    return TRUE;

  /* Initialize the BTP module */
  if (btp_init(hostname, interface, btp_port, force_btp_port))
    return TRUE;

  /* Start the BTP server as a mirror */
  backend_mtp_server = mtp_server_new (hostname, interface, 
				       mtp_port, force_mtp_port, 
				       MTP_SERVER_TYPE_BOTH, 
				       global_rendezvous);
  if (!backend_mtp_server)
    return TRUE;

  backend_hostname = g_strdup (hostname);
  backend_port = b_server->port;
  backend_interface = gnet_inetaddr_clone (interface);

  backend_initialized = TRUE;

  return FALSE;
}



/* Shutdown the backend. */
gboolean
backend_shutdown (void)
{
  /* Set backend_initialized to FALSE */
  g_return_val_if_fail (backend_initialized, TRUE);
  backend_initialized = FALSE;

  /* Stop garbage collection */
  if (gc_timer)
    {
      g_assert (g_source_remove(gc_timer));
      gc_timer = 0;
    }

  gui_freeze();

  /* Move files from url_to_file to files */
  if (url_to_file)
    {
      g_hash_table_foreach (url_to_file, backend_shutdown_hfunc, NULL);
      g_hash_table_foreach (url_to_file, backend_delete_hfunc, NULL);
      g_hash_table_destroy (url_to_file);
      url_to_file = NULL;
    }

  gui_thaw();


  /* This isn't right.  We don't know if BTP has truely shutdown.
     Eventually, we will check this in backend_stop_r or something...
     For now we assume that we need more time.  */

  /* Shutdown the BTP module */
  btp_shutdown();

  /* Shutdown MTP server */
  if (backend_mtp_server)
    {
      mtp_server_delete (backend_mtp_server);
      backend_mtp_server = NULL;
    }

  jm_url_delete (backend_rendezvous_url);
  backend_rendezvous_url = NULL;

  g_free (backend_hostname);	
  backend_hostname = NULL;
  gnet_inetaddr_delete (backend_interface);
  backend_interface = NULL;

  return FALSE;
}


static void
backend_shutdown_hfunc (gpointer key, gpointer value, gpointer user_data)
{
  JMFile* jmfile;

  jmfile = (JMFile*) value;
  g_return_if_fail (jmfile);

  gui_remove (jmfile);

  /* Send remove announcements for each local channel */
  if (jmfile->type == JMFILE_CHANNEL && jmfile->is_local)
    {
      GList* i;
	  
      for (i = jmfile->parents; i != NULL; i = i->next)
	{
	  JMFile* parent = (JMFile*) i->data;

	  backend_announce_remove (jmfile, parent);
	}
    }
}


static void
backend_delete_hfunc (gpointer key, gpointer value, gpointer user_data)
{
  JMFile* jmfile;

  jmfile = (JMFile*) value;
  g_return_if_fail (jmfile);

  jmfile_delete (jmfile);
}



gboolean
backend_is_initialized (void)
{
  return backend_initialized;
}



JMFile* 
backend_get (gchar* url)
{
  g_return_val_if_fail (url_to_file, NULL);
  g_return_val_if_fail (url, NULL);

  return (JMFile*) g_hash_table_lookup(url_to_file, url);
}



void
backend_put (JMFile* jmfile)
{
  g_return_if_fail (url_to_file);
  g_return_if_fail (jmfile);
  g_return_if_fail (jmfile->url);
  g_return_if_fail (backend_get(jmfile->url) == NULL);

  g_hash_table_insert(url_to_file, jmfile->url, jmfile);
}


void
backend_delete (JMFile* jmfile)
{
  g_return_if_fail (jmfile);

  if (url_to_file == NULL)
    return;

  if (!jmfile->url)	/* directories don't have URL's */
    return;

  /* Remove from announcer */
  backend_announce_delete (jmfile);

  g_return_if_fail (backend_get(jmfile->url));

  g_hash_table_remove (url_to_file, jmfile->url);
}


/* ******************** */


static gboolean
garbage_collect_cb (gpointer user_data)
{
  DEBUGP("garbage_collect_cb\n");

  g_return_val_if_fail (url_to_file, TRUE);

  gettimeofday (&timeofday, NULL);
  gui_freeze ();
  g_hash_table_foreach_remove (url_to_file, garbage_collect_hrfunc, NULL);
  gui_thaw ();

  return TRUE;
}


static gboolean
garbage_collect_hrfunc (gpointer key, gpointer value, gpointer user_data)
{
  JMFile* jmfile;
  struct timeval diff;

  jmfile = (JMFile*) value;
  g_return_val_if_fail (jmfile, FALSE);

  /* Ignore if local, or has a NULL parent */
  if (jmfile->is_local || jmfile_has_parent(jmfile, NULL))
    return FALSE;

  /* If older than EXPIRE_TIME, remove.  We can't remove recursively
     unfortunately. */
  timersub(&timeofday, &jmfile->last_update, &diff);
  if (diff.tv_sec > (EXPIRE_TIME / 1000))
    {
      DEBUGP ("garbage_collect_hrfunc: expire\n");
      gui_remove (jmfile);
      jmfile_delete (jmfile);
      return TRUE;
    }

  return FALSE;
}



/* ************************************************************ */



void
backend_unlink (JMFile* file, JMFile* parent)
{
  g_return_if_fail (file != NULL);

  jmfile_unlink (file, parent);
}


/*

  Called by internal backend function to add a file.

 */
JMFile*
backend_add (gchar* name, gchar* url, JMFileType type, JMFile* parent)
{
  JMFile* jmfile;

  g_return_val_if_fail (url, NULL);

  jmfile = backend_get (url);
  if (jmfile != NULL)
    {
      jmfile_link (jmfile, parent);
      gui_update (jmfile);
    }
  else
    {
      jmfile = jmfile_new (name, type, url);
      jmfile_link (jmfile, parent);

      backend_put (jmfile);
      if (parent)
	backend_announce_add (jmfile, parent);

      gui_add(jmfile);
    }

  return jmfile;
}



/*

  Called by GUI to remove file (and any children).  The file may be of
  any type.

 */
void
backend_remove (JMFile* file) /* gboolean local? */
{
  GList* i;

  g_return_if_fail (file != NULL);
  g_return_if_fail (file->is_local);

  gui_freeze();

  /* Send an update for each parent */
  for (i = file->parents; i != NULL; i = i->next)
    {
      JMFile* parent = (JMFile*) i->data;
      
      backend_announce_remove (file, parent);
    }

  /* Unlink each child */
  for (i = file->children; i != NULL; )
    {
      JMFile* child = (JMFile*) i->data;
      i = i->next;
      
      if (child != NULL)
	jmfile_unlink (child, file);
    }

  /* Unlink each parent */
  for (i = file->parents; i != NULL; )
    {
      JMFile* parent = (JMFile*) i->data;
      i = i->next;

      if (!jmfile_unlink (file, parent))
	goto done;
    }

  /* We may have NULL parents, so manually delete the file here. */
  gui_remove (file);
  backend_delete (file);
  jmfile_delete (file);

 done:
  gui_thaw();
}




void
backend_activate(JMFile* jmfile)
{
  g_return_if_fail (jmfile != NULL);

  if ((jmfile->type == JMFILE_FILE || jmfile->type == JMFILE_CHANNEL) &&
      (jmfile->url == NULL))
    {
      g_warning ("Don't have URL for %s\n", jmfile->name);
      return;
    }

  switch (jmfile->type)
    {
    case JMFILE_FILE:	 backend_file_download (jmfile);	break;
    case JMFILE_CHANNEL: backend_channel_join (jmfile);		break;
    case JMFILE_DIR:	 backend_directory_download (jmfile);	break;
    case JMFILE_URL:	 backend_url_activate (jmfile);		break;
    default:		 g_assert_not_reached();
    }
}

