/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>

#include "jmfile.h"
#include "backend_announce.h"
#include "jmpio.h"


static void 	set_channel_timer (JMFile* channel);
static gboolean backend_announce_cb(gpointer data);
static void	backend_announce (JMAnn* jmann, gchar* action);


void
backend_announce_update_most (JMFile* channel)
{
  GSList* i;
  guint n;

  g_return_if_fail (channel);

  /* Send up-to 8 updates.  This is called when we reconnect to a down
     channel.  We only send 8 because we don't want to flood the
     channel. */

  for (n = 0, i = channel->announcements; 
       n < 8 && i != NULL; ++n, i = i->next)
    {
      JMAnn* jmann = (JMAnn*) i->data;

      backend_announce (jmann, "update");
    }
}


void
backend_announce_schedule_updates (JMFile* channel)
{
  g_return_if_fail (channel);

  /* Set update timer if we don't have one and we have some
     announcements. */
  if (!channel->update_timer && channel->announcements)
    set_channel_timer (channel);
}


void
backend_announce_add (JMFile* file, JMFile* parent)
{
  JMFile* channel;
  gchar* fullname;
  GSList* i;
  JMAnn* jmann;

  g_return_if_fail (file);
  g_return_if_fail (file->is_local);

  /* Get the channel */
  if (parent != NULL)
    channel = jmfile_get_channel (file, parent);
  else if (file->type == JMFILE_CHANNEL)
    channel = file;
  else
    return;
    
  g_return_if_fail (channel);

  /* Get the fullname of the file */
  fullname = jmfile_make_fullname (parent, file->name);
  g_return_if_fail (fullname);

  /* Check if we're already announcing this file using this name.  If
     we are, return.  (Note we might announce the same file with
     different names.)  */
  for (i = channel->announcements; i != NULL; i = i->next)
    {
      jmann = (JMAnn*) i->data;

      if (!strcmp (fullname, jmann->fullname))
	{
	  g_free (fullname);
	  return;
	}
    }

  /* Create a new announcement */
  jmann = g_new0 (JMAnn, 1);
  jmann->fullname = fullname;
  jmann->file = file;
  jmann->channel = channel;
  jmann->parent = parent;

  channel->announcements = g_slist_prepend (channel->announcements, jmann);
  if (!channel->next_announcement)
    channel->next_announcement = channel->announcements;
  file->announcers = g_slist_prepend (file->announcers, channel);

  /* Update timer */
  set_channel_timer (channel);

  /* Send an announcement */
  backend_announce (jmann, "update");
}


void
backend_announce_remove (JMFile* file, JMFile* parent) 
{
  JMFile* channel;
  gchar* fullname;
  GSList* i;
  JMAnn* jmann = NULL;

  g_return_if_fail (file);
  g_return_if_fail (file->is_local);

  /* Get the channel */
  if (parent != NULL)
    channel = jmfile_get_channel (file, parent);
  else if (file->type == JMFILE_CHANNEL)
    channel = file;
  else
    return;
    
  g_return_if_fail (channel);

  /* Get the fullname of the file */
  fullname = jmfile_make_fullname (parent, file->name);
  g_return_if_fail (fullname);

  /* Find the announcement */
  for (i = channel->announcements; i != NULL; i = i->next)
    {
      JMAnn* t;

      t = (JMAnn*) i->data;

      if (!strcmp (fullname, t->fullname))
	{
	  jmann = t;
	  break;
	}
    }

  /* Return if we weren't announcing it */
  if (!jmann)
    return;

  /* Send a remove message */
  backend_announce (jmann, "remove");

  /* Remove from the channel */
  if (channel->next_announcement && channel->next_announcement->data == jmann)
    channel->next_announcement = channel->next_announcement->next;
  channel->announcements = g_slist_remove (channel->announcements, jmann);
  file->announcers = g_slist_remove (file->announcers, channel);
  g_free (jmann->fullname);
  g_free (jmann);

  /* Adjust the timer */
  set_channel_timer (channel);
}


void
backend_announce_delete (JMFile* file)
{
  GSList* i;

  /* Remove announcements for me from other channels */
  for (i = file->announcers; i != NULL; i = i->next)
    {
      JMFile* announcer = (JMFile*) i->data;
      GSList* j;

      for (j = announcer->announcements; j != NULL; )
	{
	  JMAnn* ann = (JMAnn*) j->data;
	  j = j->next;

	  if (ann->file == file)
	    {
	      announcer->announcements = 
		g_slist_remove (announcer->announcements, ann);

	      g_free (ann->fullname);
	      g_free (ann);
	    }
	}

      set_channel_timer (announcer);
    }

  /* Remove announcements for other files from me */
  for (i = file->announcements; i != NULL; i = i->next)
    {
      JMAnn* ann = (JMAnn*) i->data;

      ann->file->announcers = 
	g_slist_remove (ann->file->announcers, file);
    }
}



static gboolean
backend_announce_cb (gpointer data)
{
  JMFile* channel = (JMFile*) data;
  JMAnn* jmann;

  g_return_val_if_fail (channel != NULL, FALSE);
  g_return_val_if_fail (channel->type == JMFILE_CHANNEL, FALSE);
  g_return_val_if_fail (channel->announcements, FALSE);

  /* If BTSMP is gone, then stop */
  if (!channel->btsmp)
    {
      channel->update_timer = 0;
      return FALSE;
    }

  /* Get next announcement */
  if (!channel->next_announcement)
    channel->next_announcement = channel->announcements;

  jmann = (JMAnn*) channel->next_announcement->data;
  channel->next_announcement = channel->next_announcement->next;

  backend_announce (jmann, "update");

  return TRUE;
}


static void
set_channel_timer (JMFile* channel)
{
  g_return_if_fail (channel);

  if (channel->update_timer)
    {
      g_source_remove (channel->update_timer);
      channel->update_timer = 0;
    }

  if (channel->announcements && channel->btsmp)
    {
      guint num_files;
      guint update_time;

      num_files = g_slist_length (channel->announcements);
      update_time = MAX (JMP_UPDATE_TIME_MAX / num_files, JMP_UPDATE_TIME_MIN);

      channel->update_timer = 
	g_timeout_add (update_time, backend_announce_cb, channel);
    }
}


static void
backend_announce (JMAnn* jmann, gchar* action)
{
  gchar* buffer;
  guint length;

  g_return_if_fail (jmann);
  g_return_if_fail (jmann->file);
  g_return_if_fail (jmann->file->is_local);
  g_return_if_fail (jmann->channel);
  g_return_if_fail (action);

  /* Don't send if no BTSMP */
  if (!jmann->channel->btsmp)
    return;

  /* Serialize */
  jmpio_write(jmann, action, &buffer, &length);
  g_return_if_fail (buffer != NULL);

  /* Send packet */
  btsmp_send_packet (jmann->channel->btsmp, buffer, length);
  g_free(buffer);
}
