/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>

#include "backend_channel.h"
#include "backend_announce.h"
#include "jmpio.h"
#include "gui.h"
#include "util.h"
#include "globals.h"


static gboolean	channel_packet_cb(BTSmp* btsmp, guint8* buffer, guint length, gpointer user_data);
static void     channel_error_cb(BTSmp* btsmp, gpointer user_data);



JMFile* 
backend_channel_add (gchar* url, JMFile* parent)
{
  g_return_val_if_fail (url, NULL);

  if (strncmp (url, "jm", 2))
    return NULL;
  /* TODO: This could be improved a bit */

  return backend_add (NULL, url, JMFILE_CHANNEL, parent);
}


JMFile* 
backend_channel_create (gchar* name, JMFile* parent, gboolean is_channels_only)
{
  JMFile* channel;
  BTSmp* btsmp;
  gchar* url;

  g_return_val_if_fail (name != NULL, NULL);
  g_return_val_if_fail (is_okfilename(name), NULL);
  g_return_val_if_fail (parent == NULL ||
			(parent->type == JMFILE_CHANNEL || parent->type == JMFILE_DIR), NULL);
  g_return_val_if_fail (parent == NULL || !jmfile_get_child_by_name(parent, name), NULL);

  
  /* Create a new BTSMP session */
  btsmp = btsmp_create(name);
  if (btsmp == NULL)
    return NULL;

  /* Store only at most 16k of announcements, timeout is same as update time */
  btsmp_cache_set_policy(btsmp, 16384, JMP_UPDATE_TIME_MAX);

  url = make_url(btsmp->url, "jm", "btsmp", 0);
  g_return_val_if_fail (url, NULL);
  
  /* Warn if we collide with another URL.  This could happen if we
     reset and recreate an old channel. */
  if ((channel = backend_get(url)) != NULL)
    {
      g_return_val_if_fail (channel->is_local, NULL);

      g_warning ("Channel with URL %s already exists, deleting.\n", url);
      jmfile_delete (channel);
    }

  /* Create a new file */
  channel = jmfile_new (name, JMFILE_CHANNEL, url);
  channel->is_local = TRUE;
  channel->btsmp = btsmp;
  channel->is_channels_only = is_channels_only;
  jmfile_link (channel, parent);

  /* Initialize BTSMP */
  btsmp->packet_func = channel_packet_cb;
  btsmp->error_func  = channel_error_cb;
  btsmp->packet_user_data = channel;
  btsmp->error_user_data  = channel;

  backend_put (channel);

  /* Send announcements on this channel and my parents */
  backend_announce_add (channel, NULL);
  if (parent) backend_announce_add (channel, parent);

  /* Add the file to the GUI */
  gui_add(channel);

  g_free (url);

  return channel;
}


void
backend_channel_join (JMFile* channel)
{
  BTSmp* btsmp;
  gchar* url;

  g_return_if_fail (channel != NULL);
  g_return_if_fail (channel->url != NULL);
  g_return_if_fail (channel->type == JMFILE_CHANNEL);

  /* Ignore if already on channel */
  if (channel->btsmp != NULL)
    return;

  url = make_url(channel->url, "btsmp", "jm", 0);
  g_return_if_fail (url);

  btsmp = btsmp_join(url);
  g_free(url);
  if (btsmp != NULL)
    {
      /* Don't cache anything */
      btsmp_cache_set_policy(btsmp, 0, 0);

      channel->btsmp = btsmp;

      /* Setup call backs */
      btsmp->packet_func = channel_packet_cb;
      btsmp->error_func  = channel_error_cb;
      btsmp->packet_user_data = channel;
      btsmp->error_user_data  = channel;

      channel->status = JMFILE_STATUS_CONNECTING;
    }
  else
    {
      channel->status = JMFILE_STATUS_DOWN;
    }

  gui_update(channel);

}



void
backend_channel_leave(JMFile* channel)
{
  GList* i;

  g_return_if_fail (channel != NULL);
  g_return_if_fail (channel->type == JMFILE_CHANNEL);
  g_return_if_fail (channel->is_local == FALSE);
  
  gui_freeze();

  if (channel->btsmp)
    {
      btsmp_delete (channel->btsmp);
      channel->btsmp = NULL;
    }

  if (channel->update_timer != 0)
    {
      g_assert (g_source_remove(channel->update_timer));
      channel->update_timer = 0;
    }

  /* Unlink all children of the jmfile */
  for (i = channel->children; i != NULL; )
    {
      JMFile* child = (JMFile*) i->data;
      i = i->next;	/* child might remove itself */
      
      jmfile_unlink (child, channel);
    }

  gui_update (channel);

  gui_thaw();
}




static gboolean
channel_packet_cb (BTSmp* btsmp, guint8* buffer, guint length, 
		   gpointer user_data)
{
  JMFile* channel = (JMFile*) user_data;
  gboolean rv = FALSE;

  DEBUGP("channel_packet_cb\n");

/*    gui_freeze(); */

  if (jmpio_read(channel, buffer, length))
    {
      g_warning ("JMP parsing problem (length %d)\n", length);
      fwrite (buffer, length, 1, stderr);
      rv = TRUE;
    }

/*    gui_thaw(); */

  if (channel->status != JMFILE_STATUS_NONE)
    {
      channel->status = JMFILE_STATUS_NONE;

      backend_announce_update_most (channel);
      backend_announce_schedule_updates (channel);

      gui_update (channel);
    }

  return rv;
}


static void
channel_error_cb(BTSmp* btsmp, gpointer user_data)
{
  JMFile* jmfile = (JMFile*) user_data;

  DEBUGP ("channel_error_cb: btsmp down - %s\n", btsmp->url);

  g_return_if_fail (btsmp == jmfile->btsmp);

  btsmp_delete (btsmp);
  jmfile->btsmp = NULL;

  if (jmfile->update_timer)
    {
      g_assert (g_source_remove(jmfile->update_timer));
      jmfile->update_timer = 0;
    }

  jmfile->status = JMFILE_STATUS_DOWN;
  gui_update(jmfile);
}
