/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <errno.h>
#include <dirent.h>

#include "backend_directory.h"
#include "backend_announce.h"

#include "gui.h"
#include "util.h"


gboolean
backend_files_create (gchar* paths, JMFile* parent, gboolean swallow_top_directory)
{
  gboolean rv = FALSE;
  gchar* token;
  gchar* next = NULL;

  if (paths == NULL || paths[0] == '\0')
    return FALSE;

  g_return_val_if_fail (parent != NULL, TRUE);

  gui_freeze();

  /* For each line in paths */
  while ((token = strtokq(paths, "\n\r", &next)) != NULL)
    {
      JMFile* new_parent;
      DIR* dir;
      struct dirent* de;

      paths = NULL;	/* Don't need this - strtok has it */

      /* Skip the "file:" part, if there is one */
      if (strstr(token, "file:") == token)
	token = &token[sizeof("file:") - 1];

      /* Skip this resource if it doesn't have an absolute path -	
	 this happens when you get something like "http://baasdf/ksjd". */
      if (token[0] != '/')
	continue;

      /* Check if it's a directory */
      dir = opendir(token);
      if (dir != NULL)
	{
	  /* Create a directory (if we aren't supposed to swallow it) */
	  if (!swallow_top_directory)
	    {
	      gchar* basename;
	      guint len;

	      /* If the token ends with a /, remove it */
	      len = strlen(token);
	      --len;
	      if (token[len] == G_DIR_SEPARATOR)
		token[len] = 0;

	      /* Get the name of the directory */
	      basename = g_basename(token);
	  
	      /* Make sure it has a good name */
	      if (!is_okfilename(basename))
		{
		  g_warning("Directory name has weird letters in it (%s). "
			    "I'm not going to add it.\n", 
			    basename);
		  closedir (dir);
		  continue;
		}

	      /* Add this directory */
	      new_parent = backend_directory_create (basename, parent);
	      if (new_parent == NULL)
		{
		  rv = TRUE;
		  closedir (dir);
		  continue;
		}
	    }
	  else
	    new_parent = parent;

	  while ((de = readdir(dir)) != NULL)
	    {
	      gchar* filename;
	      gchar* path;

	      filename = de->d_name;

	      /* Skip . and .. */
	      if ((strcmp(filename, ".")) == 0 || (strcmp(filename, "..") == 0))
		continue;

	      path = g_strconcat(token, "/", filename, NULL);

	      if (backend_files_create (path, new_parent, FALSE))
		rv = TRUE;

	      g_free(path);
	    }

	  closedir(dir);
	}

      /* Check if it's a file */
      else if (errno == ENOTDIR)
	{
	  gchar* basename = g_basename(token);

	  /* Skip if no basename */
	  if (basename == NULL)
	    rv = TRUE;

	  /* Skip if we already have it */
	  else if (jmfile_get_child_by_name(parent, basename))
	    ;

	  /* Skip if a bad name */
	  else if (!is_okfilename(basename))
	    {
	      g_warning ("File name has weird letters in it (%s).  I'm not going to add it.\n", 
			 basename);
	      rv = TRUE;
	    }

	  else if (backend_file_create (basename, parent, token) == NULL)
	    {
	      g_warning ("Could not add file \"%s\"\n", token);
	      rv = TRUE;
	    }
	}

      else
	{
	  g_warning ("Couldn't open \"%s\": %s", token, g_strerror(errno));
	  continue;
	}
    }

  gui_thaw();

  return rv;
}




JMFile* 
backend_directory_create (gchar* name, JMFile* parent)
{
  JMFile* dir;

  g_return_val_if_fail (name != NULL, NULL);
  g_return_val_if_fail (is_okfilename(name), NULL);
  g_return_val_if_fail (parent != NULL, NULL);
  g_return_val_if_fail (parent->type == JMFILE_CHANNEL || 
			parent->type == JMFILE_DIR, NULL);
  g_return_val_if_fail (!jmfile_get_child_by_name (parent, name), NULL);


  dir = jmfile_new (name, JMFILE_DIR, NULL);
  dir->is_local = TRUE;
  jmfile_link (dir, parent);

  backend_announce_add (dir, parent);

  gui_add(dir);

  return dir;
}



void 
backend_directory_download (JMFile* dir)
{
  GList* i;

  g_return_if_fail (dir != NULL);
  g_return_if_fail (dir->type == JMFILE_DIR);

  /* Download everything in the directory.  We may actually want to
     ask for a confirmation. */
  for (i = dir->children; i != NULL; i = i->next)
    {
      JMFile* child = (JMFile*) i->data;

      if (child->type == JMFILE_FILE ||
	  child->type == JMFILE_DIR)
	{
	  backend_activate(child);
	}
    }
}

