/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <errno.h>
#include "jmgnet.h"
#include "sha.h"

#include "gui.h"
#include "util.h"
#include "globals.h"

#include "backend_file.h"
#include "backend_announce.h"
#include "backend.h"


#define	MUNMAP_TIME	(3 * 60 * 1000)	/* Unmap the file after 3 minutes */



static void backend_file_redownload (JMFile* jmfile);
static void backend_file_setup_path (JMFile* file);

static void file_func (MtpClient* client, MtpClientStatus status, 
		       guint8* buffer, guint length, 
		       gpointer user_data);

static gint read_func (MtpLocalMirror* mirror, gint offset, 
		       gchar** bufferp, gint blength,
		       gpointer user_data);

static gboolean munmap_timeout (gpointer p);

static void sha_cb (GSHA* sha, gpointer user_data);



JMFile* 
backend_file_create (gchar* name, JMFile* parent, gchar* path)
{
  gint length;

  gchar* fullname;
  JMGURL* url = NULL;
  gchar* url_str = NULL;

  JMFile* jmfile;
  MtpLocalMirror* mirror;

  g_return_val_if_fail (name, NULL);
  g_return_val_if_fail (is_okfilename(name), NULL);
  g_return_val_if_fail (parent == NULL ||
			(parent->type == JMFILE_CHANNEL || 
			 parent->type == JMFILE_DIR), NULL);
  g_return_val_if_fail (parent == NULL || 
			!jmfile_get_child_by_name(parent, name), NULL);
  g_return_val_if_fail (path, NULL);
  g_return_val_if_fail (file_exists (path), NULL);
  g_return_val_if_fail (backend_mtp_server, NULL);
  g_return_val_if_fail (backend_rendezvous_url, NULL);


  /* Get the file length (and check file exists) */
  length = filename_size (path);
  if (length < 0)
    {
      g_warning ("Could open file %s\n", path);
      return NULL;
    }

  /* Make the fullname */
  fullname = jmfile_make_fullname (parent, name);
  g_return_val_if_fail (fullname, NULL);

  /* Make the URL */
  url = jm_url_clone (backend_rendezvous_url);
  jm_url_set_resource (url, fullname);
  g_free (fullname);
  url_str = jm_url_get_nice_string (url);

  /* Warn if we collide with another URL.  This could happen if we
     reset and recreate an old file. */
  if ((jmfile = backend_get(url_str)) != NULL)
    {
      g_return_val_if_fail (jmfile->is_local, NULL);

      g_warning ("File with URL %s already exists, deleting.\n", url_str);
      jmfile_delete (jmfile);
    }

  /* Mirror the file using the normal file name */
  mirror = mtp_mirror (backend_mtp_server, url, read_func, NULL /* fix below */, length);
  if (!mirror)
    {
      g_warning ("Could not mirror file %s\n", name);
      goto error;
    }

  /* Create a new file */
  jmfile = jmfile_new (name, JMFILE_FILE, url_str);
  jmfile->is_local = TRUE;
  jmfile->mtp_mirror = mirror;
  mirror->user_data = jmfile;

  jmfile->path = g_strdup(path);
  jmfile->mtp_mirror = mirror;
  jmfile->buffer_length = length;
  jmfile->length = length;
  jmfile->have_file = TRUE;

  backend_put (jmfile);
  jmfile_link (jmfile, parent);

  /* Calculate the SHA - then we will send an announcment. */
  jmfile->sha_id = gnet_sha_new_file_async (path, sha_cb, jmfile);

  gui_add (jmfile);

  return jmfile;

 error:
  g_free (url_str);
  jm_url_delete (url);

  return NULL;
}


static void
sha_cb (GSHA* sha, gpointer user_data)
{
  JMFile* jmfile = (JMFile*) user_data;

  jmfile->sha_id = NULL;

  if (sha)
    {
      GList* i;

      /* Save the SHA */
      g_return_if_fail (jmfile->sha == NULL);
      jmfile->sha = sha;

      /* Send announcements if any parents */
      for (i = jmfile->parents; i != NULL; i = i->next)
	{
	  JMFile* parent = (JMFile*) i->data;

	  backend_announce_add (jmfile, parent);
	}

      /* If the mirror is not local, then mirror by SHA too */
      if (!backend_rendezvous_locally)
	{
	  gchar sha_buf[4 + GNET_SHA_HASH_LENGTH * 2 + 1] = "/SHA";
	  JMGURL* url;

	  g_return_if_fail (jmfile->sha_mirror == NULL);

	  gnet_sha_copy_string (sha, &sha_buf[4]);
	  sha_buf[4 + GNET_SHA_HASH_LENGTH * 2 + 1] = '\0';

	  url = jm_url_clone (backend_rendezvous_url);
	  jm_url_set_resource (url, sha_buf);

	  jmfile->sha_mirror = 
	    mtp_mirror (backend_mtp_server, url, read_func, 
			jmfile, jmfile->length);
	}

      gui_update (jmfile);
    }
  else
    {
      /* Something is wrong */
      jmfile->status = JMFILE_STATUS_DOWN;
      gui_update (jmfile);
    }
}



void
backend_file_download (JMFile* jmfile)
{
  JMGURL* url;
  MtpClient* mtp_client;

  g_return_if_fail (jmfile != NULL);
  g_return_if_fail (jmfile->url != NULL);
  g_return_if_fail (jmfile->type == JMFILE_FILE);

  /* Ignore if we are transfering the file already */
  if (jmfile->status == JMFILE_STATUS_TRANSFERRING)
    return;

  /* GUI activate if we have the file */
  if (jmfile->is_local || 
      (jmfile->path && file_exists (jmfile->path)))
    {
      gui_activate (jmfile);
      return;
    }

  /* Redownload if we have the file, but it's not there (the user may
     have moved it).  This still probably doesn't work right... */
  if (jmfile->file)
    {
      backend_file_redownload (jmfile);
      return;
    }
	
  /* Create URL */
  url = jm_url_new (jmfile->url);
  g_return_if_fail (url);

  /* Set up the path */
  backend_file_setup_path (jmfile);
  g_return_if_fail (jmfile->path);
  g_return_if_fail (jmfile->dirpath);

  /* Get the file */
  mtp_client = mtp_client_get_file_rendezvous (url, file_func, jmfile);
  g_return_if_fail (mtp_client);

  jmfile->mtp_client = mtp_client;
  jmfile->status = JMFILE_STATUS_TRANSFERRING;
  gui_update(jmfile);
}




static void
backend_file_redownload (JMFile* jmfile)
{
  FILE* file;

  g_return_if_fail (jmfile->path);
  g_return_if_fail (jmfile->dirpath);
  g_return_if_fail (jmfile->status != JMFILE_STATUS_TRANSFERRING);

  /* TODO: If there was an error with the transfer, this
     code might be executed and we'd create a zero
     length file.  This would be annoying. */

  /* Create the directory */
  if (!mkdirr(jmfile->dirpath, 0700))
    {
      g_warning ("Could not create directory %s\n", jmfile->dirpath);
      return;
    }

  /* Open the file for writing */
  if ((file = fopen (jmfile->path, "w")) == NULL)
    {
      g_warning ("Could not write file %s\n", jmfile->path);
      goto done;
    }

  /* Write the file */
  if (fwrite(jmfile->buffer, jmfile->buffer_length, 1, file) != 1)
    g_warning ("Could not write file %s\n", jmfile->path);

 done:
  /* Close the file */
  fclose(file);
}



static void
backend_file_setup_path (JMFile* file)
{
  gchar* path;
  gchar* temp;

  g_return_if_fail (file);
  g_return_if_fail (file->type == JMFILE_FILE);
  g_return_if_fail (jm_home);

  if (file->path != NULL)
    return;

  /* Use the first parent */
  if (file->parents)
    {
      JMFile* parent;

      /* Get the path relative to jm_home */
      parent = (JMFile*) file->parents->data;
      path = jmfile_make_fullname (parent, "");
      g_return_if_fail (is_okpathname(path));

      /* Add the path to jm_home */
      temp = path;
      path = g_strconcat(jm_home, path, NULL);
      g_free (temp);
    }
  else	/* Otherwise, just use jm_home */
    {
      path = g_strdup(jm_home);
    }

  if (!mkdirr(path, 0700))
    {
      g_warning ("Could not create directory %s\n", path);
      g_free(path);
      return;
    }

  file->dirpath = path;

  /* Create the final path name.  Add directory separator if
     needed. */
  if (*path && path[strlen(path) - 1] == G_DIR_SEPARATOR)
    path = g_strconcat(path, file->name, NULL);
  else
    path = g_strconcat(path, G_DIR_SEPARATOR_S, file->name, NULL);

  file->path = path;
}



void
backend_file_cancel (JMFile* file)
{
  g_return_if_fail (file != NULL);
  g_return_if_fail (file->type == JMFILE_FILE);
  g_return_if_fail (file->status == JMFILE_STATUS_TRANSFERRING);
  g_return_if_fail (!file->is_local);

  /* Delete the mtp client */
  if (file->mtp_client)
    {
      mtp_client_delete (file->mtp_client);
      file->mtp_client = NULL;
    }

  file->status = JMFILE_STATUS_NONE;
  
  gui_update(file);
}



/*

  Called when part of the file is received via MTP.

*/
static void
file_func (MtpClient* client, MtpClientStatus status, 
	   guint8* buffer, guint length, 
	   gpointer user_data)
{
  JMFile* jmfile = (JMFile*) user_data;

  /* If there was an error, reset myself */
  if (status != MTP_CLIENT_STATUS_OK)
    {
      mtp_client_delete (client);
      jmfile->mtp_client = NULL;

      if (jmfile->buffer)
	{
	  if (munmap(jmfile->buffer, jmfile->length) != 0)
	    g_warning ("munmap failure: %s\n", g_strerror(errno));
	  jmfile->buffer = NULL;
	}

      if (jmfile->file)
	{
	  fclose (jmfile->file);
	  jmfile->file = NULL;
	}

      jmfile->status = JMFILE_STATUS_DOWN;
      gui_update (jmfile);
    }

  /* If the is the first time, then there will be no buffer */
  else if (!buffer)
    {
      FILE* file;

      g_return_if_fail (jmfile->path);

      /* Open the file for reading and writing */
      if ((file = fopen (jmfile->path, "w+")) == NULL)
	{
	  g_warning ("Could not create file %s\n", jmfile->path);
	  goto error;
	}

      /* If the length is 0, just write a 0 byte and close it - we're
         done */
      if (client->length == 0)
	{
	  if (fwrite(NULL, 0, 1, file) != 1)
	    {
	      g_warning ("Could not write file %s\n", jmfile->path);
	      goto error;
	    }

	  /* Close the file */
	  fclose(file);

	  mtp_client_delete (client);
	  jmfile->mtp_client = NULL;

	  jmfile->status = JMFILE_STATUS_NONE;
	  jmfile->have_file = TRUE;
	}
	  
      /* Otherwise, memmap the file */
      else
	{
	  gchar* buffer;

	  /* Seek to the end of the file */
	  fseek (file, client->length - 1, SEEK_SET);

	  /* Write a byte, so the file is the right length */
	  fwrite ("", 1, 1, file);
	  fflush (file);

	  /* Map the file to memory */
	  buffer = mmap (NULL, client->length, PROT_READ | PROT_WRITE, 
			 MAP_SHARED, fileno(file), 0);
	  if ((buffer == NULL) || ((int) buffer == -1))
	    {
	      g_warning ("Could not mmap file %s\n", jmfile->path);
	      goto error;
	    }

	  jmfile->file = file;
	  jmfile->buffer = buffer;
	  jmfile->buffer_length = client->length;
	  jmfile->buffer_offset = client->offset;
	}

      gui_update (jmfile);
      return;

    error:
      if (file)
	fclose (file);
      jmfile->file = NULL;
	
      jmfile->status = JMFILE_STATUS_DOWN;

      gui_update (jmfile);
    }
					     
  /* Otherwise, this is data - copy it */
  else
    {
      g_return_if_fail (jmfile->buffer);
      g_return_if_fail (length + jmfile->buffer_offset <= jmfile->buffer_length);

      /* Copy the buffer to our buffer */
      memcpy (&jmfile->buffer[jmfile->buffer_offset], buffer, length);
      jmfile->buffer_offset += length;

      if (jmfile->buffer_offset == jmfile->buffer_length)
	{
	  gboolean bad_sha = FALSE;

	  /* Mark us as having the file */
	  jmfile->have_file = TRUE;
	  jmfile->status = JMFILE_STATUS_NONE;
	  jmfile->buffer_offset = 0;

	  /* Delete the client */
	  mtp_client_delete (client);
	  jmfile->mtp_client = NULL;
	  
	  /* Sync the file */
	  if (msync (jmfile->buffer, jmfile->buffer_length, MS_SYNC))
	    g_warning ("msync failed.\n");

	  /* Double check the SHA */
	  if (jmfile->sha)
	    {
	      GSHA* sha;

	      /* TODO: Calculate asynchrously */
	      sha = gnet_sha_new (jmfile->buffer, jmfile->buffer_length);
	      g_return_if_fail (sha);

	      if (!gnet_sha_equal (sha, jmfile->sha))
		{
		  bad_sha = TRUE;

		  g_warning ("SHA hashes not equal for file %s (is %s, should be %s, length is %d, buffer length is %d)\n", 
			     jmfile->url,
			     gnet_sha_get_string(sha),
			     gnet_sha_get_string(jmfile->sha),
			     jmfile->length, jmfile->buffer_length);
		}

	      gnet_sha_delete (sha);
	    }

	  /* Unmap the file */
	  munmap (jmfile->buffer, jmfile->buffer_length);
	  jmfile->buffer = NULL;

	  /* Close the file */
	  fclose (jmfile->file);
	  jmfile->file = NULL;

	  /* Mirror the file maybe */
	  if (!bad_sha && backend_auto_mirror_on_download && backend_mtp_server)
	    {
	      JMGURL* url = NULL;

	      url = jm_url_new (jmfile->url);
	      g_return_if_fail (url);

	      jmfile->mtp_mirror = 
		mtp_mirror (backend_mtp_server, url, 
			    read_func, jmfile, jmfile->buffer_length);

	      jm_url_delete (url);
	    }
	}

      /* Update the GUI */
      gui_update (jmfile);
    }
}


/* 

   Called by mtp_mirror to read the next section of the file.

   We try will memmap the file if it is not memmaped (and compute the
   SHA).

 */


static gint
read_func (MtpLocalMirror* mirror, gint offset, gchar** bufferp, gint blength,
	   gpointer user_data)
{
  JMFile* jmfile = (JMFile*) user_data;

  g_return_val_if_fail (jmfile, 0);
  g_return_val_if_fail (mirror, 0);
  g_return_val_if_fail (bufferp, 0);

  /* Load file if not loaded */
  if (!jmfile->file)
    {
      FILE* file;
      gchar* buffer;

      g_return_val_if_fail (jmfile->path, 0);

      /* Open file */
      file = fopen (jmfile->path, "r");
      if (!file)
	{
	  g_warning ("Couldn't open file %s.\n", jmfile->path);
	  return 0;

	  /* TODO: User probably moved it, we should kill the
             mirror. */
	}

      /* Verify length */
      if (jmfile->buffer_length != file_size (file))
	{
	  g_warning ("File length changed.\n");
	  fclose (file);
	  return 0;
	}

      /* Memmap the file */
      buffer = mmap (NULL, jmfile->buffer_length, 
		     PROT_READ, MAP_PRIVATE, fileno(file), 0);
      if (buffer == NULL || ((int) buffer == -1))
	{
	  g_warning ("Could not memmap file.\n");
	  fclose (file);
	  return 0;
	}

      jmfile->file = file;
      jmfile->buffer = buffer;
    }

  g_return_val_if_fail (jmfile->buffer, 0);
  g_return_val_if_fail (offset + blength <= jmfile->buffer_length, 0);

  /* Reset timer */
  if (jmfile->munmap_timer)
    g_source_remove (jmfile->munmap_timer);
  jmfile->munmap_timer = g_timeout_add (MUNMAP_TIME, munmap_timeout, jmfile);

  /* Get next block */
  *bufferp = &jmfile->buffer[offset];
  return blength;
}


static gboolean
munmap_timeout (gpointer p)
{
  JMFile* jmfile = (JMFile*) p;

  g_return_val_if_fail (jmfile, FALSE);
  g_return_val_if_fail (jmfile->file, FALSE);
  g_return_val_if_fail (jmfile->buffer, FALSE);

  jmfile->munmap_timer = 0;

  munmap (jmfile->buffer, jmfile->buffer_length);
  jmfile->buffer = NULL;

  fclose (jmfile->file);
  jmfile->file = NULL;

  return FALSE;
}
