/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "globals.h"
#include "jmgnet.h"
#include "util.h"

#include "b_conn.h"
#include "b_conn_accept.h"
#include "b_conn_connect.h"

#include "btp.h"
#include "btp_node.h"
#include "btp_packet.h"
#include "btp_proto.h"




/* ************************************************************ */


static gboolean		btp_initialized = FALSE;
static gchar*		btp_url_base = NULL;
static GHashTable*	name_to_tree = NULL;	



/* ************************************************************ */



/**

   Initialize the BTP module.  This calls b_conn_init, figures out the
   base URL, and creates the hash table.

   Returns FALSE on success, TRUE if there was an error.

   Last checked: 	2000-4-28 DAH

 */
gboolean
btp_init (gchar* hostname, GInetAddr* listening_interface, gint port, gboolean force_port)
{
  g_return_val_if_fail (hostname != NULL, TRUE);
  g_return_val_if_fail (listening_interface != NULL, TRUE);
  g_return_val_if_fail (btp_initialized == FALSE, TRUE);

  DEBUGP ("btp_init\n");

  if (!port)
    port = BTP_PORT;

  if (b_conn_init(hostname, listening_interface, port, FALSE))
    return TRUE;
      
  /* Hook up BTP handler */
  b_conn_set_accept_func (B_PACKET_PROTOCOL_BTP, btp_conn_accept);

  /* Create the URL base */
  if (b_server->port == BTP_PORT)
    btp_url_base = g_strdup_printf ("btp://%s", b_hostname);
  else
    btp_url_base = g_strdup_printf ("btp://%s:%d", b_hostname, b_server->port);

  /* Create the name->tree hashtable */
  name_to_tree = g_hash_table_new (g_str_hash, g_str_equal);

  btp_initialized = TRUE;

  DEBUGP ("btp_init: hostname = %s, port = %d\n", 
	  b_hostname, b_server->port);

  return FALSE;
}


gboolean
btp_is_initialized (void)
{
  DEBUGP ("btp_is_initialized\n");

  return btp_initialized;
}


/**

   Shutdown the BTP module.

   Last checked: 	2000-3-19 DAH

 */
void 
btp_shutdown (void)
{
  DEBUGP ("btp_shutdown\n");

  b_conn_shutdown();

  g_free (btp_url_base);
  btp_url_base = NULL;

  /* TODO: Delete all trees? */

  g_hash_table_destroy (name_to_tree);
  name_to_tree = NULL;
}




/**

   Create a new BTP tree.

   Last checked: 	2000-3-19 DAH

 */
BtpTree* 
btp_create (gchar* name, BtpTreeType tree_type)
{
  BtpTree* tree = NULL;

  g_return_val_if_fail (btp_initialized, NULL);
  g_return_val_if_fail (name, NULL);

  DEBUGP ("btp_create %s\n", name);

  /* ******************** */

  /* Create the URL */
  if (name[0] == '/')
    name = g_strdup (name);
  else
    name = g_strconcat ("/", name, NULL);

  /* Make sure we don't already have this tree */
  if (btp_get(name))
    return NULL;

  /* Create tree */
  tree = g_new0 (BtpTree, 1);
  tree->name = name;
  tree->url = g_strconcat (btp_url_base, name, NULL);
  tree->type = tree_type;
  tree->ref_count = 1;

  /* Add me to tree */
  tree->me = btp_node_new (tree, b_hostname, b_server->port, NULL);

  /* Make me root */
  tree->root = tree->me;
  tree->parent = NULL;

  /* Add the tree to our tables */
  btp_add (tree);

  return tree;
}



/**

   Join an existing BtpTree.

   Last checked: 	2000-3-18 DAH

 */
BtpTree* 
btp_join (gchar* url, BtpTreeType tree_type, gboolean passive)
{
  BtpTree* tree = NULL;
  gchar* proto = NULL;
  gchar* server = NULL;
  gint port = 0;
  gchar* resource = NULL;
  gchar* new_url = NULL;

  g_return_val_if_fail (url, NULL);

  DEBUGP ("btp_join %s\n", url);

  /* Parse the URL */
  parse_url (url, &proto, &server, &port, &resource);
  if (server == NULL || resource == NULL)
    {
      g_warning ("Bad URL: %s\n", url);
      goto done;
    }

  if (port == 0)
    port = BTP_PORT;

  /* Make a new canonical URL */
  if (port == BTP_PORT)
    new_url = g_strdup_printf ("btp://%s%s", server, resource);
  else
    new_url = g_strdup_printf ("btp://%s:%d%s", server, port, resource);

  /* Make sure we don't already have this tree (why?) */
  if (btp_get (resource))
    goto done;

  /* Make sure I'm not joining myself (why?) */
  if (b_server->port == port && !strcmp (b_hostname, server))
    goto done;

  /* Create tree */
  tree = g_new0(BtpTree, 1);
  tree->name = g_strdup(resource);
  tree->url = new_url;	new_url = NULL;
  tree->type = tree_type;
  tree->passive = passive;
  tree->ref_count = 1;
  
  /* Create me */
  tree->me = btp_node_new (tree, b_hostname, b_server->port, NULL);

  /* Create root and connect */
  tree->root = btp_node_new (tree, server, port, NULL);
  if (!passive)	/* If we are passive, we have no parent */
    tree->parent = tree->root;
  b_conn_connect (tree->root->conn);

  /* Add the tree to our tables */
  btp_add (tree);

 done:
  if (new_url)	g_free (new_url);
  if (proto)	g_free (proto);
  if (server)	g_free (server);
  if (resource)	g_free (resource);

  return tree;
}



/* **************************************** */


void
btp_ref (BtpTree* tree)
{
  g_return_if_fail (tree);

  DEBUGP ("btp_ref\n");

  ++tree->ref_count;
}


/* return TRUE if the tree is still valid */
gboolean
btp_unref (BtpTree* tree)
{
  g_return_val_if_fail (tree, TRUE);

  DEBUGP ("btp_unref\n");

  --tree->ref_count;

  if (!tree->ref_count)
    {
      btp_tree_delete (tree);
      return FALSE;
    }
  
  return TRUE;
}


gboolean
btp_clear_and_unref (BtpTree* tree)
{
  g_return_val_if_fail (tree, TRUE);

  DEBUGP ("btp_clear_and_unref\n");

  tree->mcast_read_func = NULL;
  tree->ucast_read_func = NULL;
  tree->hint_func = NULL;
  tree->error_func = NULL;
  tree->packet_func = NULL;

  tree->mcast_read_user_data = NULL;
  tree->ucast_read_user_data = NULL;
  tree->hint_user_data = NULL;
  tree->error_user_data = NULL;
  tree->packet_user_data = NULL;

  return btp_unref (tree);
}



/* ************************************************************ */



/**

   Get a tree using the tree name.

   Last checked:	2000-4-28 DAH

*/
BtpTree* 
btp_get (gchar* name)
{
  g_return_val_if_fail (name_to_tree, NULL);
  g_return_val_if_fail (name, NULL);

  DEBUGP ("btp_get %s\n", name);

  return (BtpTree*) g_hash_table_lookup (name_to_tree, name);
}



/**

   Add a tree to the url-to-tree table.

   Last checked:	2000-3-19 DAH

*/
void
btp_add (BtpTree* tree)
{
  g_return_if_fail (name_to_tree != NULL);
  g_return_if_fail (tree != NULL);
  g_return_if_fail (tree->name != NULL);
  g_return_if_fail (btp_get(tree->name) == NULL);

  DEBUGP ("btp_add\n");

  g_hash_table_insert (name_to_tree, tree->name, tree);
}



/**

   Remove tree from the url-to-tree table

   Last checked:	2000-3-19 DAH

*/
void
btp_remove (BtpTree* tree)
{
  if (name_to_tree)
    {
      g_return_if_fail (tree != NULL);
      g_return_if_fail (tree->name != NULL);
      g_return_if_fail (btp_get(tree->name) != NULL);

      g_hash_table_remove (name_to_tree, tree->name);
    }
}




/* ************************************************************ */
/* Send functions						*/



void
btp_send_mcast (BtpTree* tree, BtpNode* not_node, guint8* buffer, guint16 length)
{
  GList* ci;

  g_return_if_fail (tree);
  g_return_if_fail (buffer);

  /* Send to parent (if I have a parent) */
  if (tree->parent != NULL &&
      tree->parent != not_node)
    {
      BPacket* packet = b_packet_new_mcast (tree->parent->conn, buffer, length);
      b_conn_send_packet (tree->parent->conn, packet);
    }

  /* Send to children */
  for (ci = tree->children; ci != NULL; ci = ci->next)
    {
      BtpNode* child_node = (BtpNode*) ci->data;

      if (child_node != not_node)
	{
	  BPacket* packet = b_packet_new_mcast (child_node->conn, buffer, length);
	  b_conn_send_packet (child_node->conn, packet);
	}
    }
}


void
btp_send_mcast_children (BtpTree* tree, guint8* buffer, guint16 length)
{
  g_return_if_fail (tree);
  g_return_if_fail (buffer);

  btp_send_mcast (tree, tree->parent, buffer, length);
}



void
btp_send_ucast (BtpTree* tree, BtpNode* node, guint8* buffer, guint16 length)
{
  BPacket* packet;

  g_return_if_fail (tree);
  g_return_if_fail (node);
  g_return_if_fail (buffer);

  packet = b_packet_new_ucast (node->conn, buffer, length);
  b_conn_send_packet (node->conn, packet);
}
