/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <unistd.h>
#include <limits.h>
#include <errno.h>

#include "globals.h"
#include "jmgnet.h"
#include "util.h"

#include "b_conn_connect.h"
#include "b_conn_handler.h"
#include "b_util.h"

#include "btp.h"
#include "btp_proto.h"
#include "btp_node.h"
#include "btp_packet.h"



BConnAcceptType btp_conn_accept (BConn* conn, gchar* obj_name);

static void	btp_conn_connect   (BConn* conn, gpointer user_data);
static void	btp_conn_ping      (BConn* conn, gpointer user_data);
static void	btp_conn_writeable (BConn* conn, gpointer user_data);
static void	btp_conn_close     (BConn* conn, gpointer user_data);
static void	btp_conn_fail      (BConn* conn, gpointer user_data);


BConnFuncs btp_proto_funcs = 
{
  btp_conn_connect,
  btp_conn_ping,
  btp_conn_writeable,
  btp_conn_close,
  btp_conn_fail		/* TODO: Write a close function ? */
};


static void	add_handlers (BtpNode* node);
static void	remove_handlers (BtpNode* node);

static void	receive_any (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_ok_switch (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_error_switch (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_ucast (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_mcast (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_request_children (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_reply_children (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_join (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_switch (BConn* conn, BPacket* packet, gpointer user_data);
static void	receive_leave (BConn* conn, BPacket* packet, gpointer user_data);

static void	schedule_request_children (BtpTree* tree, gboolean soon);
static void	cancel_request_children (BtpTree* tree);
static gboolean	request_children_cb (gpointer data);

static void 	schedule_find_new_parent (BtpTree* tree);
/*  static void 	cancel_find_new_parent (BtpTree* tree); */
static gboolean find_new_parent_cb (gpointer data);
static BtpNode* find_new_parent (BtpTree* tree);

static gboolean	switch_timeout_cb (gpointer data);
static void	schedule_switch_timeout (BtpTree* tree);
static void	cancel_switch_timeout (BtpTree* tree);

#define	TREE	(node->tree)


/* ************************************************************ */



BConnAcceptType
btp_conn_accept (BConn* bconn, gchar* obj_name)
{
  BtpTree* tree;
  BtpNode* node;

  g_return_val_if_fail (bconn,  FALSE);
  g_return_val_if_fail (bconn->conn,  FALSE);
  g_return_val_if_fail (obj_name,  FALSE);

  tree = btp_get (obj_name);
  if (tree == NULL)
    return B_CONN_ACCEPT_NO_OBJECT;

  node = 
    btp_tree_find_node (tree, bconn->conn->hostname, bconn->conn->port);
  if (node && node->conn)
    return B_CONN_ACCEPT_ERROR;
  else if (node)
    node->conn = bconn;
  else
    node = btp_node_new (tree, bconn->conn->hostname, bconn->conn->port, bconn);

  add_handlers (node);

  ++tree->num_connect;

  return B_CONN_ACCEPT_OK;
}




static void
btp_conn_connect (BConn* conn, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;;

  g_return_if_fail (conn);
  g_return_if_fail (node);
  g_return_if_fail (TREE);

  DEBUGP ("btp_conn_connect\n");

  add_handlers (node);

  /* If this is my parent, send it a JOIN and request the siblings
     request. */
  if (btp_tree_is_parent(node))
    {
      BPacket* join_packet;

      /* Send a JOIN */
      join_packet = btp_packet_new_join (node);
      b_conn_send_packet (node->conn, join_packet);

      /* Schedule a children request. */
      schedule_request_children (TREE, TRUE);

      /* Hint that we have a new parent. */
      if (TREE->hint_func)
	{
	  (TREE->hint_func)(TREE, node, BTP_HINT_NEW_PARENT, TREE->hint_user_data);
	}
    }
}


static void
btp_conn_ping (BConn* bconn, gpointer user_data)
{
  g_return_if_fail (bconn);
  g_return_if_fail (bconn->conn);

  DEBUGP ("BTP_CONN_PING: distance to %s:%d is %d\n", 
	  bconn->conn->hostname, bconn->conn->port, bconn->distance);

  /* do nothing for now */
}


static void
btp_conn_writeable (BConn* conn, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;;

  DEBUGP ("btp_conn_writeable\n");

  g_return_if_fail (conn);
  g_return_if_fail (node);
  g_return_if_fail (TREE);

  if (TREE->hint_func)
    {
      (TREE->hint_func)(TREE, node, BTP_HINT_UCAST_WRITE, TREE->hint_user_data);
    }
}


static void
btp_conn_close (BConn* conn, gpointer user_data)
{
  /* Closing and failing are basically the same things. */
  btp_conn_fail (conn, user_data);
}


static void
btp_conn_fail (BConn* conn, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;

  g_return_if_fail (conn);
  g_return_if_fail (node);
  g_return_if_fail (!TREE || (TREE->me != node));

  DEBUGP ("btp_conn_fail\n");

  remove_handlers (node);

  /* Delete the node if we're deleting it. */
  if (node->is_deleting)
    {
      DEBUGP ("btp_conn_fail: delete\n");

      b_conn_unref (conn);
      memset (node, 0, sizeof (*node));
      g_free (node);
      return;
    }

  /* Notify the upper layer.  Note that we don't notify again if the
     node comes back up. */
  if (TREE->hint_func)
    {
      DEBUGP ("btp_conn_fail: hint\n");

      (TREE->hint_func)(TREE, node, BTP_HINT_NODE_DOWN, TREE->hint_user_data);
    }

  /* If it's our child, delete it */
  if (btp_tree_is_child(node))
    {
      DEBUGP ("btp_conn_fail: is_child\n");

      btp_tree_remove_child (node);
      btp_tree_remove_node (node);
      btp_node_delete (node);
      return;
    }

  /* If it's our parent and the root, do an error upcall */
  else if (btp_tree_is_parent(node) && btp_tree_is_root(node))
    {
      /* TODO: Try reconnecting a few times */

      DEBUGP ("btp_conn_fail: is_parent and root\n");

      /* Do an error upcall */
      if (TREE->error_func)
	{
	  (TREE->error_func)(TREE, TREE->error_user_data);
	  return;
	}
    }

  /* If it's our parent and _not_ the root, disconnect from my parent
     and connect to the root */
  else if (btp_tree_is_parent(node) && !btp_tree_is_root(node))
    {
      /* Set our parent to root */
      TREE->parent = TREE->root;

      DEBUGP ("btp_conn_fail: is_parent, not root\n");

      /* Connect to the root if unconnected. */
      if (b_conn_is_unconnected(TREE->parent->conn))
	b_conn_connect(TREE->parent->conn);

      /* Otherwise, send a JOIN */
      else
	{
	  BPacket* join_packet;
	  
	  /* Send a join */
	  join_packet = btp_packet_new_join (node);
	  b_conn_send_packet (node->conn, join_packet);

	  /* Schedule a children request. */
	  schedule_request_children (TREE, TRUE);

	  /* Hint that we have a new parent. */
	  if (TREE->hint_func)
	    {
	      (TREE->hint_func)(TREE, node, BTP_HINT_NEW_PARENT, TREE->hint_user_data);
	    }
	}
    }

  /* If it's the root and not our parent, let it stay down */
  else if (btp_tree_is_root(node) && !btp_tree_is_parent(node))
    {
      /* Do nothing */;
      DEBUGP ("btp_conn_fail: is root, not parent\n");

    }


  /* If it's our sibling, delete it.  Reset the new_parent stuff if
     it's also our new_parent. */
  else if (btp_tree_is_sibling(node))
    {
      btp_tree_remove_sibling (node);

      DEBUGP ("btp_conn_fail: is sibling\n");

      /* If the node is new_parent, then remove the SWITCH from its
	 queue and schedule a new parent search. */
      if (btp_tree_is_new_parent(node))
	{
	  DEBUGP ("btp_conn_fail: is new parent\n");

	  TREE->new_parent = NULL;

	  /* Cancel the SWITCH timeout */
	  cancel_switch_timeout (TREE);

	  /* Remove the SWITCH packet from the node's out queue */
	  b_conn_remove_packets (node->conn, BTP_PACKET_TYPE_SWITCH);

	  /* Request children again from parent */
	  schedule_request_children (TREE, TRUE);

	  /* (How do we ensure we don't connect to the node again -
             maybe we should keep it around?) */
	}
      
      btp_tree_remove_node (node);
      btp_node_delete (node);
    }

  /* Otherwise, it's completely unrelated, so delete it */
  else
    {
      DEBUGP ("btp_conn_fail: is no one\n");

      btp_tree_remove_node (node);
      btp_node_delete (node);
    }
}


static void
add_handlers (BtpNode* node)
{
  BConn* conn;

  g_return_if_fail (node);
  g_return_if_fail (node->conn);
  g_return_if_fail (TREE);

  conn = node->conn;

  b_conn_handler_add (conn, B_PACKET_TYPE_OK, BTP_PACKET_OK_SUBTYPE_SWITCH,
		      receive_ok_switch, node);
  b_conn_handler_add (conn, B_PACKET_TYPE_ERROR, BTP_PACKET_ERROR_SUBTYPE_SWITCH,
		      receive_error_switch, node);

  b_conn_handler_add (conn, B_PACKET_TYPE_UCAST, 0, receive_ucast, node);
  b_conn_handler_add (conn, B_PACKET_TYPE_MCAST, 0, receive_mcast, node);

  b_conn_handler_add (conn, B_PACKET_TYPE_INFO_REQUEST, BTP_PACKET_INFO_SUBTYPE_CHILDREN,
		      receive_request_children, node);
  b_conn_handler_add (conn, B_PACKET_TYPE_INFO_REPLY, BTP_PACKET_INFO_SUBTYPE_CHILDREN,
		      receive_reply_children, node);

  b_conn_handler_add (conn, BTP_PACKET_TYPE_JOIN, 0, receive_join, node);
  b_conn_handler_add (conn, BTP_PACKET_TYPE_SWITCH, 0, receive_switch, node);
  b_conn_handler_add (conn, BTP_PACKET_TYPE_LEAVE, 0, receive_leave, node);

  /* Start handling any packet if there's a packet func */
  if (TREE->packet_func)
    b_conn_handler_add (conn, -1, -1, receive_any, node);
}


static void
remove_handlers (BtpNode* node)
{
  BConn* conn;

  g_return_if_fail (node);
  g_return_if_fail (node->conn);
  g_return_if_fail (TREE);

  conn = node->conn;

  b_conn_handler_remove (conn, B_PACKET_TYPE_OK, BTP_PACKET_OK_SUBTYPE_SWITCH, node);
  b_conn_handler_remove (conn, B_PACKET_TYPE_ERROR, BTP_PACKET_ERROR_SUBTYPE_SWITCH, node);

  b_conn_handler_remove (conn, B_PACKET_TYPE_INFO_REQUEST, BTP_PACKET_INFO_SUBTYPE_CHILDREN, node);
  b_conn_handler_remove (conn, B_PACKET_TYPE_INFO_REPLY, BTP_PACKET_INFO_SUBTYPE_CHILDREN, node);

  b_conn_handler_remove (conn, BTP_PACKET_TYPE_JOIN, 0, node);
  b_conn_handler_remove (conn, BTP_PACKET_TYPE_SWITCH, 0, node);
  b_conn_handler_remove (conn, BTP_PACKET_TYPE_LEAVE, 0, node);


  /* Start handling any packet if there's a packet func */
  if (TREE->packet_func)  b_conn_handler_remove (conn, -1, -1, node);



}




/* ************************************************************ */


static void
receive_any (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;

  g_return_if_fail (packet);
  g_return_if_fail (node);
  g_return_if_fail (TREE);

  DEBUGP("receive_any\n");

  if (TREE->packet_func)
    TREE->packet_func (TREE, node, packet, TREE->packet_user_data);
}



static void
receive_ok_switch (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;
  BPacket* new_packet;

  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == B_PACKET_TYPE_OK);
  g_return_if_fail (packet->subtype == BTP_PACKET_OK_SUBTYPE_SWITCH);
  g_return_if_fail (TREE);

  DEBUGP ("receive_ok_switch\n");

  ++TREE->num_ok;

  /* Cancel the switch timeout */
  cancel_switch_timeout (TREE);

  /* ******************** */
  /* Checks		*/

  /* We should have a new parent.  It's possible that we btp_fail'ed
     the node and it recovered and sent us this OK. */
  if (TREE->new_parent == NULL)
    return;

  /* This node should be our new parent */
  if (TREE->new_parent != node)
    return;

  /* Don't want to switch to my current parent */
  if (btp_tree_is_parent(node))
    {
      g_warning ("Attempt to switch parents to current parent\n");
      return;
    }

  /* This node shouldn't be me, my child, or root.  It is probably my
     sibling. */
  if (btp_tree_is_child (node) || btp_tree_is_root (node) || btp_tree_is_me(node))
    {
      g_warning ("Received OK from odd relation\n");
      return;
    }


  /* ******************** */
  /* Action		*/

  /* Send a LEAVE to my old parent.  The parent will probably
     close the connection once it receives the LEAVE. */
  new_packet = btp_packet_new_leave (TREE->parent);
  b_conn_send_packet (TREE->parent->conn, new_packet);

  /* Delete my parent now (if it isn't the root).  Note this will
     actually send a BYE to the parent and then delete it. */
  if (!btp_tree_is_root(TREE->parent))
    {
      btp_tree_remove_node (TREE->parent);
      btp_node_delete (TREE->parent);
    }

  DEBUGP("switch from %s, %d to %s, %d\n",
	 TREE->parent->hostname, TREE->parent->port,
	 TREE->new_parent->hostname, TREE->new_parent->port);

  /* Switch to this parent */
  btp_tree_remove_sibling (TREE->new_parent);
  TREE->parent = TREE->new_parent;
  TREE->new_parent = NULL;

  /* Delete all siblings */
  while (TREE->siblings)
    {
      BtpNode* sibling = (BtpNode*) TREE->siblings->data;

      g_return_if_fail (TREE->me != sibling);
      g_return_if_fail (TREE->parent != sibling);

      TREE->siblings = g_list_remove_link (TREE->siblings, TREE->siblings);

      btp_tree_remove_node (sibling);
      btp_node_delete (sibling);
    }

  /* Reset sibling list */
  g_list_free (TREE->siblings);
  TREE->siblings = NULL;

  /* Get a new list of siblings */
  schedule_request_children (TREE, FALSE);

  /* Give the upper layer a hint that we have a new parent */
  if (TREE->hint_func)
    {
      (TREE->hint_func)(TREE, TREE->parent, BTP_HINT_NEW_PARENT, TREE->hint_user_data);
    }
}



static void
receive_error_switch (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;

  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == B_PACKET_TYPE_ERROR);
  g_return_if_fail (packet->subtype == BTP_PACKET_ERROR_SUBTYPE_SWITCH);
  g_return_if_fail (TREE);

  DEBUGP("receive_error_switch\n");

  ++TREE->num_error;

  /* Cancel the switch timeout */
  cancel_switch_timeout (TREE);

  /* ******************** */
  /* Checks		*/

  /* We should have a new parent.  It's possible that we
     btp_fail'ed the node and it recovered and sent us this OK.  */
  if (TREE->new_parent == NULL)
    {
      g_warning ("Received weird switch ERROR (I have no new parent)\n");
      return;
    }

  /* ******************** */
  /* Action		*/

  /* Note that this node is a bad parent so that we don't try to
     use it again (we might want to relax this later depending
     on why the switch failed). */
  TREE->new_parent->is_bad_parent = TRUE;

  /* Reset new_parent. */
  TREE->new_parent = NULL;

  /* Request children again.  Presumably we acted on bad
     information.  This will start the new parent search process
     again. */
  schedule_request_children(TREE, TRUE);
}



static void
receive_ucast (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;
  
  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == B_PACKET_TYPE_UCAST);
  g_return_if_fail (TREE);

  DEBUGP("receive_ucast\n");

  if (TREE->ucast_read_func)
    {
      TREE->ucast_read_func (TREE, node, packet->data, g_ntohs(packet->length),
			     TREE->ucast_read_user_data);
    }
}


static void
receive_mcast (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;
  
  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == B_PACKET_TYPE_MCAST);
  g_return_if_fail (TREE);

  DEBUGP("receive_mcast\n");

  if (TREE->mcast_read_func)
    {
      TREE->mcast_read_func (TREE, node, packet->data, g_ntohs(packet->length),
			     TREE->mcast_read_user_data);
    }
}


static void
receive_request_children (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;
  BPacket* new_packet;
  
  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == B_PACKET_TYPE_INFO_REQUEST);
  g_return_if_fail (packet->subtype == BTP_PACKET_INFO_SUBTYPE_CHILDREN);
  g_return_if_fail (TREE);

  DEBUGP("receive_request_children\n");

  new_packet = btp_packet_new_info_reply_children (node, TREE->children);
  b_conn_send_packet (node->conn, new_packet);
}



static void
receive_reply_children (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;
  GList* addresses;
  GList* i;

  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == B_PACKET_TYPE_INFO_REPLY);
  g_return_if_fail (packet->subtype == BTP_PACKET_INFO_SUBTYPE_CHILDREN);
  g_return_if_fail (TREE);

  DEBUGP("receive_reply_children\n");

  /* Ignore if from a node that isn't my parent or if I'm passive */
  if (node != TREE->parent || TREE->passive)
    return;

  /* Read the children addresses */
  addresses = b_util_read_addresses (packet->data, g_ntohs(packet->length));

  for (i = addresses; i != NULL; i = i->next)
    {
      BtpNode* sibling = NULL;
      BUtilAddress* address;

      /* Get the next address */
      address = (BUtilAddress*) i->data;

      /* Ignore if port is 0 */
      if (address->port == 0)
	{
	  g_warning ("Received child with port of 0\n");
	  continue;
	}

      /* Check if I already have this node */
      sibling = btp_tree_find_node (TREE, address->hostname, address->port);
      if (sibling == NULL)
	{
	  sibling = btp_node_new (TREE, address->hostname, address->port, NULL);
	  btp_tree_add_sibling (sibling);
	}

      /* If we don't have the sibling in our sibling list, put it
	 there.  I can't be my own sibling.  Ignore siblings that are
	 my children (I may have received outdated information from my
	 parent) */
      else if (sibling != TREE->me &&
	       !btp_tree_is_sibling(sibling) &&
	       !btp_tree_is_child(sibling))
	{
	  btp_tree_add_sibling (sibling);
	}

      g_free (address->hostname);
      g_free (address);
    }
  g_list_free (addresses);

  /* Connect to each of my siblings to get the ping distance */
  for (i = TREE->siblings; i != NULL; i = i->next)
    {
      BtpNode* sibling = (BtpNode*) i->data;

      b_conn_connect (sibling->conn);
    }

  /* Set a timer to later find a new parent based on ping times */
  schedule_find_new_parent (TREE);
}




static void
receive_join (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;

  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == BTP_PACKET_TYPE_JOIN);
  g_return_if_fail (TREE);

  DEBUGP("receive_join\n");

  ++TREE->num_join;

  /* ******************** */
  /* Checks		*/

  g_return_if_fail (!TREE->passive);		/* Don't allow a JOIN if passive. */
  g_return_if_fail (!btp_tree_is_me(node));  	/* Don't join myself.  */
  g_return_if_fail (!btp_tree_is_root(node));	/* Don't allow root to join. */


  /* ******************** */
  /* Action		*/

  /* If the node is already a child, 'ignore' the JOIN.  This
     might happen if the child resets and rejoins. */
  if (btp_tree_is_child(node))
    return;

  /* If the node is my parent, ignore it.  In practice, this
     shouldn't happen.  It could happen if the parent resets, is
     confused, and rejoins.  However, by the time this happen, I
     should have a new parent. */
  if (btp_tree_is_parent(node))
    return;

  /* Remove it from the sibling list.  This could happen if the
     node resets then sends us a JOIN for some reason. */
  if (btp_tree_is_sibling (node))
    btp_tree_remove_sibling (node);

  /* Add the node to my list of children */
  btp_tree_add_child (node);

  /* Give the upper level a hint */
  if (TREE->hint_func)
    {
      (TREE->hint_func)(TREE, node, BTP_HINT_NODE_JOIN, TREE->hint_user_data);
    }
}



static void
receive_switch (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;
  BUtilAddress* parent_addr = NULL;
  gboolean switch_ok = FALSE;

  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == BTP_PACKET_TYPE_SWITCH);
  g_return_if_fail (packet->subtype == 0);
  g_return_if_fail (TREE);

  DEBUGP ("receive_switch\n");

  ++TREE->num_switch;

  /* ******************** */
  /* Checks		*/

  g_return_if_fail (!TREE->passive);		/* Don't allow a SWITCH if passive */
  g_return_if_fail (!btp_tree_is_me(node));	/* Don't switch to myself */
  g_return_if_fail (!btp_tree_is_parent(node));	/* Don't allow parent to switch */
  g_return_if_fail (!btp_tree_is_root(node));   /* Don't allow root to switch */


  /* ******************** */
  /* Action		*/

  /* Can't switch if I'm also trying to switch. */
  if (TREE->new_parent != NULL)
    goto switch_done;

  /* Get the node's parent */
  parent_addr = b_util_read_address (packet->data, g_ntohs(packet->length));
  if (parent_addr == NULL)
    {
      g_warning ("Received weird data in SWITCH packet (length %d)\n", 
		 g_ntohs(packet->length));
      return;
    }

  /* Can't switch if we have different parents.  This might
     cause a loop. */
  if (parent_addr->port != TREE->parent->port ||
      strcmp(parent_addr->hostname, TREE->parent->hostname) != 0)
    {
      goto switch_done;
    }

  /* If the node is already my child, send an OK. */
  if (btp_tree_is_child(node))
    {
      switch_ok = TRUE;
      goto switch_done;
    }

  /* Ok to switch - remove the node from my sibling list (if
     it's there) and add to my children list */
  btp_tree_remove_sibling (node);
  btp_tree_add_child (node);
  
  switch_ok = TRUE;

 switch_done:

  if (parent_addr)
    {
      g_free(parent_addr->hostname);
      g_free(parent_addr);
    }

  /* Send an OK or error message */
  if (switch_ok)
    {
      BPacket* ok_packet;

      ok_packet = b_packet_new_ok (node->conn, BTP_PACKET_OK_SUBTYPE_SWITCH);
      b_conn_send_packet (node->conn, ok_packet);
    }
  else
    {
      BPacket* error_packet;

      ++TREE->num_switch_bad;

      error_packet = b_packet_new_error (node->conn, BTP_PACKET_ERROR_SUBTYPE_SWITCH);
      b_conn_send_packet (node->conn, error_packet);
    }
}



static void
receive_leave (BConn* conn, BPacket* packet, gpointer user_data)
{
  BtpNode* node = (BtpNode*) user_data;

  g_return_if_fail (node);
  g_return_if_fail (packet);
  g_return_if_fail (packet->type == BTP_PACKET_TYPE_LEAVE);
  g_return_if_fail (TREE);
  g_return_if_fail (node != TREE->me);

  DEBUGP ("receive_leave\n");

  /* Remove the child */
  if (btp_tree_is_child (node))
    {
      btp_tree_remove_child (node);
    }

  /* Delete the node via btp_conn_fail */
  btp_conn_fail (node->conn, node);
}




/* ************************************************************ */


static void
schedule_request_children (BtpTree* tree, gboolean soon)
{
  guint time;
  guint range;

  g_return_if_fail (tree);

  DEBUGP ("schedule_request_children\n");

  if (tree->request_children_timeout != 0)
    return;

  if (soon)
    {
      time =  REQUEST_CHILDREN_TIME;
      range = REQUEST_CHILDREN_RANGE;
    }
  else
    {
      time =  NEXT_REQUEST_CHILDREN_TIME;
      range = NEXT_REQUEST_CHILDREN_RANGE;
    }

  tree->request_children_timeout = 
    g_timeout_add (time + rand() % range, request_children_cb, tree);
}



static void
cancel_request_children (BtpTree* tree)
{
  g_return_if_fail (tree);

  DEBUGP ("cancel_request_children\n");

  if (tree->request_children_timeout)
    {
      g_assert (g_source_remove(tree->request_children_timeout));
      tree->request_children_timeout = 0;
    }
}



static gboolean
request_children_cb (gpointer data)
{
  BtpTree* tree = (BtpTree*) data;
  BPacket* packet;

  DEBUGP("request_children_cb\n");

  g_return_val_if_fail (tree, FALSE);
  g_return_val_if_fail (tree->parent, FALSE);

  tree->request_children_timeout = 0; 

  packet = btp_packet_new_info_request_children (tree->parent);
  b_conn_send_packet (tree->parent->conn, packet);

  return FALSE;
}




/* ************************************************************ */


static void
schedule_find_new_parent (BtpTree* tree)
{
  g_return_if_fail (tree);

  DEBUGP ("schedule_find_new_parent\n");

  if (!tree->find_new_parent_timeout)
    {
      tree->find_new_parent_timeout =
	g_timeout_add (FIND_NEW_PARENT_TIME + rand() % FIND_NEW_PARENT_RANGE, 
		       find_new_parent_cb, tree);
    }
}


#if 0
static void
cancel_find_new_parent (BtpTree* tree)
{
  g_return_if_fail (tree);

  DEBUGP ("cancel_find_new_parent\n");

  if (tree->find_new_parent_timeout)
    {
      g_assert (g_source_remove(tree->find_new_parent_timeout));
      tree->find_new_parent_timeout = 0;
    }
}
#endif


static gboolean
find_new_parent_cb (gpointer data)
{
  BtpTree* tree = (BtpTree*) data;
  BtpNode* new_parent;

  DEBUGP ("find_new_parent_cb\n");

  g_return_val_if_fail (tree, FALSE);

  /* I should not be root and should have a parent.  TODO: Is is
     possible that we can reach this point and not have a parent? */
  g_return_val_if_fail (tree->root != tree->me, FALSE);
  g_return_val_if_fail (tree->parent != NULL, FALSE);

  /* We should be in switching mode */
  g_return_val_if_fail (tree->find_new_parent_timeout != 0, FALSE);
  tree->find_new_parent_timeout = 0;

  /* We should not have a switching candidate or be in the process of
     switching. */
  g_return_val_if_fail (tree->new_parent == NULL, FALSE);
  g_return_val_if_fail (tree->switch_timeout == 0, FALSE);

  /* Figure out who the new parent should be */
  new_parent = find_new_parent (tree);

  /* Switch if we found a new parent */
  if (new_parent)
    {
      BPacket* switch_packet;

      DEBUGP("find_new_parent_cb: Switching parent\n");

      /* Cancel any request children timeout */
      cancel_request_children (tree);

      tree->new_parent = new_parent;
      /* We will remove the parent from the list of siblings in the
	 OK/ERROR code. */

      switch_packet = btp_packet_new_switch (new_parent, tree->parent);
      b_conn_send_packet (new_parent->conn, switch_packet);

      schedule_switch_timeout (tree);
    }

  else
    {
      DEBUGP("find_new_parent_cb: Did not find a better parent\n");

      /* Wait a while, request siblings, and try again. */
      schedule_request_children(tree, FALSE);
    }

  DEBUGP ("find_new_parent_cb DONE\n");

  return FALSE;
}



static BtpNode*
find_new_parent(BtpTree* tree)
{
  BtpNode* parent;
  BtpNode* closest;
  guint distance;
  GList* si;

  g_return_val_if_fail (tree != NULL, NULL);
  g_return_val_if_fail (tree->parent != NULL, NULL);

  DEBUGP ("find_new_parent\n");

  parent = tree->parent;

  /* Closest is our parent (unless we don't know the distance or the
     parent has missed more than one pong) */
  if (parent->conn->distance != 0)
    {
      closest = parent;
      distance = parent->conn->distance;
    }
  else
    {
      closest = NULL;
      distance = G_MAXUINT;
    }

  /* Find the closest sibling */
  for (si = tree->siblings; si != NULL; si = si->next)
    {
      BtpNode* sibling = (BtpNode*) si->data;

      /* Skip siblings that we don't have a distance for or are known
         bad parents. */
      if (sibling->conn->distance == 0 ||  sibling->is_bad_parent == TRUE)
	continue;

      /* If the sibling is closer, then save it */
      if (sibling->conn->distance <= distance)
	{
	  closest = sibling;
	  distance = sibling->conn->distance;
	}
    }

  /* Return NULL if we didn't find one or didn't find one better than
     our current parent */
  if (closest == NULL || closest == parent)
    return NULL;

  /* Heuristic: If the closest is less than 10 ms away, then
     definately switch to it. */
  if (closest->conn->distance <= 10)
    return closest;

  /* Otherwise, only switch if it's better than 25% closer _or_ the
     parent has more than 4 children.  This means that a node won't
     switch if a sibling is about as close as the parent _and_ the
     parent doesn't have many children.
  */
  if (((double) closest->conn->distance <= (.75 * (double) parent->conn->distance)) ||
      (g_list_length (tree->siblings) > 4))
    return closest;

  return NULL;
}



/* ************************************************************ */



static void
schedule_switch_timeout (BtpTree* tree)
{
  g_return_if_fail (tree != NULL);

  DEBUGP ("schedule_switch_timeout\n");

  if (!tree->switch_timeout)
    {
      tree->switch_timeout = 
	g_timeout_add (SWITCH_TIMEOUT_TIME + rand() % SWITCH_TIMEOUT_RANGE, 
		       switch_timeout_cb, tree);
    }
}


static void
cancel_switch_timeout (BtpTree* tree)
{
  g_return_if_fail (tree != NULL);

  DEBUGP ("cancel_schedule_switch_timeout\n");

  if (tree->switch_timeout)
    {
      g_assert (g_source_remove(tree->switch_timeout));
      tree->switch_timeout = 0;
    }
}


static gboolean
switch_timeout_cb(gpointer data)
{
  BtpTree* tree = (BtpTree*) data;

  g_return_val_if_fail (tree->new_parent, FALSE);
  g_return_val_if_fail (tree->switch_timeout != 0, FALSE);

  DEBUGP ("switch_timeout_cb\n");

  /* Mark the node as a bad parent */
  tree->new_parent->is_bad_parent = TRUE;

  /* Remove any switches that we might be about to send */
  b_conn_remove_packets (tree->new_parent->conn, BTP_PACKET_TYPE_SWITCH);

  /* Reset the new parent search */
  tree->new_parent = 0;
  tree->switch_timeout = 0;

  /* Request children and start again */
  schedule_request_children (tree, TRUE);

  return FALSE;
}

