/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _BTP_TREE_H
#define _BTP_TREE_H


#include <glib.h>
#include <gnet/gnet.h>

#include "btp_node.h"
#include "btp_packet.h"



/**

   Callbacks

   These callbacks provide an interface for BTP to communcate with
   higher level protocols.  Set the callbacks by modifying the BtpTree
   structure.

   The read callbacks (mcast and ucast) are called when there's data
   to be read.  The data is passed in the buffer and should be copied
   and read immediately.  BTP OWNS THE DATA - DO NOT DELETE IT.

   The hint callback gives the upper level a hint when various
   conditions happen.  Some of these aren't used, some are.  This is
   experimental.

   The error callback is called if there is an error.  The details
   haven't been figured out yet.

   The packet callback is called when a packet is received.  Return
   TRUE if you have fully processed the packet and BTP should process
   it no more; FALSE if BTP should process it.  This is not used.

*/

typedef void (*BtpMcastReadFunc)(BtpTree* tree, BtpNode* from_node,
				 guint8* buffer, guint length, 
				 gpointer user_data);


typedef void (*BtpUcastReadFunc)(BtpTree* tree, BtpNode* from_node, 
				 guint8* buffer, guint length,
				 gpointer user_data);

typedef enum {
  BTP_HINT_NEW_PARENT,		/* This is the new parent of the node	*/
  BTP_HINT_NODE_JOIN,		/* Node just JOINed			*/
  BTP_HINT_NODE_DOWN,		/* This node is down (or went away) - don't use it */
  BTP_HINT_UCAST_WRITE,		/* Node's out buffer is empty		*/
  BTP_HINT_MCAST_WRITE		/* Multicast out buffers are empty	*/
} BtpHint;

typedef void (*BtpHintFunc)(BtpTree* tree, BtpNode* node, 
			    BtpHint hint, gpointer user_data);

typedef void (*BtpErrorFunc)(BtpTree* tree, gpointer user_data);

typedef gboolean (*BtpPacketFunc)(BtpTree* tree, BtpNode* node, 
				  BPacket* packet, gpointer user_data);



/* 

   Tree types give a hint to BTP about what the tree will be used for.
   Right now this is just used to set the TOS bit on the outgoing
   packets.

 */
typedef enum 
{
  BTP_TREE_TYPE_NONE,		/* Tree is just a regular tree 			*/
  BTP_TREE_TYPE_BULK		/* Tree will be used for bulk file transfer	*/
} BtpTreeType;



/**

   BtpTree represents a multicast tree.  Call btp_connect_root to
   create one.

 */
struct _BtpTree
{
  gchar* name;			/* Name of the group the tree represents */
  gchar* url;			/* URL of the tree			 */
  BtpTreeType type;		/* Type of the tree			 */
  gboolean passive;		/* Am I a passive member (I don't join)  */
  guint ref_count;		/* Reference count			 */

  BtpNode* me;			/* Me					 */
  BtpNode* root;		/* Root (might be me)			 */
  BtpNode* parent;		/* My parent (might be NULL)		 */

  BtpNode* new_parent;		/* Parent I am switching too		 */
  guint switch_timeout;		/* Timeout on switch to new parent 	 */

  GList* nodes;			/* List of all nodes 			 */
  GList* children;		/* List of children (subset of nodes)	 */
  GList* siblings;		/* List of siblings (subset of nodes)	 */

  guint request_children_timeout; /* Timeout to request children	 */
  guint find_new_parent_timeout;  /* Timeout to find new parent		 */

  /* Statistics */
  guint num_connect;		/* Number of CONNECTs received	 	*/
  guint num_join;		/* Number of JOINs received	 	*/
  guint num_switch;		/* Number of SWITCHs received		*/
  guint num_switch_bad;		/* Number of bad SWITCHes received	*/
  guint num_ok;			/* Number of OKs received		*/
  guint num_error;		/* Number of ERRORs received		*/

  /* Upper level callbacks */
  BtpMcastReadFunc  mcast_read_func;
  BtpUcastReadFunc  ucast_read_func;
  BtpHintFunc	    hint_func;
  BtpErrorFunc	    error_func;

  BtpPacketFunc	    packet_func;
  /* node_fail_func */
  /* node_lost_connection_func */
  /* error_func */


  gpointer 	    mcast_read_user_data;
  gpointer 	    ucast_read_user_data;
  gpointer	    hint_user_data;
  gpointer	    error_user_data;
  gpointer	    packet_user_data;

};



/* **************************************** */


void 	 btp_tree_delete (BtpTree* tree);
/*  void	 btp_tree_print (FILE* file, BtpTree* tree); */


/* **************************************** */

gboolean btp_tree_is_me (BtpNode* node);
gboolean btp_tree_is_root (BtpNode* node);
gboolean btp_tree_is_parent (BtpNode* node);
gboolean btp_tree_is_new_parent (BtpNode* node);
gboolean btp_tree_is_related (BtpNode* node);
	    
gboolean btp_tree_is_sibling (BtpNode* node);
void	 btp_tree_add_sibling (BtpNode* node);
void	 btp_tree_remove_sibling (BtpNode* node);
	    
gboolean btp_tree_is_child (BtpNode* node);
void	 btp_tree_add_child (BtpNode* node);
void	 btp_tree_remove_child (BtpNode* node);

gboolean btp_tree_is_node (BtpNode* node);
void	 btp_tree_add_node (BtpNode* node);
void	 btp_tree_remove_node (BtpNode* node);
BtpNode* btp_tree_find_node (BtpTree* tree, gchar* hostname, gint port);

/* **************************************** */


#endif /* _BTP_TREE_H */
