/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <sys/time.h>
#include <unistd.h>

#include "btsmp.h"

#include "globals.h"
#include "util.h"

#include "btp.h"
#include "b_conn.h"


#define BTSMP_MAX_CACHE_SIZE		16384		/* 16 kb is default max */

typedef struct _BTSmpCacheEntry
{
  guint length;
  guint8* buffer;
  struct timeval time;

} BTSmpCacheEntry;


static void	btsmp_mcast_read_cb(BtpTree* tree, BtpNode* from_node, 
				    guint8* buffer, guint length, gpointer user_data);
static void	btsmp_ucast_read_cb(BtpTree* tree, BtpNode* from_node, 
				    guint8* buffer, guint length, gpointer user_data);
static void 	btsmp_error_cb(BtpTree* tree, gpointer user_data);
static void 	btsmp_hint_cb(BtpTree* tree, BtpNode* node, BtpHint hint, gpointer user_data);

static void 	btsmp_cache_add(BTSmp* btsmp, guint8* buffer, guint length);
static void	btsmp_cache_check(BTSmp* btsmp);
 


BTSmp* 
btsmp_create(gchar* groupname)
{
  BtpTree* tree;
  BTSmp* btsmp = NULL;

  g_return_val_if_fail (groupname != NULL, NULL);

  DEBUGP ("btsmp_create\n");

  /* Create the BTP tree */
  tree = btp_create (groupname, BTP_TREE_TYPE_NONE);
  if (tree == NULL)
    return NULL;

  /* Create the BTSmp */
  btsmp = g_new0 (BTSmp, 1);
  btsmp->tree = tree;

  /* Set up cache */
  btsmp->cache = NULL;
  btsmp->cache_size = 0;
  btsmp->max_cache_size = BTSMP_MAX_CACHE_SIZE;
  btsmp->cache_timeout = 0;

  /* Set up call backs */
  tree->mcast_read_func  	= btsmp_mcast_read_cb;
  tree->ucast_read_func  	= btsmp_ucast_read_cb;
  tree->error_func	 	= btsmp_error_cb;
  tree->hint_func	 	= btsmp_hint_cb;
  tree->mcast_read_user_data 	= btsmp;
  tree->ucast_read_user_data 	= btsmp;
  tree->error_user_data      	= btsmp;
  tree->hint_user_data      	= btsmp;

  /* Create the url */
  g_assert (tree->url != NULL);
  g_assert (strstr(tree->url, "btp://") == tree->url);
  btsmp->url = g_strdup_printf("btsmp://%s", &tree->url[sizeof("btp://") - 1]);

  /* TODO?: Don't return the BTSMP yet - wait until we actually connect
     to the tree. */

  return btsmp;
}



BTSmp* 
btsmp_join(gchar* url)
{
  BtpTree* tree;
  BTSmp* btsmp = NULL;

  g_return_val_if_fail (url != NULL, NULL);


  DEBUGP ("btsmp_join\n");

  /* Create the BTP tree */
  tree = btp_join (url, BTP_TREE_TYPE_NONE, FALSE);
  if (tree == NULL)
    return NULL;

  /* Create the BTSmp */
  btsmp = g_new0 (BTSmp, 1);
  btsmp->tree = tree;
  btsmp->url = g_strdup (url);

  /* Set up cache */
  btsmp->cache = NULL;
  btsmp->cache_size = 0;
  btsmp->max_cache_size = BTSMP_MAX_CACHE_SIZE;
  btsmp->cache_timeout = 0;

  /* Set up call backs */
  tree->mcast_read_func  	= btsmp_mcast_read_cb;
  tree->ucast_read_func  	= btsmp_mcast_read_cb;
  tree->error_func	 	= btsmp_error_cb;
  tree->hint_func	 	= btsmp_hint_cb;
  tree->mcast_read_user_data 	= btsmp;
  tree->ucast_read_user_data 	= btsmp;
  tree->error_user_data      	= btsmp;
  tree->hint_user_data      	= btsmp;

  return btsmp;
}


void
btsmp_delete(BTSmp* btsmp)
{
  if (btsmp)
    {
      btp_clear_and_unref (btsmp->tree);
      g_free (btsmp->url);

      btsmp_cache_clear (btsmp);

      /* Zero out struct for safety */
      memset (btsmp, 0, sizeof(*btsmp));

      g_free (btsmp);
    }
}



void
btsmp_send_packet(BTSmp* btsmp, guint8* buffer, guint length)
{
  g_return_if_fail (btsmp);
  g_return_if_fail (buffer);
  g_return_if_fail (length < BTSMP_MAX_PACKET_SIZE);

  DEBUGP("btsmp_send_packet %s\n", btsmp->url);

  /* Send the packet */
  btp_send_mcast (btsmp->tree, NULL, buffer, length);

  /* Add to cache */
  btsmp_cache_add (btsmp, buffer, length);
}


void
btsmp_cache_set_policy(BTSmp* btsmp, guint max_cache_size, guint cache_timeout)
{
  g_return_if_fail (btsmp);

  btsmp->max_cache_size = MIN(BTSMP_MAX_CACHE_SIZE, max_cache_size);
  btsmp->cache_timeout = cache_timeout;

  btsmp_cache_check (btsmp);
}


void
btsmp_cache_clear(BTSmp* btsmp)
{
  GList* i;

  g_return_if_fail (btsmp);

  for (i = btsmp->cache; i != NULL; i = i->next)
    {
      BTSmpCacheEntry* entry = (BTSmpCacheEntry*) i->data;
      g_assert (entry != NULL);

      g_free (entry->buffer);
      g_free (entry);
    }

  g_list_free (btsmp->cache);
  btsmp->cache = NULL;

  btsmp->cache_size = 0;
}



static void
btsmp_mcast_read_cb(BtpTree* tree, BtpNode* from_node, 
		    guint8* buffer, guint length, gpointer user_data)
{
  BTSmp* btsmp = (BTSmp*) user_data;

  g_return_if_fail (tree);
  g_return_if_fail (from_node);
  g_return_if_fail (buffer);	/* Can send 0 length packets? */
  g_return_if_fail (btsmp);


  DEBUGP ("btsmp_mcast_read_cb %s\n", btsmp->url);

  /* Pass packet up */
  if (btsmp->packet_func)
    {
      if (!(btsmp->packet_func)(btsmp, buffer, length, btsmp->packet_user_data))
	{
	  /* Send on */
	  btp_send_mcast (tree, from_node, buffer, length);

	  /* Add to cache */
	  btsmp_cache_add (btsmp, buffer, length);
	}
    }
}


static void
btsmp_ucast_read_cb(BtpTree* tree, BtpNode* from_node, 
		    guint8* buffer, guint length, gpointer user_data)
{
  BTSmp* btsmp = (BTSmp*) user_data;

  g_return_if_fail (tree);
  g_return_if_fail (from_node);
  g_return_if_fail (buffer);	/* Can send 0 length packets? */
  g_return_if_fail (btsmp);

  DEBUGP ("btsmp_ucast_read_cb %s\n", btsmp->url);

  /* Pass packet up */
  if (btsmp->packet_func)
    {
      if (!(btsmp->packet_func)(btsmp, buffer, length, btsmp->packet_user_data))
	{
	  /* Don't pass mcast on */

	  /* Add to cache.  We assume mcast and ucast packets are the
             same.  We may want to make this an option. */
	  btsmp_cache_add (btsmp, buffer, length);
	}
    }
}


static void 	
btsmp_error_cb(BtpTree* tree, gpointer user_data)
{
  BTSmp* btsmp = (BTSmp*) user_data;

  g_return_if_fail (tree);
  g_return_if_fail (btsmp);


  DEBUGP ("btsmp_error_cb\n");

  /* Pass error up */
  if (btsmp->error_func)
    (btsmp->error_func)(btsmp, btsmp->error_user_data);
}


static void
btsmp_hint_cb(BtpTree* tree, BtpNode* node, BtpHint hint, gpointer user_data)
{
  BTSmp* btsmp = (BTSmp*) user_data;
  GList* i;

  g_return_if_fail (tree);
  g_return_if_fail (btsmp);


  /* Look at JOIN hints only */
  if (hint != BTP_HINT_NODE_JOIN)
    return;

  DEBUGP ("btsmp_hint_cb\n");

  g_return_if_fail (node != NULL);

  /* Check the cache.  This will remove any old entries. */
  btsmp_cache_check (btsmp);

  /* Send the new node all the packets in our cache */
  for (i = btsmp->cache; i != NULL; i = i->next)
    {
      BTSmpCacheEntry* entry = (BTSmpCacheEntry*) i->data;

      btp_send_ucast (btsmp->tree, node, entry->buffer, entry->length);
    }
}


static void
btsmp_cache_add(BTSmp* btsmp, guint8* buffer, guint length)
{
  BTSmpCacheEntry* entry;

  if (btsmp->max_cache_size == 0)
    return;

  entry = g_new(BTSmpCacheEntry, 1);
  entry->length = length;
  entry->buffer = g_new(guint8, length);
  memcpy(entry->buffer, buffer, length);
  g_assert (gettimeofday(&entry->time, NULL) == 0);

  btsmp->cache = g_list_append (btsmp->cache, entry);
  btsmp->cache_size += length;

  btsmp_cache_check (btsmp);
}


static void
btsmp_cache_check(BTSmp* btsmp)
{
  BTSmpCacheEntry* entry;

  /* If the cache is too big, remove entries until small enough */
  while (btsmp->cache_size > btsmp->max_cache_size)
    {
      g_assert (btsmp->cache != NULL);

      entry = (BTSmpCacheEntry*) btsmp->cache->data;
      g_assert (entry != NULL);

      btsmp->cache_size -= entry->length;
      btsmp->cache = g_list_remove(btsmp->cache, entry);

      g_free(entry->buffer);
      g_free(entry);
    }


  /* Remove entries that have expired */
  if (btsmp->cache_timeout > 0 && btsmp->cache != NULL)
    {
      struct timeval timeofday;

      g_assert (gettimeofday(&timeofday, NULL) == 0);

      while (btsmp->cache != NULL)
	{
	  struct timeval diff;

	  entry = (BTSmpCacheEntry*) btsmp->cache->data;
	  g_assert (entry != NULL);

	  timersub(&timeofday, &entry->time, &diff);

	  if ((diff.tv_sec * 1000 + diff.tv_usec / 1000) > btsmp->cache_timeout)
	    {
	      btsmp->cache_size -= entry->length;
	      btsmp->cache = g_list_remove(btsmp->cache, entry);

	      g_free(entry->buffer);
	      g_free(entry);
	    }
	  else
	    break;
	}
    }
}


gboolean
btsmp_is_up (BTSmp* btsmp)
{
  if (btsmp && btsmp->tree)
    {
      if (btp_tree_is_root (btsmp->tree->me))
	return TRUE;
      
      if (btsmp->tree->parent &&
	  btsmp->tree->parent->conn &&
	  b_conn_is_connected (btsmp->tree->parent->conn))
	return TRUE;
    }

  return FALSE;
}
