/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "ggui.h"
#include "ggui_channels.h"
#include "ggui_config.h"
#include "ggui_files.h"
#include "ggui_menu.h"


/* ******************** */

static GladeXML* 	menu_xml = NULL;

static GtkCList* 	channels_clist = NULL;
static GtkWidget* 	channels_popup = NULL;

static JMFile*		selected_channel = NULL;

static GList*		bookmarked_channels = NULL;


/* ******************** */

static void channels_add (JMFile* jmfile);


void on_channels_select_row(GtkCList* list, gint row, gint column, GdkEvent* event);
void on_channels_unselect_row(GtkCList* list, gint row, gint column, GdkEvent* event);
gint on_channels_button_press (GtkCList* list, GdkEventButton *event, gpointer user_data);

void on_channels_properties(gpointer callback_data, guint callback_action, GtkWidget* widget);


/* ************************************************************ */


void
ggui_channels_init(void)
{
  menu_xml = ggui_get_jm_glade_xml("channels_popup", "master_popup.glade");
  g_assert (menu_xml != NULL);
  glade_xml_signal_autoconnect(menu_xml);

  channels_clist = GTK_CLIST(glade_xml_get_widget(jmw_xml, "channels_clist"));
  channels_popup = GTK_WIDGET(glade_xml_get_widget(menu_xml, "channels_popup"));

  g_assert (channels_clist != NULL);
  g_assert (channels_popup != NULL);

  ggui_set_tooltip (GTK_WIDGET(channels_clist), _("Channels you have joined."));
}


void
ggui_channels_shutdown (void)
{
  char** argv;
  gint argc;
  GList* i;

  argc = g_list_length (bookmarked_channels);
  argv = g_new0 (gchar*, argc + 1);

  for (argc = 0, i = bookmarked_channels; i != NULL; i = i->next)
    {
      JMFile* jmfile = (JMFile*) i->data;

      if (jmfile->url)
	argv[argc++] = jmfile->url;
    }

  ggui_config_set_vector ("/Jungle Monkey/General/Bookmarks", argv);

  g_free (argv);
}



void
ggui_channels_load_bookmarks (JMFile* root_channel)
{
  gchar** argv;
  gint i;

  argv = ggui_config_get_vector ("/Jungle Monkey/General/Bookmarks");
  if (!argv)
    return;

  for (i = 0; argv[i] != NULL; ++i)
    {
      JMFile* jmfile;

      /* Skip if NULL */
      if (!*argv[i])
	continue;

      /* Skip if equal to the root channel */
      if (root_channel && root_channel->url && 
	  !strcmp(root_channel->url, argv[i]) )
	continue;

      /* Add it */
      jmfile = backend_channel_add (argv[i], NULL);
      if (jmfile)
	{
	  ggui_channels_show (jmfile);
	  ggui_channels_set_bookmark (jmfile, TRUE);
	}
    }

  g_free (argv);
}



void
ggui_channels_select(JMFile* jmfile)
{
  gint row;

  g_return_if_fail (jmfile != NULL);

  row = gtk_clist_find_row_from_data (channels_clist, jmfile);
  if (row >= 0)
    gtk_clist_select_row (channels_clist, row, 0);
}


JMFile* 
ggui_channels_get_selected(void)
{
  return selected_channel;
}


void
ggui_channels_set_bookmark (JMFile* jmfile, gboolean bookmark)
{
  g_return_if_fail (jmfile != NULL);
  g_return_if_fail (jmfile->type == JMFILE_CHANNEL);

  if (!g_list_find (bookmarked_channels, jmfile))
    {
      bookmarked_channels = g_list_prepend (bookmarked_channels, jmfile);
    }
}


void
ggui_channels_show(JMFile* jmfile)
{
  g_return_if_fail (jmfile != NULL);
  g_return_if_fail (jmfile->type == JMFILE_CHANNEL);

  /* Ignore if we're already showing this file */
  if (gtk_clist_find_row_from_data(channels_clist, jmfile) >= 0)
    return;

  channels_add (jmfile);
}


void
ggui_channels_add(JMFile* jmfile)
{
  g_return_if_fail (jmfile != NULL);

  if (jmfile->type != JMFILE_CHANNEL)
    return;

  /* Ignore if we aren't on it, or it's a normal file */
  if (!jmfile_on_channel(jmfile) && 
      jmfile->status == JMFILE_NONE &&
      !jmfile_has_parent(jmfile, NULL))
    return;

  channels_add (jmfile);
}


static void
channels_add (JMFile* jmfile)
{
  gchar* texts[2] = {NULL, NULL};
  gint row;

  DEBUGP("channels_add\n");

  g_return_if_fail (jmfile != NULL);
  g_return_if_fail (jmfile->name != NULL);
  g_return_if_fail (jmfile->type == JMFILE_CHANNEL);

  texts[0] = jmfile->name;
  row = gtk_clist_append (channels_clist, texts);
  gtk_clist_set_row_data(channels_clist, row, jmfile);

  ggui_channels_update(jmfile);
}


void
ggui_channels_update(JMFile* jmfile)
{
  GdkPixmap* pixmap;
  GdkBitmap* mask;
  gint row;

  g_return_if_fail (jmfile);

  DEBUGP ("gui_channels_update\n");

  if (jmfile->type != JMFILE_CHANNEL)
    return;
  
 try_again:
  row = gtk_clist_find_row_from_data (channels_clist, jmfile);
  if (row < 0)
    {
      /* Ignore if we aren't on it, or it's just a normal channel */
      if (!jmfile_on_channel(jmfile) && jmfile->status == JMFILE_NONE)
	return;

      channels_add (jmfile);
      goto try_again;
    }

  ggui_get_pixmaps (jmfile, &pixmap, &mask);

  gtk_clist_set_pixtext (channels_clist, row, 0, 
			 jmfile->name, 5, pixmap, mask);
}


void
ggui_channels_remove(JMFile* jmfile)
{
  gint row;

  g_return_if_fail (jmfile);

  bookmarked_channels = g_list_remove (bookmarked_channels, jmfile);

  row = gtk_clist_find_row_from_data(channels_clist, jmfile);
  if (row >= 0)
    gtk_clist_remove(channels_clist, row);
}


void
ggui_channels_freeze(void)
{
  gtk_clist_freeze (channels_clist);
}


void
ggui_channels_thaw(void)
{
  gtk_clist_thaw (channels_clist);
}


/* **************************************** */


void 
on_channels_select_row(GtkCList* list, gint row, gint column, GdkEvent* event)
{
  selected_channel = (JMFile*) gtk_clist_get_row_data(list, row);
  ggui_files_show (selected_channel);
}


void 
on_channels_unselect_row(GtkCList* list, gint row, gint column, GdkEvent* event)
{
  selected_channel = NULL;
  ggui_files_show (NULL);
}



#define HIDE_MENUITEM(MI) do {       					\
  GtkWidget* menu_item = glade_xml_get_widget(menu_xml, (MI));		\
  g_return_val_if_fail (menu_item != NULL, TRUE);			\
  gtk_widget_hide(menu_item);} while (0)

#define SHOW_MENUITEM(MI) do {       					\
  GtkWidget* menu_item = glade_xml_get_widget(menu_xml, (MI));		\
  g_return_val_if_fail (menu_item != NULL, TRUE);			\
  gtk_widget_show(menu_item);} while (0)



/* Return TRUE if we handle event; FALSE to pass on */
gint
on_channels_button_press (GtkCList* list, GdkEventButton *event, gpointer user_data)
{
  gint row, column;
  gboolean rv = FALSE;

  /* User double clicked - activate the thing selected */
  if (event->type == GDK_2BUTTON_PRESS && 
      (event->button == 1 || event->button == 2))
    {
      /* Get the selection information */
      if (gtk_clist_get_selection_info(list, event->x, event->y, &row, &column))
	{
	  /* Select the clicked row, if there is one */
	  gtk_clist_select_row (list, row, 0);
	
	  /* Double clicks on files and channels activate them.  We
	     don't do anything for directories because that might be
	     annoying. */
	  if (selected_channel)
	    backend_activate(selected_channel);
	}
      else
	gtk_clist_unselect_all(list);

      rv = TRUE;
    }

  /* User right-clicked - show context menu */
  else if (event->type == GDK_BUTTON_PRESS && event->button == 3)
    {
      GList* i;

      /* Select the clicked row, if there is one */
      if (gtk_clist_get_selection_info (list, event->x, event->y, &row, &column))
	gtk_clist_select_row (list, row, 0);
      else
	gtk_clist_unselect_all(list);

      /* Hide everything */
      for (i = GTK_MENU_SHELL(channels_popup)->children; i != NULL; i = i->next)
	{
	  GtkWidget* menu_item = GTK_WIDGET(i->data);
	  g_return_val_if_fail (GTK_MENU_ITEM(menu_item) != NULL, FALSE);
	  gtk_widget_hide(menu_item);
	}

      SHOW_MENUITEM("open_channel_mi");
      SHOW_MENUITEM("create_root_channel_mi");

      /* Build the menu */
      if (selected_channel)
	{
	  if (selected_channel->is_local)
	    {
	      SHOW_MENUITEM("remove_channel_mi");
	      SHOW_MENUITEM("separator1");
	    }
	  else if (btsmp_is_up(selected_channel->btsmp))
	    {
	      SHOW_MENUITEM("leave_channel_mi");
	      SHOW_MENUITEM("remove_channel_mi");
	      SHOW_MENUITEM("separator1");
	    }
	  else if (selected_channel->status != JMFILE_STATUS_CONNECTING)
	    {
	      SHOW_MENUITEM("join_channel_mi");
	      SHOW_MENUITEM("remove_channel_mi");
	      SHOW_MENUITEM("separator1");
	    }

	  SHOW_MENUITEM("separator3");
	  SHOW_MENUITEM("properties_mi");

	  if (!selected_channel->is_local)
	    {
	      GtkCheckMenuItem* item = 
		GTK_CHECK_MENU_ITEM(glade_xml_get_widget(menu_xml, "bookmark_mi"));
	      g_return_val_if_fail (item != NULL, TRUE);

	      if (g_list_find(bookmarked_channels, selected_channel))
		gtk_check_menu_item_set_active (item, TRUE);
	      else
		gtk_check_menu_item_set_active (item, FALSE);

	      gtk_widget_show(GTK_WIDGET(item));
	    }
	}
      else
	{
	  SHOW_MENUITEM("separator2");
	  SHOW_MENUITEM("exit_mi");
	}

      gtk_menu_popup (GTK_MENU(channels_popup), NULL, NULL, NULL, NULL, 3, event->time);

      rv = TRUE;
    }

  return rv;
}



void
on_channels_join_channel(gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_join_channel(selected_channel);
}


void
on_channels_open_channel(gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_open_channel();
}

void
on_channels_create_root_channel(gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_create_root_channel();
}

void
on_channels_remove_channel(gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  g_return_if_fail (selected_channel != NULL);

  /*

    TODO:

    We do not elegantly handle the case where I'm leaving a channel
    where I'm offering stuff.  We might want to ask the user if
    they're sure they want to leave the channel.

  */


  if (selected_channel->is_local)
    {
      ggui_menu_remove_channel (selected_channel);
    }
  else if (btsmp_is_up(selected_channel->btsmp))
    {
      ggui_menu_leave_channel (selected_channel);
      ggui_channels_remove (selected_channel);
    }
  else if (selected_channel->status != JMFILE_STATUS_CONNECTING)
    {
      ggui_channels_remove (selected_channel);
    }
}

void
on_channels_leave_channel(gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_leave_channel(selected_channel);
}

void
on_channels_properties(gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_properties(selected_channel);
}

void
on_channels_bookmark(gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  GtkCheckMenuItem* item = GTK_CHECK_MENU_ITEM(callback_data);

  g_return_if_fail (selected_channel != NULL);

  if (item->active)
    {
      if (!g_list_find(bookmarked_channels, selected_channel))
	bookmarked_channels = g_list_prepend(bookmarked_channels, selected_channel);
    }
  else
    {
      bookmarked_channels = g_list_remove(bookmarked_channels, selected_channel);
    }
}
