/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "ggui.h"
#include "ggui_channels.h"
#include "ggui_chat.h"
#include "ggui_menu.h"
#include "ggui_files.h"
#include "util.h"

/* ******************** */

static GladeXML* 	menu_xml = NULL;

static GtkCList* 	files_clist = NULL;
static GtkWidget*	files_popup = NULL;

static JMFile*		selected_file = NULL;
static JMFile*		shown_file = NULL;


enum {
  DND_TARGET_STRING,
  DND_TARGET_URI_LIST,
};

static GtkTargetEntry dnd_target_table[] = {
  { "text/uri-list", 	0, 	DND_TARGET_URI_LIST },
  { "text/plain", 	0, 	DND_TARGET_STRING },
  { "STRING",     	0, 	DND_TARGET_STRING },
};




/* ******************** */

static gint files_row_compare (GtkCList* clist, gconstpointer ptr1, gconstpointer ptr2);

void on_files_select_row (GtkCList* list, gint row, gint column, GdkEvent* event);
void on_files_unselect_row (GtkCList* list, gint row, gint column, GdkEvent* event);
gint on_files_button_press (GtkCList* list, GdkEventButton *event, gpointer user_data);

gboolean on_files_drag_motion (GtkCList* list, GdkDragContext* context,
			       gint x, gint y, guint time);
void on_files_drag_data_received (GtkCList* list, GdkDragContext* context,
				  gint x, gint y, GtkSelectionData* data,
				  guint info, guint time, gpointer user_data);



void
ggui_files_init(void)
{
  menu_xml = ggui_get_jm_glade_xml("files_popup", "master_popup.glade");
  g_assert (menu_xml != NULL);
  glade_xml_signal_autoconnect(menu_xml);

  /* Get the widgets */
  files_clist = GTK_CLIST(glade_xml_get_widget(jmw_xml,   "files_clist"));
  files_popup = GTK_WIDGET(glade_xml_get_widget(menu_xml, "files_popup"));

  g_assert (files_clist != NULL);
  g_assert (files_popup != NULL);

  gtk_clist_set_compare_func (files_clist, files_row_compare);
  gtk_clist_set_auto_sort (files_clist, TRUE);

  gtk_clist_column_title_passive (files_clist, 0);
  gtk_clist_column_title_passive (files_clist, 1);

  /* Set up tooltips */
  ggui_set_tooltip (GTK_WIDGET(files_clist), 
		    _("Files, directories, and channels available. "
		      "Right-click on something to get a popup-menu of options.  "
		      "Double-click on something to activate it."));

  /* Set up drop and drag */
  gtk_drag_dest_set (GTK_WIDGET(files_clist),
		     GTK_DEST_DEFAULT_ALL,
		     dnd_target_table, 
		     sizeof(dnd_target_table) / sizeof(dnd_target_table[0]),
		     GDK_ACTION_COPY /* | GDK_ACTION_MOVE */);
}



static gint
files_row_compare (GtkCList* clist, gconstpointer ptr1, gconstpointer ptr2)
{
  char *text1 = NULL;
  char *text2 = NULL;
  JMFile* file1 = NULL;
  JMFile* file2 = NULL;

  GtkCListRow* row1 = (GtkCListRow *) ptr1;
  GtkCListRow* row2 = (GtkCListRow *) ptr2;

  text1 = GTK_CELL_PIXTEXT (row1->cell[clist->sort_column])->text;
  text2 = GTK_CELL_PIXTEXT (row2->cell[clist->sort_column])->text;

  if (!text2)
    return (text1 != NULL);

  if (!text1)
    return -1;

  if (!strcmp(text1, _("..")))
    return -1;

  if (!strcmp(text2, _("..")))
    return 1;

  file1 = (JMFile*) row1->data;
  file2 = (JMFile*) row1->data;

  if (file1 && file2 && (file1->type != file2->type))
    {
      if (file1->type == JMFILE_CHANNEL && 
	  (file2->type == JMFILE_FILE || file2->type == JMFILE_DIR))
	return -1;

      else if (file1->type == JMFILE_DIR && file2->type == JMFILE_FILE)
	return -1;

      return 1;
    }

  return strcmp (text1, text2);
}


void
ggui_files_show(JMFile* jmfile)
{
  GList* i;

  /* Ignore if we're already showing this file */
  if (jmfile == shown_file)
    return;

  ggui_files_freeze ();

  /* Reset the GuiData's */
  if (shown_file)
    {
      for (i = shown_file->children; i != NULL; i = i->next)
	{
	  JMFile* child = (JMFile*) i->data;

	  if (child->gui_data)
	    child->gui_data->in_files_list = FALSE;
	}
    }

  /* Reset selected_file */
  selected_file = NULL;

  /* Clear the list */
  gtk_clist_clear (files_clist);

  /* Set the shown file */
  shown_file = jmfile;

  ggui_files_thaw ();

  /* We're done if the file is NULL */
  if (jmfile == NULL)
    return;

  /* Add all children files */
  for (i = jmfile->children; i != NULL; i = i->next)
    ggui_files_add ((JMFile*) i->data);

  /* Add a back button if this is a directory */
  if (jmfile->type == JMFILE_DIR)
    {
      int row;
      gchar* texts[2] = {NULL, NULL};

      texts[0] = _("..");
      row = gtk_clist_insert (files_clist, 0, texts);  
      gtk_clist_set_row_data(files_clist, row, back_pixmap);
      ggui_clist_update_row2(files_clist, 0, _(".."), NULL, back_pixmap, back_mask);
    }
}


void
ggui_files_add(JMFile* jmfile)
{
  gchar* texts[2] = {NULL, NULL};
  gint row;
  GuiData* gui_data;

  g_return_if_fail (jmfile != NULL);
  g_return_if_fail (files_clist != NULL);

  /* Ignore if no shown file or does not have this parent */
  if (!shown_file || !jmfile_has_parent(jmfile, shown_file))
    return;

  /* Ignore searches and orphaned files */
  if ((jmfile->type == JMFILE_FILE && jmfile_is_orphan(jmfile)) ||
      jmfile->type == JMFILE_SEARCH)
    {
      return;
    }

  ggui_files_freeze();

  /* Add to clist */ 
  texts[0] = jmfile->name;
  row = gtk_clist_insert (files_clist, 0, texts);
  gtk_clist_set_row_data (files_clist, row, jmfile);
  gtk_clist_sort (files_clist);

  /* Set in_files_list flag */
  gui_data = ggui_get_gui_data(jmfile);
  gui_data->in_files_list = TRUE;

  /* Update the row (will set icon and size) */
  ggui_files_update(jmfile);

  ggui_files_thaw();
}


void
ggui_files_remove (JMFile* jmfile)
{
  GuiData* gui_data;

  g_return_if_fail (jmfile != NULL);
  g_return_if_fail (files_clist != NULL);

  gui_data = jmfile->gui_data;
  if (gui_data && gui_data->in_files_list)
    {
      gint row;

      row = gtk_clist_find_row_from_data(files_clist, jmfile);
      if (row >= 0)
	gtk_clist_remove(files_clist, row);

      gui_data->in_files_list = FALSE;
    }

  if (jmfile == selected_file)
    selected_file = NULL;

  if (jmfile == shown_file)
    {
      shown_file = NULL;
      gtk_clist_clear (files_clist);
    }
}


void
ggui_files_update(JMFile* jmfile)
{
  GuiData* gui_data;
  gint row;
  gchar* text = NULL;
  gchar* length = NULL;
  GdkPixmap* pixmap = NULL;
  GdkBitmap* mask = NULL;

  g_return_if_fail (jmfile != NULL);
  g_return_if_fail (files_clist != NULL);

  gui_data = jmfile->gui_data;
  if (!gui_data || !gui_data->in_files_list)
    return;

  row = gtk_clist_find_row_from_data(files_clist, jmfile);
  g_return_if_fail (row >= 0);

  if (jmfile->type == JMFILE_FILE && jmfile->length)
    length = filesize_to_string(jmfile->length);

  ggui_get_pixmaps(jmfile, &pixmap, &mask);

  switch (jmfile->status)
    {
    case JMFILE_STATUS_CONNECTING:
      text = g_strdup_printf(_("%s (connecting)"), jmfile->name); 	
      break;

    case JMFILE_STATUS_TRANSFERRING:
      text = g_strdup_printf(_("%s (downloading)"), jmfile->name); 
      break;

    case JMFILE_STATUS_DOWN:
      text = g_strdup_printf(_("%s (down)"), jmfile->name); 
      break;

    default:
      text = g_strdup(jmfile->name); 
      break;
    }

  ggui_clist_update_row2 (files_clist, row, text, length, pixmap, mask);

  g_free(text);
  g_free(length);
}


void
ggui_files_freeze(void)
{
  g_return_if_fail (files_clist);

  gtk_clist_freeze (files_clist);
}


void	
ggui_files_thaw(void)
{
  g_return_if_fail (files_clist);

  gtk_clist_thaw (files_clist);
}



/* ************************************************************ */


void 
on_files_select_row (GtkCList* list, gint row, gint column, GdkEvent* event)
{
  gpointer selectee;
  
  selectee = gtk_clist_get_row_data(list, row);
  if (selectee != back_pixmap)
    selected_file = (JMFile*) selectee;
}


void
on_files_unselect_row (GtkCList* list, gint row, gint column, GdkEvent* event)
{
  selected_file = NULL;
}



#define HIDE_MENUITEM(MI) do {       					\
  GtkWidget* menu_item = glade_xml_get_widget(menu_xml, (MI));		\
  g_return_val_if_fail (menu_item != NULL, FALSE);			\
  gtk_widget_hide(menu_item);} while(0)

#define SHOW_MENUITEM(MI) do {       					\
  GtkWidget* menu_item = glade_xml_get_widget(menu_xml, (MI));		\
  g_return_val_if_fail (menu_item != NULL, FALSE);			\
  show_menu = TRUE;							\
  gtk_widget_show(menu_item);} while(0)


/* Return TRUE if we handle event; FALSE to pass on */
gint
on_files_button_press (GtkCList* list, GdkEventButton *event, gpointer user_data)
{
  gint row, column;
  gboolean rv = FALSE;

  /* User double clicked - activate the thing selected */
  if (event->type == GDK_2BUTTON_PRESS && 
      (event->button == 1 || event->button == 2))
    {
      /* Get the selection information */
      if (gtk_clist_get_selection_info(list, event->x, event->y, &row, &column))
	{
	  /* Select the clicked row, if there is one */
	  gtk_clist_select_row (list, row, 0);
	
	  /* The first row is 'back' */
	  if (row == 0 && shown_file && shown_file->type == JMFILE_DIR)
	    {
	      ggui_files_show (jmfile_get_parent(shown_file));
	      return TRUE;
	    }

	  /* Double clicks on files and channels activate them.  We
	     don't do anything for directories because that might be
	     annoying. */
	  if (selected_file && selected_file->type != JMFILE_DIR)
	    backend_activate(selected_file);

	  if (selected_file)
	    { 
	      /* Double clicks on channels also show them */
	      if (selected_file->type == JMFILE_CHANNEL)
		{
		  ggui_channels_select (selected_file);
		  /* WARNING: this changes selected_file as a side
		     effect. */
		}
	      else if (selected_file->type == JMFILE_DIR)
		{
		  ggui_files_show (selected_file);
		}
	    }
	}
      else
	gtk_clist_unselect_all(list);

      rv = TRUE;
    }

  /* User right-clicked - show context menu */
  else if (event->type == GDK_BUTTON_PRESS && event->button == 3)
    {
      GList* i;
      gboolean show_menu;
      JMFile* selected_channel;

      /* Select the clicked row, if there is one */
      if (gtk_clist_get_selection_info (list, event->x, event->y, &row, &column))
	gtk_clist_select_row (list, row, 0);
      else
	gtk_clist_unselect_all(list);

      /* Get the selected channel */
      selected_channel = ggui_channels_get_selected();

      /* Hide everything */
      for (i = GTK_MENU_SHELL(files_popup)->children; i != NULL; i = i->next)
	{
	  GtkWidget* menu_item = GTK_WIDGET(i->data);
	  g_assert (GTK_MENU_ITEM(menu_item) != NULL);
	  gtk_widget_hide(menu_item);
	}


      SHOW_MENUITEM("open_channel_mi");
      SHOW_MENUITEM("create_root_channel_mi");

      if (shown_file)
	{
	  g_return_val_if_fail (selected_channel != NULL, FALSE);

	  if (jmfile_on_channel (selected_channel))
	    {
	      if (!selected_channel->is_channels_only)
		{
		  SHOW_MENUITEM("add_file_mi");
		  SHOW_MENUITEM("add_url_mi");
		  SHOW_MENUITEM("add_directory_mi");
		  SHOW_MENUITEM("create_chat_mi");
		  SHOW_MENUITEM("create_directory_mi");
		}

	      SHOW_MENUITEM("create_channel_mi");
	      SHOW_MENUITEM("create_channel_from_directory_mi");

	      SHOW_MENUITEM("separator2");

	      if (selected_channel->is_local)
		SHOW_MENUITEM("remove_current_channel_mi");
	      else if (jmfile_on_channel (selected_channel))
		SHOW_MENUITEM("leave_current_channel_mi");
	    }
	}


      /* Build the menu */
      if (selected_file)
	{
	  SHOW_MENUITEM("separator3");
	  SHOW_MENUITEM("properties_mi");

	  g_return_val_if_fail (shown_file != NULL, FALSE);

	  switch (selected_file->type)
	    {
		  
	    case JMFILE_FILE:
	      {
		SHOW_MENUITEM("separator1");

		if (selected_file->is_local)
		  SHOW_MENUITEM("remove_file_mi");
		else if (selected_file->status == JMFILE_STATUS_TRANSFERRING)
		  SHOW_MENUITEM("cancel_download_mi");
		else if (!selected_file->have_file)
		  SHOW_MENUITEM("get_file_mi");

		break;
	      }

	    case JMFILE_DIR:
	      {
		SHOW_MENUITEM("separator1");

		if (selected_file->is_local)
		  SHOW_MENUITEM("remove_directory_mi");
		else
		  SHOW_MENUITEM("get_directory_mi");

		break;
	      }

	    case JMFILE_CHANNEL:
	      {
		SHOW_MENUITEM("separator1");

		if (selected_file->is_local)
		  SHOW_MENUITEM("remove_channel_mi");
		else if (btsmp_is_up(selected_file->btsmp))
		  SHOW_MENUITEM("leave_channel_mi");
		else if (selected_file->status != JMFILE_STATUS_CONNECTING)
		  SHOW_MENUITEM("join_channel_mi");

		break;
	      }

	    case JMFILE_URL:
	      {
		SHOW_MENUITEM("separator1");

		if (!strncmp(selected_file->url, "jmchat", 6))
		  {
		    if (selected_file->is_local)
		      SHOW_MENUITEM("remove_chat_mi");
		    else if (ggui_chat_on (selected_file))
		      SHOW_MENUITEM("leave_chat_mi");
		    else
		      SHOW_MENUITEM("join_chat_mi");
		  }

		else
		  {
		    if (selected_file->is_local)
		      SHOW_MENUITEM("remove_url_mi");
		    SHOW_MENUITEM("open_url_mi");
		  }

		break;
	      }


	    case JMFILE_SEARCH:
	    default:
	      break;
	    }
	}

      else
	{
	  SHOW_MENUITEM("separator3");
	  SHOW_MENUITEM("exit_mi");
	}

      if (show_menu)
	gtk_menu_popup (GTK_MENU(files_popup), NULL, NULL, NULL, NULL, 3, event->time);

      rv = TRUE;
    }

  return rv;
}


/* ************************************************************ */


gboolean
on_files_drag_motion (GtkCList* list, GdkDragContext* context,
		      gint x, gint y, guint time)
{
  JMFile* selected_channel;

  selected_channel = ggui_channels_get_selected();

  /* Get the selection information */
  if (shown_file && selected_channel &&
      (selected_channel->is_local || !selected_channel->is_channels_only))
    {
      gdk_drag_status (context, context->suggested_action, time);
    }
    
  return TRUE;
}


void  
on_files_drag_data_received (GtkCList* list, GdkDragContext* context,
			     gint x, gint y, GtkSelectionData* sel_data,
			     guint info, guint time, gpointer user_data)
{
  /* Check if the data is valid */
  if ((sel_data->length >= 0) && (sel_data->format == 8))
    {
      JMFile* selected_channel;

      selected_channel = ggui_channels_get_selected();

      if (shown_file && selected_channel)
	{
	  if (!selected_channel->is_local && selected_channel->is_channels_only)
	    {
	      ggui_show_message ("warning", _("This channel allows channels only"));
	      goto error;
	    }

	  switch (info)
	    {
	    case DND_TARGET_STRING:
	    case DND_TARGET_URI_LIST:
	      {
		if (!backend_files_create(sel_data->data, shown_file, FALSE))
		  {
		    gtk_drag_finish (context, TRUE, TRUE, time);
		    return;
		  }
	      }
	    }
	}
    }

 error:
  gtk_drag_finish (context, FALSE, TRUE, time);

  return;
}


/* ************************************************************ */


void
on_files_get_file (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_get_file (selected_file);
}

void
on_files_cancel_download (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_cancel_download (selected_file);
}

void
on_files_remove_file (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_remove_file (selected_file);
}

/* get_directory    ~ get_file */

/* remove_directory ~ remove_file */

void
on_files_open_url (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_open_url (selected_file);
}

void
on_files_remove_url (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_remove_url (selected_file);
}

void
on_files_join_channel (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_join_channel (selected_file);
}

void
on_files_remove_channel (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_remove_channel (selected_file);
}

void
on_files_leave_channel (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_leave_channel (selected_file);
}

void
on_files_join_chat (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_join_chat (selected_file);
}

void
on_files_remove_chat (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_remove_chat (selected_file);
}

void
on_files_leave_chat (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_leave_chat (selected_file);
}


/* ********** */

void
on_files_add_file (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_add_file (&shown_file);
}

void
on_files_add_directory (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_add_directory (&shown_file);
}

void
on_files_add_url (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_add_url (&shown_file);
}

void
on_files_create_chat (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_create_chat (&shown_file);
}


void
on_files_create_directory (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_create_directory (&shown_file);
}

void
on_files_create_channel (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_create_channel (&shown_file);
}

void
on_files_create_channel_from_directory (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_create_channel_from_directory (&shown_file);
}


/* ********** */

void
on_files_open_channel(gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_open_channel ();
}

void
on_files_create_root_channel (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_create_root_channel();
}


/* ********** */


void
on_files_remove_current_channel (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  JMFile* selected_channel;

  selected_channel = ggui_channels_get_selected();
  g_return_if_fail (selected_channel != NULL);

  ggui_menu_remove_channel (selected_channel);
}


void
on_files_leave_current_channel (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  JMFile* selected_channel;

  selected_channel = ggui_channels_get_selected();
  g_return_if_fail (selected_channel != NULL);

  ggui_menu_leave_channel (selected_channel);
}


/* ********** */

void
on_files_properties (gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_properties (selected_file);
}

