/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "ggui.h"
#include "ggui_menu.h"
#include "ggui_search.h"
#include "util.h"
#include "jmsearch.h"


static GladeXML* 	menu_xml = NULL;

static GtkEntry* 	search_entry = NULL;
static GtkCList* 	search_clist = NULL;
static GtkOptionMenu* 	search_menu = NULL;
static GtkWidget*	search_popup = NULL;

static GList* 		search_servers = NULL;
static gpointer 	search_id = NULL;

static GList*		search_results = NULL;

static JMFile*		selected_result = NULL;



static void	ggui_search_update_servers(void);

void 		on_search_select_row(GtkCList* list, gint row, gint column, GdkEvent* event);
void 		on_search_unselect_row(GtkCList* list, gint row, gint column, GdkEvent* event);
gint		on_search_button_press (GtkCList* list, GdkEventButton *event, gpointer user_data);

void 		on_search_get_file(gpointer callback_data, guint callback_action, GtkWidget* widget);
void 		on_search_cancel_download(gpointer callback_data, guint callback_action, GtkWidget* widget);
void 		on_search_properties(gpointer callback_data, guint callback_action, GtkWidget* widget);

void 		on_search_entry_activate (GtkEntry* entry, gpointer user_data);

static void 	jmsearch_cb(JMSearchStatus status, JMGURL* url, gchar* str, GList* results);





void
ggui_search_init(void)
{
  menu_xml = ggui_get_jm_glade_xml("search_popup", "master_popup.glade");
  g_assert (menu_xml != NULL);
  glade_xml_signal_autoconnect(menu_xml);

  search_entry = 	GTK_ENTRY(glade_xml_get_widget(jmw_xml, "jmsearchentry"));
  search_clist = 	GTK_CLIST(glade_xml_get_widget(jmw_xml, "jmsearchclist"));
  search_menu = 	GTK_OPTION_MENU(glade_xml_get_widget(jmw_xml, "jmsearchservers"));
  search_popup = 	GTK_WIDGET(glade_xml_get_widget(menu_xml, "search_popup"));

  g_assert (search_entry   != NULL);
  g_assert (search_clist   != NULL);
  g_assert (search_menu != NULL);
  g_assert (search_popup    != NULL);

  gtk_clist_column_title_passive (search_clist, 0);
  gtk_clist_column_title_passive (search_clist, 1);

  ggui_set_tooltip (GTK_WIDGET(search_entry), 
		    _("Seach query.  "
		      "Enter a string to search for and press return.  "  
		      "Regular expressions are _not_ supported yet."));
  ggui_set_tooltip (GTK_WIDGET(search_clist), 
		    _("Results from the last search."
		      "Double click to download."));
  ggui_set_tooltip (GTK_WIDGET(search_menu), 
		    _("Search servers available.  "
		      "Choose a server before searching.  "
		      "Try to choose a server near you.  "
		      "You need to be on some channel to see a search server."));
}


void
ggui_search_add(JMFile* jmfile)
{
  g_return_if_fail (jmfile != NULL);

  if (jmfile->type == JMFILE_SEARCH)
    {
      search_servers = g_list_append(search_servers, jmfile);
      ggui_search_update_servers();
    }
}


void
ggui_search_remove(JMFile* jmfile)
{
  g_return_if_fail (jmfile != NULL);

  if (jmfile->type == JMFILE_SEARCH)
    {
      search_servers = g_list_remove (search_servers, jmfile);
      ggui_search_update_servers ();
    }

  else if (jmfile->type == JMFILE_FILE)
    {
      int row;

      row = gtk_clist_find_row_from_data (search_clist, jmfile);
      if (row >= 0)
	{
	  search_results = g_list_remove (search_results, jmfile);
	  gtk_clist_remove (search_clist, row);
	}
    }
}


void
ggui_search_update (JMFile* jmfile)
{
  int row;
  gchar* length;
  GdkPixmap* pixmap;
  GdkBitmap* mask;

  if (jmfile->type != JMFILE_FILE)
    return;

  row = gtk_clist_find_row_from_data (search_clist, jmfile);
  if (row < 0)
    return;

  if (jmfile->length)
    length = filesize_to_string (jmfile->length);

  ggui_get_pixmaps (jmfile, &pixmap, &mask);

  ggui_clist_update_row2 (search_clist, row,
			  jmfile->name, length,
			  pixmap, mask);

  g_free (length);
}


void
ggui_search_freeze(void)
{
  gtk_clist_freeze (search_clist);
}


void
ggui_search_thaw(void)
{
  gtk_clist_thaw (search_clist);
}


void
ggui_search_update_servers(void)
{
  GList* i;
  GtkMenu* menu;

  /* Remove the old menu */
  gtk_option_menu_remove_menu(search_menu);

  /* Build the new menu */
  menu = GTK_MENU(gtk_menu_new());
  for (i = search_servers; i != NULL; i = i->next)
    {
      JMFile* jmfile = (JMFile*) i->data;
      GtkWidget* item;

      item = gtk_menu_item_new_with_label(jmfile->name);
      gtk_object_set_user_data(GTK_OBJECT(item), jmfile);
      gtk_menu_append(menu, item);
      gtk_widget_show(item);
    }
  gtk_widget_show(GTK_WIDGET(menu));

  /* Set the new menu */
  gtk_option_menu_set_menu(search_menu, GTK_WIDGET(menu));
}



/* **************************************** */


void 
on_search_select_row(GtkCList* list, gint row, gint column, GdkEvent* event)
{
  selected_result = (JMFile*) gtk_clist_get_row_data(list, row);
}


void 
on_search_unselect_row(GtkCList* list, gint row, gint column, GdkEvent* event)
{
  selected_result = NULL;
}



#define HIDE_MENUITEM(MI) do {       					\
  GtkWidget* menu_item = glade_xml_get_widget(menu_xml, (MI));		\
  g_return_val_if_fail (menu_item != NULL, FALSE);			\
  gtk_widget_hide(menu_item);} while (0)

#define SHOW_MENUITEM(MI) do {       					\
  GtkWidget* menu_item = glade_xml_get_widget(menu_xml, (MI));		\
  g_return_val_if_fail (menu_item != NULL, FALSE);			\
  do_show = TRUE;							\
  gtk_widget_show(menu_item);} while (0)



/* Return TRUE if we handle event; FALSE to pass on */
gint
on_search_button_press (GtkCList* list, GdkEventButton *event, gpointer user_data)
{
  gint row, column;
  gboolean rv = FALSE;


  /* User double clicked - activate the thing selected */
  if (event->type == GDK_2BUTTON_PRESS && 
      (event->button == 1 || event->button == 2))
    {
      /* Get the selection information */
      if (gtk_clist_get_selection_info(list, event->x, event->y, &row, &column))
	{
	  /* Select the clicked row, if there is one */
	  gtk_clist_select_row (list, row, 0);
	
	  /* Double clicks on files and channels activate them.  We
	     don't do anything for directories because that might be
	     annoying. */
	  if (selected_result && 
	      (selected_result->type == JMFILE_CHANNEL ||
	       selected_result->type == JMFILE_FILE))
	    backend_activate(selected_result);
	}
      else
	gtk_clist_unselect_all(list);

      rv = TRUE;
    }

  /* User right-clicked - show context menu */
  else if (event->type == GDK_BUTTON_PRESS && event->button == 3)
    {
      GList* i;
      gboolean do_show = FALSE;

      /* Select the clicked row, if there is one */
      if (gtk_clist_get_selection_info (list, event->x, event->y, &row, &column))
	gtk_clist_select_row (list, row, 0);
      else
	gtk_clist_unselect_all(list);

      /* Hide everything */
      for (i = GTK_MENU_SHELL(search_popup)->children; i != NULL; i = i->next)
	{
	  GtkWidget* menu_item = GTK_WIDGET(i->data);
	  g_assert (GTK_MENU_ITEM(menu_item) != NULL);
	  gtk_widget_hide(menu_item);
	}

      /* Build the menu */
      if (selected_result)
	{
	  if (!selected_result->is_local &&
	      !selected_result->mtp_client)
	    SHOW_MENUITEM("search_get_file_mi");

	  else if (selected_result->status == JMFILE_STATUS_TRANSFERRING)
	    SHOW_MENUITEM("search_cancel_download_mi");

	  SHOW_MENUITEM("separator1");
	  SHOW_MENUITEM("properties_mi");
	}

      if (do_show)
	gtk_menu_popup (GTK_MENU(search_popup), NULL, NULL, NULL, NULL, 3, event->time);

      rv = TRUE;
    }

  return rv;
}



void
on_search_get_file(gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_get_file(selected_result);
}


void
on_search_cancel_download(gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_cancel_download(selected_result);
}


void
on_search_properties(gpointer callback_data, guint callback_action, GtkWidget* widget)
{
  ggui_menu_properties(selected_result);
}


void
on_search_entry_activate (GtkEntry* entry, gpointer user_data)
{
  GtkMenu* menu;
  GtkWidget* item;
  JMFile* server = NULL;
  gchar* word;
  GList* i;
  JMGURL* url;

  menu = GTK_MENU(gtk_option_menu_get_menu(search_menu));
  g_assert (menu != NULL);

  /* Get search server */
  item = gtk_menu_get_active(menu);
  if (item == NULL)
    {
      ggui_show_message ("warning", _("Select a search server first\n"));
      return;
    }

  server = (JMFile*) gtk_object_get_user_data(GTK_OBJECT(item));
  g_return_if_fail (server);
  
  /* Get search server URL */
  url = jm_url_new (server->url);
  g_return_if_fail (url);

  /* Get search query */
  word = gtk_entry_get_text(search_entry);
  if (word == NULL || word[0] == '\0')
    {
      ggui_show_message ("warning", _("Enter some text to search for first\n"));
      return;
    }

  /* Unlink old search results */
  for (i = search_results; i != NULL; i = i->next)
    {
      JMFile* result = (JMFile*) i->data;

      backend_unlink (result, NULL);
    }
  g_list_free (search_results);
  search_results = NULL;

  /* Clear list of old search results */
  gtk_clist_clear(search_clist);

  /* Cancel any pending seach */
  if (search_id)
    jmsearch_cancel(search_id);

  search_id = jmsearch_search(url, word, jmsearch_cb);

  jm_url_delete (url);
}


static void
jmsearch_cb(JMSearchStatus status, JMGURL* url, gchar* str, /* const */ GList* results)
{
  GList* i;

  search_id = NULL;

  if (status != JMSEARCH_OK)
    {
      gchar* texts[3] = {"Search error", "", NULL};
      gtk_clist_insert (search_clist, 0, texts);
      return;
    }

  if (results == NULL)
    {
      gchar* text;
      gchar* texts[2] = {NULL, NULL};

      text = g_strdup_printf("No matches for \"%s\" found.", str);
      texts[0] = text;

      gtk_clist_insert (search_clist, 0, texts);

      g_free (text);

      return;
    }

  gui_freeze();

  for (i = results; i != NULL; i = i->next)
    {
      JMSearchResult* result = (JMSearchResult*) i->data;
      gchar* url_str;

      g_return_if_fail (result->name);
      g_return_if_fail (result->url);

      url_str = jm_url_get_nice_string (result->url);
      if (url_str)
	{
	  JMFile* jmfile;

	  jmfile = backend_file_add (result->name, url_str, NULL);
	  if (jmfile)
	    {
	      gchar* texts[2] = {NULL, NULL};
	      gint row;

	      jmfile->length = result->length;

	      row = gtk_clist_insert (search_clist, 0, texts);
	      gtk_clist_set_row_data(search_clist, row, jmfile);

	      ggui_search_update (jmfile);
	    }

	  g_free (url_str);
	}
    }

  gui_thaw();
}
