/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "jmchat.h"
#include "jmp.h"
#include "globals.h"
#include "util.h"

static void	send_join (JMChat* chat);
static void 	send_leave (JMChat* chat);

static gboolean	chat_packet_cb (BTSmp* btsmp, guint8* buffer, guint length, gpointer user_data);
static void     chat_error_cb (BTSmp* btsmp, gpointer user_data);



JMChat* 
jmchat_create (gchar* name, gchar* nick)
{
  JMChat* chat;
  BTSmp* btsmp;
  gchar* url;

  g_return_val_if_fail (name, NULL);
  g_return_val_if_fail (nick, NULL);

  /* Create a new BTSMP session */
  btsmp = btsmp_create (name);
  if (btsmp == NULL)
    return NULL;

  /* Create the URL */
  url = make_url (btsmp->url, "jmchat", "btsmp", 0);
  g_return_val_if_fail (url, NULL);

  /* Store only at most 4k or 3 minutes of chat */
  btsmp_cache_set_policy (btsmp, 4096, 180000);
  
  /* Create a new JMChat */
  chat = g_new0 (JMChat, 1);
  chat->url = url;
  chat->nick = g_strdup (nick);
  chat->btsmp = btsmp;
  chat->is_local = TRUE;

  /* Initialize BTSMP */
  btsmp->packet_func = chat_packet_cb;
  btsmp->error_func  = chat_error_cb;
  btsmp->packet_user_data = chat;
  btsmp->error_user_data  = chat;

  /* Send a JOIN */
  send_join (chat);

  return chat;
}


void
jmchat_delete (JMChat* chat)
{
  g_return_if_fail (chat);

  g_free (chat->url);
  g_free (chat->nick);

  btsmp_delete (chat->btsmp);

  memset (chat, 0, sizeof(*chat));
  g_free (chat);
}


JMChat*
jmchat_join (gchar* url, gchar* nick)
{
  JMChat* chat;
  BTSmp* btsmp;

  g_return_val_if_fail (url, NULL);
  g_return_val_if_fail (nick, NULL);

  /* Join the BTSMP */
  btsmp = btsmp_join (url);
  if (!btsmp)
    return NULL;

  /* Don't cache anything */
  btsmp_cache_set_policy (btsmp, 0, 0);

  /* Create a new JMChat */
  chat = g_new0 (JMChat, 1);
  chat->url = g_strdup (url);
  chat->nick = g_strdup (nick);
  chat->btsmp = btsmp;

  /* Initialize BTSMP */
  btsmp->packet_func = chat_packet_cb;
  btsmp->error_func  = chat_error_cb;
  btsmp->packet_user_data = chat;
  btsmp->error_user_data  = chat;

  /* Send a JOIN */
  send_join (chat);

  return chat;
}



gboolean
jmchat_rejoin (JMChat* chat)
{
  BTSmp* btsmp;

  g_return_val_if_fail (chat, TRUE);
  g_return_val_if_fail (!jmchat_is_up(chat), TRUE);

  /* Join the BTSMP */
  btsmp = btsmp_join (chat->url);
  if (!btsmp)
    return TRUE;
  
  chat->btsmp = btsmp;

  /* Don't cache anything */
  btsmp_cache_set_policy (btsmp, 0, 0);

  /* Initialize BTSMP */
  btsmp->packet_func = chat_packet_cb;
  btsmp->error_func  = chat_error_cb;
  btsmp->packet_user_data = chat;
  btsmp->error_user_data  = chat;

  /* We aren't down any more */
  chat->is_down = FALSE;

  /* Send a JOIN */
  send_join (chat);

  return FALSE;
}



void
jmchat_leave (JMChat* chat)
{
  g_return_if_fail (chat);
  
  /* Close the BTSMP */
  if (chat->btsmp)
    {
      /* Send a LEAVE */
      send_leave (chat);

      /* Delete the BTSMP */
      btsmp_delete (chat->btsmp);
      chat->btsmp = NULL;
    }
}


void
jmchat_say (JMChat* chat, gchar* message)
{
  JMPAnnouncement* a;
  gchar* buffer;
  guint length;

  g_return_if_fail (chat);
  g_return_if_fail (message);

  if (!chat->btsmp)
    return;
  
  /* Create a SAY */
  a = jmp_announcement_new ("say");
  jmp_announcement_set_attribute (a, "name", chat->nick);
  jmp_announcement_set_attribute (a, "message", message);
  jmp_write (a, &buffer, &length);
  g_return_if_fail (buffer);

  /* Send the SAY */
  btsmp_send_packet (chat->btsmp, buffer, length);
  g_free (buffer);
  jmp_announcement_delete (a);
}


gboolean
jmchat_is_up (JMChat* chat)
{
  g_return_val_if_fail (chat, FALSE);

  if (!chat->btsmp)
    return FALSE;
  
  return TRUE;
}


static void
send_join (JMChat* chat)
{
  JMPAnnouncement* a;
  gchar* buffer;
  guint length;

  g_return_if_fail (chat);
  g_return_if_fail (chat->btsmp);
  
  /* Create a JOIN */
  a = jmp_announcement_new ("join");
  jmp_announcement_set_attribute (a, "name", chat->nick);
  jmp_write (a, &buffer, &length);
  g_return_if_fail (buffer);

  /* Send the JOIN */
  btsmp_send_packet (chat->btsmp, buffer, length);
  g_free (buffer);
  jmp_announcement_delete (a);
}


static void
send_leave (JMChat* chat)
{
  JMPAnnouncement* a;
  gchar* buffer;
  guint length;

  DEBUGP ("send_leave\n");

  g_return_if_fail (chat);
  g_return_if_fail (chat->btsmp);
  
  /* Create a LEAVE */
  a = jmp_announcement_new ("leave");
  jmp_announcement_set_attribute (a, "name", chat->nick);
  jmp_write (a, &buffer, &length);
  g_return_if_fail (buffer);

  /* Send the LEAVE */
  btsmp_send_packet (chat->btsmp, buffer, length);
  g_free (buffer);
  jmp_announcement_delete (a);
}



/* ************************************************************ */

static gboolean
chat_packet_cb (BTSmp* btsmp, guint8* buffer, guint length, gpointer user_data)
{
  JMChat* chat = (JMChat*) user_data;
  GList* announcements;
  GList* i;

  DEBUGP("chat_packet_cb: %s\n", chat->url);

  if (!chat->info_func)
    return FALSE;

  announcements = jmp_read (buffer, length);
  for (i = announcements; i != NULL; i = i->next)
    {
      JMPAnnouncement* a = (JMPAnnouncement*) i->data;
      gchar* nick;

      nick = jmp_announcement_get_attribute (a, "name");
      if (!nick) continue;

      if (!strcmp(a->name, "join"))
	{
	  chat->info_func (chat, JMCHAT_JOIN, nick, NULL, chat->user_data);
	}
      else if (!strcmp(a->name, "leave"))
	{
	  chat->info_func (chat, JMCHAT_LEAVE, nick, NULL, chat->user_data);
	}
      else if (!strcmp(a->name, "say"))
	{
	  gchar* msg;

	  msg = jmp_announcement_get_attribute (a, "message");
	  if (!msg) continue;

	  chat->info_func (chat, JMCHAT_SAY, nick, msg, chat->user_data);
	}
    }

  jmp_announcement_delete_list(announcements);

  return FALSE;
}


static void
chat_error_cb (BTSmp* btsmp, gpointer user_data)
{
  JMChat* chat = (JMChat*) user_data;

  DEBUGP ("chat_error_cb: %s\n", chat->url);

  btsmp_delete (btsmp);
  chat->btsmp = NULL;

  chat->is_down = TRUE;
  
  if (chat->error_func)
    chat->error_func (chat, chat->user_data);
}
