/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <errno.h>
#include <string.h>

#include "gui.h"
#include "jmfile.h"
#include "backend.h"
#include "globals.h"




JMFile* 
jmfile_new (const gchar* name, JMFileType type, const gchar* url)
{
  JMFile* jmfile;

  /* Figure out the name from the URL if the name isn't given */
  if (name == NULL && url != NULL)
    {
      name = strrchr(url, '/');
      if (name)
	name = &name[1];
      else
	name = url;
    }

  jmfile = g_new0(JMFile, 1);
  if (name) jmfile->name = g_strdup(name);
  jmfile->type = type;
  if (url) jmfile->url = g_strdup(url);
  jmfile->status = JMFILE_STATUS_NONE;

  return jmfile;
}



void
jmfile_delete (JMFile* file)
{
  GList* i;
  GSList* si;

  g_return_if_fail (file);

  DEBUGP ("jmfile_delete %s\n", file->url);

  /* Remove myself from my parents */
  for (i = file->parents; i != NULL; i = i->next)
    {
      JMFile* parent = (JMFile*) i->data;

      g_return_if_fail (parent);
      parent->children = g_list_remove (parent->children, file);
    }
  g_list_free (file->parents);
  file->parents = NULL;

  /* Remove myself from my children */
  for (i = file->children; i != NULL; i = i->next)
    {
      JMFile* child = (JMFile*) i->data;

      g_return_if_fail (child);
      child->parents = g_list_remove (child->parents, file);
    }
  g_list_free (file->children);
  file->children = NULL;

  if (file->mtp_client)
    mtp_client_delete (file->mtp_client);

  if (file->mtp_mirror)
    mtp_unmirror (file->mtp_mirror);

  if (file->sha_mirror)
    mtp_unmirror (file->sha_mirror);

  if (file->btsmp)
    btsmp_delete (file->btsmp);

  if (file->update_timer != 0)
    g_assert (g_source_remove(file->update_timer));

  for (si = file->announcements; si != NULL; si = si->next)
    {
      JMAnn* jmann = (JMAnn*) si->data;

      g_free (jmann->fullname);	
      g_free (jmann);

    }
  g_slist_free (file->announcements);
  file->announcements = NULL;

  g_slist_free (file->announcers);
  file->announcers = NULL;

  /* Delete stuff */
  g_free (file->name);
  g_free (file->url);
  g_free (file->path);
  g_free (file->dirpath);
  gnet_sha_delete (file->sha);
  if (file->sha_id)
    gnet_sha_new_file_async_cancel (file->sha_id);

  if (file->buffer)
    {
      g_return_if_fail (file->file);

      if (munmap(file->buffer, file->length) != 0)
	g_warning ("munmap failure: %s\n", g_strerror(errno));

      fclose (file->file);
    }

  if (file->munmap_timer)
    g_source_remove (file->munmap_timer);


  /* Zero out myself for safety */
  memset (file, 0, sizeof(*file));

  g_free (file);

  DEBUGP ("jmfile_delete done\n");
}


gint
jmfile_compare (gconstpointer a, gconstpointer b)
{
  gint rv = strcmp(((JMFile*) a)->name, ((JMFile*) b)->name);

  if (rv < 0)
    return 1;
  
  return 0;
}



/* ************************************************************ */


gboolean
jmfile_has_parent (JMFile* file, JMFile* parent)
{
  g_return_val_if_fail (file, FALSE);

  if (parent == NULL && file->num_null_parents)
    return TRUE;

  return (g_list_find (file->parents, parent) != NULL);
}


gboolean
jmfile_is_orphan (JMFile* file)
{
  g_return_val_if_fail (file, FALSE);

  if (file->parents == NULL && file->num_null_parents)
    return TRUE;

  return FALSE;
}


void
jmfile_link (JMFile* file, JMFile* parent)
{
  g_return_if_fail (file);

  /* Don't link directories to multipe parents or NULL */
  if (file->type == JMFILE_DIR && (file->parents || !parent))
    {
      g_warning ("Linking directory twice or to NULL.");
      return;
    }

  if (parent != NULL)
    {
      if (!jmfile_has_parent (file, parent))
	file->parents = g_list_prepend (file->parents, parent);

      if (!jmfile_has_child (parent, file))
	parent->children = g_list_prepend (parent->children, file);
    }
  else
    ++file->num_null_parents;
}


/* return TRUE if the JMFile is still valid */
gboolean
jmfile_unlink (JMFile* file, JMFile* parent)
{
  GList* i;

  g_return_val_if_fail (file, FALSE);

  file->parents = g_list_remove (file->parents, parent);

  if (parent)
    parent->children = g_list_remove (parent->children, file);
  else
    --file->num_null_parents;

  /* Unlink each child */
  for (i = file->children; i != NULL; )
    {
      JMFile* child = (JMFile*) i->data;
      i = i->next;
      
      if (child != NULL)
	jmfile_unlink (child, file);
    }

  /* If I have no parents, delete me */
  if (file->parents == NULL && !file->num_null_parents)
    {
      gui_remove (file);
      backend_delete (file);
      jmfile_delete (file);
      return FALSE;
    }

  return TRUE;
}


JMFile*		
jmfile_get_parent (JMFile* file)
{
  g_return_val_if_fail (file, NULL);

  if (!file->parents)
    return NULL;

  return (JMFile*) file->parents->data;
}



gboolean
jmfile_has_child (JMFile* file, JMFile* child)
{
  g_return_val_if_fail (file, FALSE);

  return (g_list_find (file->children, child) != NULL);
}



JMFile* 
jmfile_get_child_by_name (JMFile* file, gchar* name)
{
  GList* i;
  JMFile* child;

  if (!file)
    return NULL;

  for (i = file->children; i != NULL; i = i->next)
    {
      child = (JMFile*) i->data;

      if (strcmp(child->name, name) == 0)
	{
	  return child;
	}
    }
  
  return NULL;
}



JMFile*
jmfile_get_channel (JMFile* file, JMFile* parent)
{
  g_return_val_if_fail (file, NULL);

  while (parent != NULL && parent->type != JMFILE_CHANNEL)
    {
      g_return_val_if_fail (parent->parents, NULL);

      parent = (JMFile*) parent->parents->data;
    }

  return parent;
}



gchar*
jmfile_make_fullname (JMFile* parent, gchar* name)
{
  gchar* temp;
  gchar* rv;

  g_return_val_if_fail (name, NULL);

  if (parent == NULL)
    return g_strconcat ("/", name, NULL);

  g_return_val_if_fail (parent->name, NULL);
  g_return_val_if_fail (parent->type == JMFILE_CHANNEL || 
			parent->type == JMFILE_DIR, NULL);

  if (parent->type == JMFILE_CHANNEL)
    return g_strconcat ("/", parent->name, "/", name, NULL);

  /* Otherwise the parent is a directory, recurse */
  g_return_val_if_fail (parent->parents, NULL);

  temp = jmfile_make_fullname((JMFile*) parent->parents->data, parent->name);
  rv = g_strconcat (temp, "/", name, NULL);
  g_free (temp);

  return rv;
}





/* ************************************************************ */

gboolean
jmfile_have_file (JMFile* jmfile)
{
  g_return_val_if_fail (jmfile, FALSE);

  if (jmfile->type != JMFILE_FILE)
    return FALSE;

  if (jmfile->is_local)
    return TRUE;

  return jmfile->have_file;
}


gboolean
jmfile_on_channel (JMFile* jmfile)
{
  g_return_val_if_fail (jmfile, FALSE);

  if (jmfile->type != JMFILE_CHANNEL)
    return FALSE;

  if (btsmp_is_up(jmfile->btsmp))
    return TRUE;

  return FALSE;
}


gboolean
jmfile_is_url_proto (JMFile* jmfile, gchar* proto)
{
  g_return_val_if_fail (jmfile, FALSE);
  g_return_val_if_fail (proto, FALSE);

  if (jmfile->type != JMFILE_URL)
    return FALSE;

  g_return_val_if_fail (jmfile->url, FALSE);

  if (!strncmp(jmfile->url, proto, strlen(proto)))
    return TRUE;

  return FALSE;
}


gint
jmfile_get_length(JMFile* jmfile)
{
  g_return_val_if_fail (jmfile, 0);
  g_return_val_if_fail (jmfile->type == JMFILE_FILE, 0);

  if (jmfile->have_file || jmfile->buffer_length)
    return jmfile->buffer_length;

  return jmfile->length;
}


gint
jmfile_get_bytes_received(JMFile* jmfile)
{
  g_return_val_if_fail (jmfile, 0);
  g_return_val_if_fail (jmfile->type == JMFILE_FILE, 0);

  if (jmfile->have_file)
    return jmfile->buffer_length;

  if (jmfile->mtp_client)
    return jmfile->mtp_client->offset;

  return 0;
}



gint
jmfile_get_percent_complete(JMFile* jmfile)
{
  g_return_val_if_fail (jmfile, 0);
  g_return_val_if_fail (jmfile->type == JMFILE_FILE, 0);

  if (jmfile->have_file)
    return 100;

  if (!jmfile->mtp_client || !jmfile->mtp_client->length)
    return 0;

  return (100.0 * (double) jmfile->mtp_client->offset) / ((double) jmfile->mtp_client->length);
}
