/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _JMFILE_H
#define _JMFILE_H


#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#include <glib.h>
#include <gnet/gnet.h>
#include "sha.h"

#include "mtp_types.h"
#include "mtp_client.h"
#include "mtp_mirror.h"
#include "btsmp.h"


typedef enum {
  JMFILE_NONE, 
  JMFILE_CHANNEL, 
  JMFILE_DIR, 
  JMFILE_FILE,
  JMFILE_SEARCH,	/* FIX: Change to service? */
  JMFILE_URL
} JMFileType;

typedef enum {
  JMFILE_STATUS_NONE, 
  JMFILE_STATUS_CONNECTING, 
  JMFILE_STATUS_TRANSFERRING,
  JMFILE_STATUS_DOWN
} JMFileStatus;


typedef struct _JMFile
{
  /* Basic file information */
  gchar* name;			/* Name of file (derived from URL)	*/
  JMFileType type;		/* Type of file				*/
  gchar* url;			/* URL of file				*/
  gboolean is_local;		/* Is the file local (did I create?)	*/
  JMFileStatus status;		/* Status of the file			*/

  guint num_null_parents;	/* Number of NULL parents		*/
  GList* parents;		/* Parent files.  A directory can have  */
				/*   only one parent. 			*/
  GList* children;		/* Children files (for a channel or	*/
				/*   directory) 			*/

  struct timeval last_update;	/* Last time we received an 		*/
				/*   announcement for this file. 	*/

  /* For files only */
  MtpClient* mtp_client;	/* MTP Client 				*/
  MtpLocalMirror* mtp_mirror;	/* MTP Mirror				*/
  MtpLocalMirror* sha_mirror;	/* MTP Mirror of SHA of file		*/
  gchar* path;			/* Path to file 			*/
  gchar* dirpath;		/* Path to directory with file		*/
  FILE* file;			/* Pointer to mmap'ed file		*/
  gchar* buffer;		/* File buffer (mmap'ed)		*/
  guint buffer_length;		/* Length of the buffer			*/
  guint buffer_offset;		/* Offset into buffer 			*/
  guint length;			/* Advertised length of file		*/
  guint munmap_timer;		/* Timer for munmaping the file		*/
  gboolean have_file;		/* Do I have the file?  (The file might
				   be zero length.)			*/
  GSHAAsyncID* sha_id;		/* Async SHA calculation ID		*/
  GSHA* sha;			/* SHA hash				*/

  /* For channels only */
  BTSmp* btsmp;			/* BTSmp		 		*/ 
  guint update_timer;		/* Timer for sending updates		*/
  gboolean is_channels_only;	/* Allow channels in channel only	*/

  /* Announcement management (for backend_announce only) */
  GSList* announcers;		/* Channels this file is announced on	*/
  GSList* announcements;	/* Local channel announcements.		*/
  GSList* next_announcement;	/* Next announcement.			*/


  struct _GuiData* gui_data;	/* gui_data for frontend		*/

} JMFile;


/* 

   JM announcement.  Each announcement announces a single file.
   However, a file may have several announcements.  A channel keeps a
   list of announcements for each of its local files.

*/
typedef struct _JMAnn
{
  gchar* 	fullname;	/* Full announcement name		*/

  JMFile*	file;		/* File this announcement represents	*/
  JMFile*	channel;	/* Channel the file was announced on	*/
  JMFile*	parent;		/* Parent of this file			*/

} JMAnn;





JMFile* 	jmfile_new (const gchar* name, JMFileType type, const gchar* url);
void 		jmfile_delete (JMFile* file);
gint 		jmfile_compare (gconstpointer a, gconstpointer b);

gboolean	jmfile_has_parent (JMFile* file, JMFile* parent);
gboolean	jmfile_is_orphan (JMFile* file);
void		jmfile_link (JMFile* file, JMFile* parent);
gboolean	jmfile_unlink (JMFile* file, JMFile* parent);
JMFile*		jmfile_get_parent (JMFile* file);


gboolean	jmfile_has_child (JMFile* file, JMFile* child);
JMFile* 	jmfile_get_child_by_name (JMFile* file, gchar* name);

JMFile*		jmfile_get_channel (JMFile* file, JMFile* parent);
gchar*		jmfile_make_fullname (JMFile* parent, gchar* name);


gboolean	jmfile_have_file (JMFile* jmfile);
gboolean	jmfile_on_channel (JMFile* jmfile);
gboolean	jmfile_is_url_proto (JMFile* jmfile, gchar* proto);
#define		jmfile_is_down(JMFILE)	((JMFILE)->status == JMFILE_STATUS_DOWN)

gint		jmfile_get_length (JMFile* jmfile);
gint		jmfile_get_bytes_received (JMFile* jmfile);
gint		jmfile_get_percent_complete (JMFile* jmfile);


#endif /* _JMFILE_H */
