/* GNet - Networking library
 * Copyright (C) 2000  David Helder
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include <sys/types.h>
#include <sys/ioctl.h>
#ifdef HAVE_SYS_POLL_H
#include <sys/poll.h>
#endif
#include <sys/socket.h>
#ifdef HAVE_SYS_SOCKIO_H
#include <sys/sockio.h>
#endif
#include <sys/time.h>

#include <sys/utsname.h>
#include <sys/wait.h>
#include <signal.h>

#include <errno.h>
#include <fcntl.h>

#include <netinet/in_systm.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/ip.h>		/* Need for TOS */
#include <arpa/inet.h>

#include <arpa/nameser.h>
#include <resolv.h>
#include <netdb.h>

#include <glib.h>

#ifndef socklen_t
#define socklen_t size_t
#endif

#include "jmgnet.h"


/* Hidden in gnet/gnet-private.h */
GList* gnet_private_inetaddr_list_interfaces(void);

#define JMGNET_SOCKADDR_IN(s) (*((struct sockaddr_in*) &s))

struct _GTcpSocket
{
  gint sockfd;
  struct sockaddr sa;
  guint ref_count;
  GIOChannel* iochannel;
};


struct _GInetAddr
{
  gchar* name;
  struct sockaddr sa;
  guint ref_count;
};


void
jmgnet_tcp_socket_set_tos(GTcpSocket* socket, JMGTOS tos)
{
  int sotos;

  g_return_if_fail (socket != NULL);

  /* Some systems (e.g. OpenBSD) do not have IPTOS_*.  Other systems
     have some of them, but not others.  And some systems have them,
     but with different names (e.g. FreeBSD has IPTOS_MINCOST).  If a
     system does not have a IPTOS, or any of them, then this function
     does nothing.  */
  switch (tos)
    {
#ifdef IPTOS_LOWDELAY
    case JMGTOS_LOWDELAY:	sotos = IPTOS_LOWDELAY;		break;
#endif
#ifdef IPTOS_THROUGHPUT
    case JMGTOS_THROUGHPUT:	sotos = IPTOS_THROUGHPUT;	break;
#endif
#ifdef IPTOS_RELIABILITY
    case JMGTOS_RELIABILITY:	sotos = IPTOS_RELIABILITY;	break;
#endif
#ifdef IPTOS_LOWCOST
    case JMGTOS_LOWCOST:	sotos = IPTOS_LOWCOST;		break;
#else
#ifdef IPTOS_MINCOST	/* Called MINCOST in FreeBSD 4.0 */
    case JMGTOS_LOWCOST:	sotos = IPTOS_MINCOST;		break;
#endif
#endif
    default: return;
    }

#ifdef IP_TOS
  if (setsockopt(socket->sockfd, IPPROTO_IP, IP_TOS, (void*) &sotos, sizeof(sotos)) != 0)
    g_warning ("Can't set TOS on TCP socket\n");
#endif


}



GTcpSocket* 
jm_gnet_tcp_socket_server_new(GInetAddr* iface, const gint port)
{
  GTcpSocket* s;
  struct sockaddr_in* sa_in;
  const int on = 1;
  socklen_t socklen;
  gint flags;


  /* Create socket */
  s = g_new0(GTcpSocket, 1);
  s->ref_count = 1;
  s->sockfd = socket(AF_INET, SOCK_STREAM, 0);
  if (s->sockfd < 0)
    goto error;

  /* Set up address and port for connection */
  sa_in = (struct sockaddr_in*) &s->sa;
  sa_in->sin_family = AF_INET;
  if (iface)
    sa_in->sin_addr.s_addr = JMGNET_SOCKADDR_IN(iface->sa).sin_addr.s_addr;
  else
    sa_in->sin_addr.s_addr = g_htonl(INADDR_ANY);
  sa_in->sin_port = g_htons(port);

  /* Set REUSEADDR so we can reuse the port */
  if (setsockopt(s->sockfd, SOL_SOCKET, SO_REUSEADDR, 
		     (void*) &on, sizeof(on)) != 0)
    g_warning("Can't set reuse on tcp socket\n");

  /* Get the flags (should all be 0?) */
  flags = fcntl(s->sockfd, F_GETFL, 0);
  if (flags == -1)
    goto error;

  /* Make the socket non-blocking */
  if (fcntl(s->sockfd, F_SETFL, flags | O_NONBLOCK) == -1)
    goto error;

  /* Bind */
  if (bind(s->sockfd, &s->sa, sizeof(s->sa)) != 0)
    goto error;

  /* Get the socket name - don't care if it fails */
  socklen = sizeof(s->sa);
  if (getsockname(s->sockfd, &s->sa, &socklen) != 0)
    goto error;

  /* Listen */
  if (listen(s->sockfd, 10) != 0)
    goto error;

  return s;

 error:
  if (s)    		g_free(s);
  return NULL;
}


/* ************************************************************ */

GList* 
jm_gnet_inetaddr_list_interfaces(void)
{
  GList* list = NULL;
  gint len, lastlen;
  gchar* buf;
  gchar* ptr;
  gint sockfd;
  struct ifconf ifc;


  /* Create a dummy socket */
  sockfd = socket(AF_INET, SOCK_DGRAM, 0);
  if (sockfd == -1) return NULL;

  len = 8 * sizeof(struct ifreq);
  lastlen = 0;

  /* Get the list of interfaces.  We might have to try multiple times
     if there are a lot of interfaces. */
  while(1)
    {
      buf = g_new0(gchar, len);
      
      ifc.ifc_len = len;
      ifc.ifc_buf = buf;
      if (ioctl(sockfd, SIOCGIFCONF, &ifc) < 0)
	{
	  /* Might have failed because our buffer was too small */
	  if (errno != EINVAL || lastlen != 0)
	    {
	      g_free(buf);
	      return NULL;
	    }
	}
      else
	{
	  /* Break if we got all the interfaces */
	  if (ifc.ifc_len == lastlen)
	    break;

	  lastlen = ifc.ifc_len;
	}

      /* Did not allocate big enough buffer - try again */
      len += 8 * sizeof(struct ifreq);
      g_free(buf);
    }


  /* Create the list.  Stevens has a much more complex way of doing
     this, but his is probably much more correct portable.  */
  for (ptr = buf; ptr < (buf + ifc.ifc_len); )
    {
      struct ifreq* ifr = (struct ifreq*) ptr;
      struct sockaddr addr;
      GInetAddr* ia;

#ifdef HAVE_SOCKADDR_SA_LEN      
      ptr += sizeof(ifr->ifr_name) + ifr->ifr_addr.sa_len;
#else
      ptr += sizeof(struct ifreq);
#endif

      /* Ignore non-AF_INET */
      if (ifr->ifr_addr.sa_family != AF_INET)
	continue;

      /* FIX: Skip colons in name?  Can happen if aliases, maybe. */

      /* Save the address - the next call will clobber it */
      memcpy(&addr, &ifr->ifr_addr, sizeof(addr));
      
      /* Get the flags */
      ioctl(sockfd, SIOCGIFFLAGS, ifr);

      /* Ignore entries that aren't up or loopback.  Someday we'll
	 write an interface structure and include this stuff. */
      if (!(ifr->ifr_flags & IFF_UP) ||
	  (ifr->ifr_flags & IFF_LOOPBACK))
	continue;

      /* Create an InetAddr for this one and add it to our list */
      ia = g_new0 (GInetAddr, 1);
      ia->ref_count = 1;
      memcpy(&ia->sa, &addr, sizeof(addr));
      list = g_list_prepend (list, ia);
    }

  list = g_list_reverse (list);

  return list;
}


GInetAddr* 
jm_gnet_inetaddr_new_nonblock (const gchar* name, const gint port)
{
  struct sockaddr_in* sa_in;
  struct in_addr inaddr;
  GInetAddr* ia = NULL;

  g_return_val_if_fail (name, NULL);

  /* Try to read the name as if were dotted decimal */
 try_again:
  if (inet_aton(name, &inaddr) != 0)
    {
      ia = g_new0(GInetAddr, 1);

      ia->ref_count = 1;
      sa_in = (struct sockaddr_in*) &ia->sa;
      sa_in->sin_family = AF_INET;
      sa_in->sin_port = g_htons(port);
      memcpy(&sa_in->sin_addr, (char*) &inaddr, sizeof(struct in_addr));
    }

  /* If the name is localhost, change it to 127.0.0.1 and try again */
  /* FIX: I don't think this is legal... */
  else if (!strcmp (name, "localhost"))
    {
      name = "127.0.0.1";
      goto try_again;
    }

  return ia;
}


gboolean
jm_gnet_inetaddr_is_canonical (const gchar* name)
{
  struct in_addr inaddr;

  g_return_val_if_fail (name, FALSE);

  return (inet_aton(name, &inaddr) != 0);
}



/* ************************************************************ */

GInetAddr* 
jm_gnet_inetaddr_autodetect_interface (void)
{
  GInetAddr* jm_addr;
  GInetAddr* interface;

  /* First try to get the interface with a route to
     junglemonkey.net (141.213.11.1).  This uses the connected UDP
     socket method described by Stevens.  It does not work on all
     systems.  (see Stevens UNPv1 pp 231-3) */
  jm_addr = jm_gnet_inetaddr_new_nonblock ("141.213.11.1", 0);
  g_assert (jm_addr);

  interface = jm_gnet_inetaddr_get_interface_to (jm_addr);
  gnet_inetaddr_delete (jm_addr);

  /* If that didn't work, then try getting an internet interface
     from the list via SIOCGIFCONF. (see Stevens pp 428-) */
  if (!interface)
    interface = jm_gnet_inetaddr_get_internet_interface ();

  return interface;
}



GInetAddr* 
jm_gnet_inetaddr_get_interface_to (GInetAddr* addr)
{
  int sockfd;
  struct sockaddr_in myaddr;
  socklen_t len;
  GInetAddr* interface;

  g_return_val_if_fail (addr, NULL);

  sockfd = socket (AF_INET, SOCK_DGRAM, 0);
  if (sockfd == -1)
    return NULL;

  if (connect (sockfd, &addr->sa, sizeof(addr->sa)) == -1)
    {
      close (sockfd);
      return NULL;
    }

  len = sizeof (myaddr);
  if (getsockname (sockfd, (struct sockaddr*) &myaddr, &len) != 0)
    {
      close (sockfd);
      return NULL;
    }

  interface = g_new0 (GInetAddr, 1);
  interface->ref_count = 1;
  memcpy (&interface->sa, (char*) &myaddr, sizeof (struct sockaddr_in));

  return interface;
}



GInetAddr* 
jm_gnet_inetaddr_get_internet_interface (void)
{
  GInetAddr* interface = NULL;
  GList* interfaces;
  GList* i;

  /* Get a list of interfaces */
  interfaces = jm_gnet_inetaddr_list_interfaces ();
  if (interfaces == NULL)
    return NULL;

  /* Reverse the list.  TODO: We should be doing this in GNet. */
  interfaces = g_list_reverse (interfaces);

  /* Find the first interface that's an internet interface */
  for (i = interfaces; i != NULL; i = i->next)
    {
      GInetAddr* ia;

      ia = (GInetAddr*) i->data;

      if (jm_gnet_inetaddr_is_internet (ia))
	{
	  interface = gnet_inetaddr_clone (ia);
	  break;
	}
    }

  /* If we didn't find one, return the first interface. */
  if (interface == NULL)
    interface = gnet_inetaddr_clone ((GInetAddr*) interfaces->data);

  /* Delete the interface list */
  for (i = interfaces; i != NULL; i = i->next)
    gnet_inetaddr_delete ((GInetAddr*) i->data);
  g_list_free (interfaces);

  return interface;
}



gboolean
jm_gnet_inetaddr_is_internet_domainname (gchar* domainname)
{
  GInetAddr* addr;

  g_return_val_if_fail (domainname, FALSE);

  /* The domainname can't be localhost or localhost.localdomain */
  if (!strcmp(domainname, "localhost") || 
      !strcmp(domainname, "localhost.localdomain"))
    {
      return FALSE;
    }

  /* The domainname must have a dot in it */
  if (!strchr(domainname, '.'))
    {
      return FALSE;
    }
	
  /* If it's dotted decimal, make sure it's valid */
  addr = jm_gnet_inetaddr_new_nonblock (domainname, 0);
  if (addr)
    {
      gboolean rv;

      rv = jm_gnet_inetaddr_is_internet (addr);
      gnet_inetaddr_delete (addr);

      if (!rv)
	return FALSE;
    }

  return TRUE;
}



/**

   Check if the interface is a sensible internet address.  This mean
   it is not private, reserved, loopback, multicast, or broadcast.

   Note that private and loopback address are often valid addresses,
   so this should only be used to check for general internet
   connectivity.

   Returns TRUE if it an 'internet' address, FALSE otherwise.

*/
gboolean
jm_gnet_inetaddr_is_internet (GInetAddr* inetaddr)
{
  g_return_val_if_fail(inetaddr != NULL, FALSE);

  if (!jm_gnet_inetaddr_is_private(inetaddr) 	&&
      !jm_gnet_inetaddr_is_reserved(inetaddr) 	&&
      !jm_gnet_inetaddr_is_loopback(inetaddr) 	&&
      !jm_gnet_inetaddr_is_multicast(inetaddr) 	&&
      !jm_gnet_inetaddr_is_broadcast(inetaddr))
    {
      return TRUE;
    }

  return FALSE;
}



/* 

   TRUE if inetaddr is an address reserved for private networks or
   something else.  This includes:

     10.0.0.0        -   10.255.255.255  (10/8 prefix)
     172.16.0.0      -   172.31.255.255  (172.16/12 prefix)
     192.168.0.0     -   192.168.255.255 (192.168/16 prefix)

   (from RFC 1918.  See also draft-manning-dsua-02.txt)

 */
gboolean
jm_gnet_inetaddr_is_private (GInetAddr* inetaddr)
{
  guint addr;

  g_return_val_if_fail (inetaddr != NULL, FALSE);

  addr = JMGNET_SOCKADDR_IN(inetaddr->sa).sin_addr.s_addr;
  addr = g_ntohl(addr);

  if ((addr & 0xFF000000) == (10 << 24))
    return TRUE;

  if ((addr & 0xFFF00000) == 0xAC100000)
    return TRUE;

  if ((addr & 0xFFFF0000) == 0xC0A80000)
    return TRUE;

  return FALSE;
}


/* 

   TRUE if inetaddr is reserved for 'something'.  This excludes
   address reserved for private networks.

   We check for:
     0.0.0.0/16  (top 16 bits are 0's)
     Class E (top 5 bits are 11110)

 */
gboolean
jm_gnet_inetaddr_is_reserved (GInetAddr* inetaddr)
{
  guint addr;

  g_return_val_if_fail (inetaddr != NULL, FALSE);

  addr = JMGNET_SOCKADDR_IN(inetaddr->sa).sin_addr.s_addr;
  addr = g_ntohl(addr);

  if ((addr & 0xFFFF0000) == 0)
    return TRUE;

  if ((addr & 0xF8000000) == 0xF0000000)
    return TRUE;

  return FALSE;
}



/* 

   TRUE if inetaddr is a loopback address 

   Loopback addresses: 127.0.0.0/8

*/
gboolean
jm_gnet_inetaddr_is_loopback(GInetAddr* inetaddr)
{
  guint addr;

  g_return_val_if_fail (inetaddr != NULL, FALSE);

  addr = JMGNET_SOCKADDR_IN(inetaddr->sa).sin_addr.s_addr;
  addr = g_ntohl(addr);

  if ((addr & 0xFF000000) == (127 << 24))
    return TRUE;

  return FALSE;
}




/* 
   
   TRUE if inetaddr is a multicast address 

   Multicast address: 224.0.0.1 - 239.255.255.255.
    or top four bits are 1110, 
    or (addr & 0xF0000000) == 0xE0000000)

*/
gboolean 
jm_gnet_inetaddr_is_multicast(GInetAddr* inetaddr)
{
  guint addr;

  g_return_val_if_fail (inetaddr != NULL, FALSE);

  addr = JMGNET_SOCKADDR_IN(inetaddr->sa).sin_addr.s_addr;
  addr = g_htonl(addr);

  if ((addr & 0xF0000000) == 0xE0000000)
    return TRUE;

  return FALSE;
}


/* 
   
   TRUE if inetaddr is a broadcast address 

   Broadcast address: 255.255.255.255.255


*/
gboolean 
jm_gnet_inetaddr_is_broadcast(GInetAddr* inetaddr)
{
  guint addr;

  g_return_val_if_fail (inetaddr != NULL, FALSE);

  addr = JMGNET_SOCKADDR_IN(inetaddr->sa).sin_addr.s_addr;

  if (addr == 0xFFFFFFFF)
    return TRUE;

  return FALSE;
}



/* ************************************************************ */

typedef struct _GNetWriteAsyncState
{
  GIOChannel* iochannel;

  gchar* buffer;
  guint length;
  guint n;

  GNetWriteAsyncFunc func;
  gpointer user_data;

} GNetWriteAsyncState;

static gboolean gnet_write_async_write_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data);
static gboolean gnet_write_async_error_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data);
static gboolean gnet_write_async_timeout_cb (gpointer data);



GNetWriteAsyncID
jm_gnet_write_async(GIOChannel* iochannel, gchar* buffer, guint length, guint timeout,
		    GNetWriteAsyncFunc func, gpointer user_data)
{
  GNetWriteAsyncState* state;

  g_return_val_if_fail (iochannel != NULL, NULL);
  g_return_val_if_fail ((buffer != NULL && length != 0) || (buffer == NULL && length == 0), NULL);
  g_return_val_if_fail (func != NULL, NULL);

  if (buffer == NULL)
    {
      (func)(iochannel, buffer, length, 0, GNET_WRITE_ASYNC_STATUS_OK, user_data);
      return NULL;
    }

  state = g_new0(GNetWriteAsyncState, 1);

  state->iochannel = iochannel;
  state->buffer = buffer;
  state->length = length;
  state->n = 0;

  state->func = func;
  state->user_data = user_data;

  g_io_add_watch(iochannel, G_IO_OUT,  gnet_write_async_write_cb, state);
  g_io_add_watch(iochannel, G_IO_ERR,  gnet_write_async_error_cb, state);
  g_io_add_watch(iochannel, G_IO_HUP,  gnet_write_async_error_cb, state);
  g_io_add_watch(iochannel, G_IO_NVAL, gnet_write_async_error_cb, state);

  if (timeout > 0)
    g_timeout_add(timeout, gnet_write_async_timeout_cb, state);

  return state;
}


void
jm_gnet_write_async_cancel(GNetWriteAsyncID id, gboolean delete_buffer)
{
  GNetWriteAsyncState* state;

  g_return_if_fail (id != NULL);

  state = (GNetWriteAsyncState*) id;

  if (delete_buffer)
    g_free(state->buffer);

  while (g_source_remove_by_user_data(state))
    ;

  g_free(state);
}


static gboolean 
gnet_write_async_write_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data)
{
  GNetWriteAsyncState* state;
  guint bytes_writen;

  state = (GNetWriteAsyncState*) data;

  g_return_val_if_fail (iochannel != NULL, FALSE);
  g_return_val_if_fail (state != NULL, FALSE);
  g_return_val_if_fail (condition == G_IO_OUT, FALSE);
  g_return_val_if_fail (iochannel == state->iochannel, FALSE);

  if (g_io_channel_write(iochannel, 
			 &state->buffer[state->n], 
			 state->length - state->n,
			 &bytes_writen) != G_IO_ERROR_NONE)
    {
      state->func(iochannel, state->buffer, state->length, state->n,
		  GNET_WRITE_ASYNC_STATUS_ERROR, state->user_data);
      jm_gnet_write_async_cancel(state, FALSE);
      return FALSE;
    }

  state->n += bytes_writen;

  if (state->n == state->length)
    {
      state->func(iochannel, state->buffer, state->length, state->n,
		  GNET_WRITE_ASYNC_STATUS_OK, state->user_data);
      jm_gnet_write_async_cancel(state, FALSE);
      return FALSE;
    }

  return TRUE;
}


static gboolean 
gnet_write_async_error_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data)
{
  GNetWriteAsyncState* state;

  state = (GNetWriteAsyncState*) data;

  g_return_val_if_fail (iochannel != NULL, FALSE);
  g_return_val_if_fail (state != NULL, FALSE);
  g_return_val_if_fail (iochannel == state->iochannel, FALSE);

  state->func(iochannel, state->buffer, state->length, state->n,
	      GNET_WRITE_ASYNC_STATUS_ERROR, state->user_data);
  jm_gnet_write_async_cancel(state, FALSE);

  return FALSE;
}


static gboolean 
gnet_write_async_timeout_cb (gpointer data)
{
  GNetWriteAsyncState* state;

  state = (GNetWriteAsyncState*) data;

  g_return_val_if_fail (state != NULL, FALSE);

  state->func(state->iochannel, state->buffer, state->length, state->n,
	      GNET_WRITE_ASYNC_STATUS_TIMEOUT, state->user_data);
  jm_gnet_write_async_cancel(state, FALSE);

  return FALSE;
}



/* ************************************************************ */

typedef struct _GNetReadAsyncState
{
  GIOChannel* iochannel;

  gboolean read_one;
  gboolean my_buffer;
  gchar* buffer;
  guint max_len;
  guint length;
  guint offset;
  guint timeout;

  guint read_watch;
  guint other_watch;
  guint timer;

  GNetReadAsyncCheckFunc check_func;
  gpointer check_user_data;

  GNetReadAsyncFunc func;
  gpointer user_data;

} GNetReadAsyncState;

static gboolean gnet_read_async_read_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data);
static gboolean gnet_read_async_error_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data);
static gboolean gnet_read_async_timeout_cb (gpointer data);



GNetReadAsyncID
jm_gnet_read_async (GIOChannel* iochannel, 
		    gchar* buffer, guint length, guint timeout, 
		    gboolean read_one_byte_at_a_time, 
		    GNetReadAsyncCheckFunc check_func, gpointer check_user_data,
		    GNetReadAsyncFunc func, 	  gpointer user_data)
{
  GNetReadAsyncState* state;

  g_return_val_if_fail (iochannel, NULL);
  g_return_val_if_fail (check_func, NULL);
  g_return_val_if_fail (func, NULL);
  g_return_val_if_fail (buffer || (!buffer && length), NULL);

  state = g_new0 (GNetReadAsyncState, 1);

  state->iochannel = iochannel;
  state->read_one = read_one_byte_at_a_time;

  if (buffer)
    {
      state->my_buffer = FALSE;
      state->buffer = buffer;
      state->max_len = length;
      state->length = length;
      state->offset = 0;
    }
  else
    {
      state->my_buffer = TRUE;
      state->buffer = NULL;
      state->max_len = length;
      state->length = 0;
      state->offset = 0;
    }

  state->check_func = 		check_func;
  state->check_user_data = 	check_user_data;
  state->func = 		func;
  state->user_data = 		user_data;

  state->read_watch = g_io_add_watch(iochannel, G_IO_IN,   
				     gnet_read_async_read_cb, state);
  state->other_watch = g_io_add_watch(iochannel, G_IO_ERR | G_IO_HUP | G_IO_NVAL, 
				      gnet_read_async_error_cb, state);

  state->timeout = timeout;
  if (timeout > 0)
    state->timer = g_timeout_add(timeout, gnet_read_async_timeout_cb, state);

  return state;
}


void
jm_gnet_read_async_cancel (GNetReadAsyncID id)
{
  GNetReadAsyncState* state;

  g_return_if_fail (id != NULL);

  state = (GNetReadAsyncState*) id;

  if (state->read_watch)
    g_source_remove (state->read_watch);
  if (state->other_watch)
    g_source_remove (state->other_watch);
  if (state->timer)
    g_source_remove (state->timer);

  if (state->my_buffer)
    g_free(state->buffer);

  g_free(state);
}


static gboolean 
gnet_read_async_read_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data)
{
  GNetReadAsyncState* state;
  GIOError error;
  guint bytes_to_read;
  guint bytes_read;
  gint bytes_processed;

  state = (GNetReadAsyncState*) data;

  g_return_val_if_fail (iochannel != NULL, FALSE);
  g_return_val_if_fail (state != NULL, FALSE);
  g_return_val_if_fail (condition == G_IO_IN, FALSE);
  g_return_val_if_fail (iochannel == state->iochannel, FALSE);


  /* Check if we should make the buffer larger */
  if (state->my_buffer && state->length == state->offset)
    {
      if (state->length)
	{
	  state->length *= 2;
	  state->buffer = g_realloc(state->buffer, state->length);
	}
      else
	{
	  state->length = MIN (128, state->max_len);
	  state->buffer = g_malloc(state->length);
	}
    }


  if (state->read_one)
    bytes_to_read = 1;
  else
    bytes_to_read = state->length - state->offset;

  /* Read in some stuff */
  error = g_io_channel_read(iochannel, &state->buffer[state->offset], 
			    bytes_to_read, &bytes_read);
  state->offset += bytes_read;

  if (error == G_IO_ERROR_AGAIN)
    return TRUE;

  else if (error != G_IO_ERROR_NONE)
    {
      state->func (iochannel, GNET_READ_ASYNC_STATUS_ERROR, NULL, 0, state->user_data);
      jm_gnet_read_async_cancel (state);
      return FALSE;
    }

  /* If we read nothing, that means EOF and we're done.  Note we do
     not send anything that might be in the buffer */
  else if (bytes_read == 0)
    {
      state->func (iochannel, GNET_READ_ASYNC_STATUS_OK, NULL, 0, state->user_data);
      jm_gnet_read_async_cancel (state);
      return FALSE;
    }

  /* Check if we read something */
 again:
  bytes_processed = (state->check_func)(state->buffer, state->offset, state->check_user_data);
  if (bytes_processed)
    {
      if (!state->func (iochannel, GNET_READ_ASYNC_STATUS_OK, 
			state->buffer, bytes_processed, state->user_data))
	{
	  jm_gnet_read_async_cancel (state);
	  return FALSE;
	}

      /* Move it over */
      g_memmove (state->buffer, &state->buffer[bytes_processed], 
		 state->offset - bytes_processed);
      state->offset -= bytes_processed;
      goto again;
    }

  /* Check if we hit the max length.  If so, it's an error. */
  if (state->offset >= state->max_len)
    {
      state->func(iochannel, GNET_READ_ASYNC_STATUS_ERROR, 
		  state->buffer, state->offset, state->user_data);
      jm_gnet_read_async_cancel (state);
      return FALSE;
    }

  /* Reset the timer */
  if (state->timer)
    {
      g_assert (g_source_remove (state->timer));
      state->timer = g_timeout_add (state->timeout, gnet_read_async_timeout_cb, state);
    }

  return TRUE;
}


static gboolean 
gnet_read_async_error_cb (GIOChannel* iochannel, GIOCondition condition, gpointer data)
{
  GNetReadAsyncState* state;

  state = (GNetReadAsyncState*) data;

  g_return_val_if_fail (iochannel, FALSE);
  g_return_val_if_fail (state, FALSE);

  (state->func)(iochannel, GNET_READ_ASYNC_STATUS_ERROR, NULL, 0, state->user_data);
  jm_gnet_read_async_cancel(state);

  return FALSE;
}


static gboolean 
gnet_read_async_timeout_cb (gpointer data)
{
  GNetReadAsyncState* state;

  state = (GNetReadAsyncState*) data;

  g_return_val_if_fail (state, FALSE);

  (state->func)(state->iochannel, GNET_READ_ASYNC_STATUS_TIMEOUT, NULL, 0, state->user_data);
  jm_gnet_read_async_cancel(state);

  return FALSE;
}


gint
jm_gnet_readany_check_func (gchar* buffer, guint length, gpointer user_data)
{
  return length;
}


gint
jm_gnet_readline_check_func (gchar* buffer, guint length, gpointer user_data)
{
  guint i;

  for (i = 0; i < length; ++i)
    {
      if (buffer[i] == '\n')
	{
	  buffer[i] = '\0';
	  return i + 1;
	}
    }

  return 0;
}



/* ************************************************************ */


typedef struct _QueuedWrite
{
  gchar* buffer;
  gint length;
  gint timeout;

} QueuedWrite;

static void jm_conn_check_queued_writes (JMGConn* conn);

static void jm_conn_connect_cb (GTcpSocket* socket, GInetAddr* ia,
				 GTcpSocketConnectAsyncStatus status, 
				 gpointer user_data);

static void jm_conn_new_cb (GTcpSocket* socket, 
			     GTcpSocketNewAsyncStatus status, 
			     gpointer user_data);

static void jm_conn_write_cb (GIOChannel* iochannel, gchar* buffer, guint length, guint bytes_writen,
			       GNetWriteAsyncStatus status, gpointer user_data);

static gboolean jm_conn_read_cb (GIOChannel* iochannel, GNetReadAsyncStatus status, 
				 gchar* buffer, guint length, gpointer user_data);

static gboolean jm_conn_timeout_cb (gpointer data);


JMGConn*
jm_conn_new (gchar* hostname, gint port, JMGConnFunc func, gpointer user_data)
{
  JMGConn* conn;

  g_return_val_if_fail (hostname, NULL);

  conn = g_new0 (JMGConn, 1);
  conn->ref_count = 1;
  conn->hostname = g_strdup(hostname);
  conn->port = port;
  conn->inetaddr = jm_gnet_inetaddr_new_nonblock (hostname, port);
  conn->func = func;
  conn->user_data = user_data;

  return conn;
}


JMGConn*   
jm_conn_new_inetaddr (GInetAddr* inetaddr, JMGConnFunc func, gpointer user_data)
{
  JMGConn* conn;

  g_return_val_if_fail (inetaddr, NULL);

  conn = g_new0 (JMGConn, 1);
  conn->ref_count = 1;
  conn->hostname = gnet_inetaddr_get_canonical_name (inetaddr);
  conn->port = gnet_inetaddr_get_port (inetaddr);
  conn->inetaddr = gnet_inetaddr_clone (inetaddr);
  conn->func = func;
  conn->user_data = user_data;

  return conn;
}



void
jm_conn_delete (JMGConn* conn, gboolean delete_buffers)
{
  if (conn)
    {
      jm_conn_disconnect (conn, delete_buffers);

      if (conn->inetaddr)
	gnet_inetaddr_delete (conn->inetaddr);

      g_free (conn->hostname);

      if (conn->timer)
	g_source_remove (conn->timer);

      memset (conn, 0, sizeof(*conn));
      g_free (conn);
    }
}


void
jm_conn_ref (JMGConn* conn)
{
  g_return_if_fail (conn);

  ++conn->ref_count;
}


void
jm_conn_unref (JMGConn* conn, gboolean delete_buffers)
{
  g_return_if_fail (conn);

  --conn->ref_count;

  if (conn->ref_count == 0)
    jm_conn_delete (conn, delete_buffers);
}


void
jm_conn_connect (JMGConn* conn, guint timeout)
{
  g_return_if_fail (conn);
  g_return_if_fail (conn->func);

  if (conn->connect_id || conn->new_id || conn->socket)
    return;

  if (conn->inetaddr)
    conn->new_id = 
      gnet_tcp_socket_new_async (conn->inetaddr, jm_conn_new_cb, conn);
  
  else if (conn->hostname)
    conn->connect_id = 
      gnet_tcp_socket_connect_async (conn->hostname, conn->port, 
				     jm_conn_connect_cb, conn);
  else
    g_return_if_fail (FALSE);
}


static void
jm_conn_connect_cb (GTcpSocket* socket, GInetAddr* ia,
		    GTcpSocketConnectAsyncStatus status, 
		    gpointer user_data)
{
  JMGConn* conn = (JMGConn*) user_data;
  JMGConnStatus st = GNET_CONN_STATUS_ERROR;

  g_return_if_fail (conn);

  conn->connect_id = NULL;

  if (status == GTCP_SOCKET_CONNECT_ASYNC_STATUS_OK)
    {
      conn->socket = socket;
      conn->inetaddr = ia;
      conn->iochannel = gnet_tcp_socket_get_iochannel (socket);

      st = GNET_CONN_STATUS_CONNECT;

      jm_conn_check_queued_writes (conn);
    }

  (conn->func)(conn, st, NULL, 0, conn->user_data);
}


static void
jm_conn_new_cb (GTcpSocket* socket, 
		GTcpSocketNewAsyncStatus status, 
		gpointer user_data)
{
  JMGConn* conn = (JMGConn*) user_data;
  JMGConnStatus st = GNET_CONN_STATUS_ERROR;

  g_return_if_fail (conn);

  conn->new_id = NULL;

  if (status == GTCP_SOCKET_NEW_ASYNC_STATUS_OK)
    {
      conn->socket = socket;
      conn->iochannel = gnet_tcp_socket_get_iochannel (socket);

      st = GNET_CONN_STATUS_CONNECT;

      jm_conn_check_queued_writes (conn);
    }

  (conn->func)(conn, st, NULL, 0, conn->user_data);
}



void
jm_conn_disconnect (JMGConn* conn, gboolean delete_buffers)
{
  GList* i;

  g_return_if_fail (conn);

  if (conn->connect_id)
    {
      gnet_tcp_socket_connect_async_cancel (conn->connect_id);
      conn->connect_id = NULL;
    }

  if (conn->new_id)
    {
      gnet_tcp_socket_new_async_cancel (conn->new_id);
      conn->new_id = NULL;
    }

  jm_conn_watch_remove_read  (conn);
  jm_conn_watch_remove_write (conn);
  jm_conn_watch_remove_error (conn);

  for (i = conn->queued_writes; i != NULL; i = i->next)
    {
      QueuedWrite* queued_write = i->data;

      if (delete_buffers)
	g_free (queued_write->buffer);

      g_free (queued_write);
    }
  g_list_free (conn->queued_writes);
  conn->queued_writes = NULL;

  if (conn->write_id)
    {
      jm_gnet_write_async_cancel (conn->write_id, delete_buffers);
      conn->write_id = NULL;
    }
      
  if (conn->read_id)
    {
      jm_gnet_read_async_cancel (conn->read_id);
      conn->read_id = NULL;
    }

  if (conn->iochannel)
    {
      g_io_channel_unref (conn->iochannel);
      conn->iochannel = NULL;
    }

  if (conn->socket)
    {
      gnet_tcp_socket_delete (conn->socket);
      conn->socket = NULL;
    }
}


gboolean
jm_conn_is_connected (JMGConn* conn)
{
  g_return_val_if_fail (conn, FALSE);

  return (conn->socket != NULL);
}


void
jm_conn_watch_add_read (JMGConn* conn, GIOFunc func, gpointer user_data)
{
  g_return_if_fail (conn);
  g_return_if_fail (conn->iochannel);
  g_return_if_fail (!conn->read_id);

  if (!conn->read_watch)
    conn->read_watch = g_io_add_watch (conn->iochannel, G_IO_IN, func, user_data);
}


void
jm_conn_watch_add_write (JMGConn* conn, GIOFunc func, gpointer user_data)
{
  g_return_if_fail (conn);
  g_return_if_fail (conn->iochannel);
  g_return_if_fail (!conn->write_id);

  if (!conn->write_watch)
    conn->write_watch = g_io_add_watch (conn->iochannel, G_IO_OUT, func, user_data);
}


void
jm_conn_watch_add_error (JMGConn* conn, GIOFunc func, gpointer user_data)
{
  g_return_if_fail (conn);
  g_return_if_fail (conn->iochannel);

  if (!conn->err_watch)
    conn->err_watch  = 
      g_io_add_watch (conn->iochannel, G_IO_ERR | G_IO_HUP | G_IO_NVAL, func, user_data);
}


void
jm_conn_watch_remove_read (JMGConn* conn)
{
  g_return_if_fail (conn);

  if (conn->read_watch)
    {
      g_assert (g_source_remove(conn->read_watch));
      conn->read_watch = 0;
    }
}


void
jm_conn_watch_remove_write (JMGConn* conn)
{
  g_return_if_fail (conn);

  if (conn->write_watch)
    {
      g_assert (g_source_remove(conn->write_watch));
      conn->write_watch = 0;
    }
}


void
jm_conn_watch_remove_error (JMGConn* conn)
{
  g_return_if_fail (conn);

  if (conn->err_watch)
    {
      g_assert (g_source_remove(conn->err_watch));
      conn->err_watch = 0;
    }
}


void
jm_conn_read (JMGConn* conn, gchar* buffer, guint length, guint timeout,
	      gboolean read_one_byte_at_a_time,
	      GNetReadAsyncCheckFunc check_func, gpointer check_user_data)
{
  g_return_if_fail (conn);
  g_return_if_fail (conn->iochannel);
  g_return_if_fail (conn->func);
  g_return_if_fail (!conn->read_id);

  conn->read_id = jm_gnet_read_async (conn->iochannel, buffer, length, timeout,
				      read_one_byte_at_a_time,
				      check_func, check_user_data,
				      jm_conn_read_cb, conn);
}


void
jm_conn_readany (JMGConn* conn, gchar* buffer, guint length, guint timeout)
{
  g_return_if_fail (conn);
  g_return_if_fail (buffer);
  g_return_if_fail (conn->func);
  g_return_if_fail (conn->iochannel);
  g_return_if_fail (!conn->read_id);

  conn->read_id = jm_gnet_readany_async (conn->iochannel, buffer, length, timeout,
					 jm_conn_read_cb, conn);
}


void
jm_conn_readline (JMGConn* conn, gchar* buffer, guint length, guint timeout)
{
  g_return_if_fail (conn);
  g_return_if_fail (conn->func);
  g_return_if_fail (conn->iochannel);
  g_return_if_fail (!conn->read_id);

  conn->read_id = jm_gnet_readline_async (conn->iochannel, buffer, length, timeout,
					  jm_conn_read_cb, conn);
}


static gboolean
jm_conn_read_cb (GIOChannel* iochannel, GNetReadAsyncStatus status, 
		 gchar* buffer, guint length, gpointer user_data)
{
  JMGConn* conn = (JMGConn*) user_data;
  gpointer read_id;

  g_return_val_if_fail (conn, FALSE);

  /* If the upper level calls disconnect, we don't want it to delete
     the read_id. */
  read_id = conn->read_id;
  conn->read_id = NULL;

  if (status == GNET_READ_ASYNC_STATUS_OK)
    {
      if (length)
	{
	  gboolean rv;

	  rv = (conn->func)(conn, GNET_CONN_STATUS_READ, buffer, length, conn->user_data);
	  if (rv)
	    conn->read_id = read_id;
	  return rv;
	}
      else
	{
	  (conn->func)(conn, GNET_CONN_STATUS_CLOSE, NULL, 0, conn->user_data);
	}
    }
  else
    {
      (conn->func)(conn, GNET_CONN_STATUS_ERROR, NULL, 0, conn->user_data);
    }

  return FALSE;
}


void
jm_conn_write (JMGConn* conn, gchar* buffer, gint length, guint timeout)
{
  g_return_if_fail (conn);
  g_return_if_fail (conn->func);

  if (conn->iochannel && !conn->write_id)
    {
      conn->write_id = jm_gnet_write_async (conn->iochannel, buffer, length,
					    timeout, jm_conn_write_cb, conn);
    }
  else
    {
      QueuedWrite* queued_write = g_new0 (QueuedWrite, 1);
      queued_write->buffer = buffer;
      queued_write->length = length;
      queued_write->timeout = timeout;

      conn->queued_writes = g_list_append (conn->queued_writes, queued_write);
    }
}


static void
jm_conn_write_cb (GIOChannel* iochannel, gchar* buffer, guint length, guint bytes_writen,
		   GNetWriteAsyncStatus status, gpointer user_data)
{
  JMGConn* conn = (JMGConn*) user_data;

  g_return_if_fail (conn);

  conn->write_id = NULL;

  if (status == GNET_WRITE_ASYNC_STATUS_OK)
    {
      jm_conn_check_queued_writes (conn);

      (conn->func)(conn, GNET_CONN_STATUS_WRITE, buffer, length, conn->user_data);
    }
  else
    {
      (conn->func)(conn, GNET_CONN_STATUS_ERROR, NULL, 0, conn->user_data);
    }
}


void
jm_conn_timeout (JMGConn* conn, guint timeout)
{
  g_return_if_fail (conn);
  g_return_if_fail (conn->func);

  if (conn->timer)
    {
      g_source_remove (conn->timer);
      conn->timer = 0;
    }

  if (timeout)
    {
      conn->timer = g_timeout_add (timeout, jm_conn_timeout_cb, conn);
    }
}


static gboolean 
jm_conn_timeout_cb (gpointer data)
{
  JMGConn* conn = (JMGConn*) data;

  g_return_val_if_fail (conn, FALSE);

  (conn->func)(conn, GNET_CONN_STATUS_TIMEOUT, NULL, 0, conn->user_data);

  return FALSE;
}


static void
jm_conn_check_queued_writes (JMGConn* conn)
{
  g_return_if_fail (conn);
  g_return_if_fail (conn->iochannel);

  if (conn->write_id)
    g_return_if_fail (FALSE);

  if (conn->queued_writes)
    {
      QueuedWrite* queued_write = conn->queued_writes->data;
      conn->queued_writes = g_list_remove_link (conn->queued_writes, 
						conn->queued_writes);

      conn->write_id = jm_gnet_write_async (conn->iochannel, 
					    queued_write->buffer, queued_write->length,
					    queued_write->timeout, 
					    jm_conn_write_cb, conn);
      g_free (queued_write);
    }
}


/* ************************************************************ */

static gboolean jm_server_accept_cb (GIOChannel* listen_iochannel, 
				      GIOCondition condition, gpointer user_data);


JMGServer*  
jm_server_new (GInetAddr* interface, gint port, gboolean force_port, 
		JMGServerFunc func, gpointer user_data)
{
  JMGServer* server = NULL;

  g_return_val_if_fail (func, NULL);

  server = g_new0 (JMGServer, 1);
  server->func = func;
  server->user_data = user_data;

  /* Create a listening socket */
  if (port)
    {
      server->socket = jm_gnet_tcp_socket_server_new (interface, port);
      if (!server->socket && force_port)
	goto error;
    }

  if (!server->socket)
    server->socket = jm_gnet_tcp_socket_server_new (interface, 0);

  if (!server->socket)
    goto error;

  /* Get the port number */
  server->port = gnet_tcp_socket_get_port (server->socket);
  if (server->port == 0)
    goto error;

  /* Get the address */
  server->interface = gnet_tcp_socket_get_inetaddr (server->socket);

  /* Get the iochannel */
  server->iochannel = gnet_tcp_socket_get_iochannel (server->socket);
  if (server->iochannel == NULL)
    goto error;

  /* Wait for new connections */
  server->watch_id = g_io_add_watch (server->iochannel, G_IO_IN | G_IO_ERR | G_IO_NVAL, 
				     jm_server_accept_cb, server);

  return server;

 error:
  jm_server_delete (server);
  return NULL;
}



void
jm_server_delete (JMGServer* server)
{
  if (server)
    {
      if (server->watch_id)      g_source_remove (server->watch_id);
      if (server->iochannel)     g_io_channel_unref (server->iochannel);
      if (server->socket)	 gnet_tcp_socket_delete (server->socket);
      if (server->interface)     gnet_inetaddr_delete (server->interface);

      memset (server, 0, sizeof(*server));
      g_free (server);
    }
}



static gboolean
jm_server_accept_cb (GIOChannel* listen_iochannel, GIOCondition condition, gpointer user_data)
{
  JMGServer* server = (JMGServer*) user_data;

  g_return_val_if_fail (server, FALSE);

  if (condition == G_IO_IN)
    {
      GTcpSocket* socket = NULL;
      GIOChannel* iochannel = NULL;
      JMGConn* conn; 

      /* Accept the connection */
      socket = gnet_tcp_socket_server_accept_nonblock (server->socket);

      /* Ignore if the connection is now gone */
      if (socket == NULL) 
	return TRUE;

      /* Get the iochannel */
      iochannel = gnet_tcp_socket_get_iochannel (socket);
      g_return_val_if_fail (iochannel, FALSE);

      /* Create a Connection */
      conn = g_new0 (JMGConn, 1);
      conn->socket = socket;
      conn->iochannel = iochannel;
      conn->inetaddr = gnet_tcp_socket_get_inetaddr (socket);
      conn->hostname = gnet_inetaddr_get_canonical_name (conn->inetaddr);
      conn->port = gnet_inetaddr_get_port (conn->inetaddr);

      (server->func)(server, GNET_SERVER_STATUS_CONNECT, conn, server->user_data);
    }
  else
    {
      (server->func)(server, GNET_SERVER_STATUS_ERROR, NULL, server->user_data);
      return FALSE;
    }

  return TRUE;
}



/* ************************************************************ */


JMGURL* 
jm_url_new (const gchar* url)
{
  JMGURL* gurl = NULL;
  const gchar* p;
  const gchar* temp;

  g_return_val_if_fail (url, NULL);

  gurl = g_new0 (JMGURL, 1);

  /* Skip initial whitespace */
  while (*url != 0 && isspace(*url))
    ++url;
  p = url;

  /* Read in protocol (protocol is optional) */
  temp = p;
  while (*p != 0 && *p != ':' && *p != '/')
    ++p;
  if (*p == ':' && p[1] == '/')
    {
      gurl->protocol = g_strndup (url, p - url);
      while (*p != 0 && (*p == ':' || *p == '/')) ++p;
    }
  else	/* This is a \0 or /, so it must be the hostname */
    p = temp;

  /* Read in the hostname */
  temp = p;
  while (*p != 0 && *p != '/' && *p != ':') ++p;
  if ((p - temp) == 0) 
    {
      jm_url_delete (gurl);
      return NULL;
    }
  gurl->hostname = g_strndup (temp, p - temp);

  /* Read in the port */
  if (*p == ':')
    {
      temp = ++p;
      while (*p != 0 && *p != '/') ++p;
      if ((p - temp) != 0)
	{
	  gchar* port_str = g_strndup (temp, p - temp);
	  gurl->port = atoi (port_str);
	  g_free (port_str);
	}
    }

  /* Read in the file */
  if (*p == 0)
    gurl->resource = g_strdup("/");
  else
    gurl->resource = g_strdup(p);

  return gurl;
}


JMGURL*     
jm_url_new_fields (const gchar* protocol, const gchar* hostname, 
		   const gint port, const gchar* resource)
{
  JMGURL* url = NULL;

  url = g_new0 (JMGURL, 1);
  if (protocol)		url->protocol = g_strdup (protocol);
  if (hostname)		url->hostname = g_strdup (hostname);
  url->port = port;
  if (resource)		url->resource = g_strdup (resource);

  return url;
}


JMGURL*     
jm_url_clone (const JMGURL* url)
{
  JMGURL* url2;

  g_return_val_if_fail (url, NULL);

  url2 = g_new0 (JMGURL, 1);
  if (url->protocol)
    url2->protocol = g_strdup (url->protocol);
  if (url->hostname)
    url2->hostname = g_strdup (url->hostname);
  url2->port = url->port;
  if (url->resource)
    url2->resource = g_strdup (url->resource);

  return url2;
}


void
jm_url_delete (JMGURL* url)
{
  if (url)
    {
      g_free (url->protocol);
      g_free (url->hostname);
      g_free (url->resource);
      g_free (url);
    }
}


guint
jm_url_hash (const gpointer p)
{
  const JMGURL* url = (const JMGURL*) p;
  guint h = 0;

  g_return_val_if_fail (url, 0);

  if (url->protocol)	h =  g_str_hash (url->protocol);
  if (url->hostname)	h ^= g_str_hash (url->hostname);
  h ^= url->port;
  if (url->resource)	h ^= g_str_hash (url->resource);
  
  return h;
}


gint
jm_url_equal (const gpointer p1, const gpointer p2)
{
  const JMGURL* url1 = (const JMGURL*) p1;
  const JMGURL* url2 = (const JMGURL*) p2;

  g_return_val_if_fail (url1, 0);
  g_return_val_if_fail (url2, 0);

  if (url1->port == url2->port &&

      ( (!url1->protocol && !url2->protocol) ||
	 (url1->protocol &&  url2->protocol &&
	  !strcmp (url1->protocol, url2->protocol)) ) &&

      ( (!url1->hostname && !url2->hostname) ||
	 (url1->hostname &&  url2->hostname &&
	  !strcmp (url1->hostname, url2->hostname)) ) &&

      ( (!url1->resource && !url2->resource) ||
	 (url1->resource &&  url2->resource &&
	  !strcmp (url1->resource, url2->resource)) ) )
    return 1;

  return 0;
}


void
jm_url_set_protocol (JMGURL* url, gchar* protocol)
{
  g_return_if_fail (url);

  if (url->protocol)
    {
      g_free (url->protocol);
      url->protocol = NULL;
    }

  if (protocol)
    url->protocol = g_strdup (protocol);
}


void
jm_url_set_hostname (JMGURL* url, gchar* hostname)
{
  g_return_if_fail (url);

  if (url->hostname)
    {
      g_free (url->hostname);
      url->hostname = NULL;
    }

  if (hostname)
    url->hostname = g_strdup (hostname);
}


void	
jm_url_set_port     (JMGURL* url, gint port)
{
  url->port = port;
}


void
jm_url_set_resource (JMGURL* url, gchar* resource)
{
  g_return_if_fail (url);

  if (url->resource)
    {
      g_free (url->resource);
      url->resource = NULL;
    }

  if (resource)
    {
      if (resource[0] == '/')
	url->resource = g_strdup (resource);
      else
	url->resource = g_strconcat ("/", resource, NULL);
    }
}


gchar*
jm_url_get_nice_string (JMGURL* url)
{
  gchar* rv = NULL;
  gchar* resource = NULL;

  g_return_val_if_fail (url, NULL);
  g_return_val_if_fail (url->hostname, NULL);

  if (url->resource)
    {
      if (*url->resource == '/')
	resource = g_strdup (url->resource);
      else
	resource = g_strconcat ("/", url->resource, NULL);
    }
  else
    resource = g_strdup ("");

  if (url->port)
    rv = g_strdup_printf ("%s:%d%s", url->hostname, url->port, resource);
  else
    {
      if (*resource)
	rv = g_strconcat (url->hostname, resource, NULL);
      else
	rv = g_strdup (url->hostname);
    }

  if (url->protocol)
    {
      gchar* temp;

      temp = g_strconcat (url->protocol, "://", rv, NULL);
      g_free (rv);
      rv = temp;
    }

  g_free (resource);

  return rv;
}
