/* GNet - Networking library
 * Copyright (C) 2000  David Helder
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the 
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */


/* 

   This file contains functions that will probably be moved to GNet
   eventually. 

   The naming conventions here are crappy.

*/

#ifndef _JMGNET_H
#define _JMGNET_H


#include <glib.h>
#include <gnet/gnet.h>

#include <limits.h>

#ifndef G_MAXUINT
#define G_MAXUINT UINT_MAX
#endif

typedef enum
{
  JMGTOS_NONE,
  JMGTOS_LOWDELAY,
  JMGTOS_THROUGHPUT,
  JMGTOS_RELIABILITY,
  JMGTOS_LOWCOST

} JMGTOS;

void jmgnet_tcp_socket_set_tos (GTcpSocket* socket, JMGTOS tos);

GTcpSocket* jm_gnet_tcp_socket_server_new (GInetAddr* iface, const gint port);


/* ******************** */

GList* jm_gnet_inetaddr_list_interfaces (void);


/* Create address, but don't block.  In other words, if the name is a
   canonical name or "localhost", it returns the address.  Otherwise,
   it returns NULL. */
GInetAddr* jm_gnet_inetaddr_new_nonblock (const gchar* name, const gint port);

gboolean jm_gnet_inetaddr_is_canonical (const gchar* name);


GInetAddr* jm_gnet_inetaddr_autodetect_interface (void);
GInetAddr* jm_gnet_inetaddr_get_interface_to (GInetAddr* addr);
GInetAddr* jm_gnet_inetaddr_get_internet_interface (void);

gchar*	   jm_gnet_inetaddr_autodetect_address (GInetAddr* interface);


/**

   Get the domain name based on the interface.  If no_dns is
   specified, it will use the dotted decimal address as the domain
   name and not use DNS.

 */
gchar*
jm_gnet_inetaddr_get_internet_domainname (GInetAddr* interface, gboolean no_dns);




/**

   Use heuristics to check if the domain name is a sensible internet
   domain name.  This function does not use DNS.  Returns TRUE if it
   is, FALSE otherwise.

 */
gboolean jm_gnet_inetaddr_is_internet_domainname (gchar* domainname);



/**

   Check if the interface is a sensible internet address.  This mean
   it is not private, loopback, multicast, or broadcast.

   Not that private and lop

   Returns TRUE if it is, FALSE otherwise. 

*/
gboolean jm_gnet_inetaddr_is_internet (GInetAddr* inetaddr);

/* TRUE if inetaddr is an address reserved for private networks */
gboolean jm_gnet_inetaddr_is_private (GInetAddr* inetaddr);

/* TRUE if inetaddr is reserved in general */
gboolean jm_gnet_inetaddr_is_reserved (GInetAddr* inetaddr);

/* TRUE if inetaddr is a loopback address */
gboolean jm_gnet_inetaddr_is_loopback (GInetAddr* inetaddr);

/* TRUE if inetaddr is a multicast address */
gboolean jm_gnet_inetaddr_is_multicast (GInetAddr* inetaddr);

/* TRUE if inetaddr is a broadcast address */
gboolean jm_gnet_inetaddr_is_broadcast (GInetAddr* inetaddr);



/* **************************************** */

/*
  Notes

  _Do_not_ call cancel if the callback has been called.  For example, if
  you receive a callback with the status of TIMEOUT, _do_not_ cancel
  the read or write.

  A timeout of 0 means there is no timeout.  If you are writing a
  server, you should use a timeout.  2 minutes is reasonable for most
  small things.

*/


typedef enum {
  GNET_WRITE_ASYNC_STATUS_OK,
  GNET_WRITE_ASYNC_STATUS_TIMEOUT,
  GNET_WRITE_ASYNC_STATUS_ERROR
} GNetWriteAsyncStatus;

typedef gpointer GNetWriteAsyncID;

typedef void (*GNetWriteAsyncFunc)(GIOChannel* iochannel,
				   gchar* buffer,		/* you own (passed in write_async) */
				   guint length,
				   guint bytes_writen,
				   GNetWriteAsyncStatus status, 
				   gpointer user_data);

GNetWriteAsyncID jm_gnet_write_async(GIOChannel* iochannel, gchar* buffer, guint length, guint timeout,
				     GNetWriteAsyncFunc func, gpointer user_data);

void jm_gnet_write_async_cancel(GNetWriteAsyncID id, gboolean delete_buffer);


/* **************************************** */

typedef enum {
  GNET_READ_ASYNC_STATUS_OK,
  GNET_READ_ASYNC_STATUS_TIMEOUT,
  GNET_READ_ASYNC_STATUS_ERROR
} GNetReadAsyncStatus;

typedef gpointer GNetReadAsyncID;

typedef gboolean (*GNetReadAsyncFunc)(GIOChannel* iochannel, GNetReadAsyncStatus status, 
				      gchar* buffer, guint length, gpointer user_data);
/* ID is invalid if there is an error.  If OK and buffer and length is
   0, then it's an EOF. */

typedef gint (*GNetReadAsyncCheckFunc)(gchar* buffer, guint length, gpointer user_data);
/* Return -1 if error, otherwise number of bytes read. */

/* Set buffer to NULL if you want the buffer created dynamicly (length
   will be the maximum length of the buffer.  

   If read_one_byte_at_a_time is TRUE, it will read one byte at a time
   and call the check function.  This is really used for the readline
   function when you read a line, return FALSE, and then call readany.
   The problem is that readline may have stuff still in its buffer
   that now gets lost.  This would be bad.

   If you call read_async once and always return TRUE, then don't set
   read_one_byte_at_a_time.

*/


GNetReadAsyncID jm_gnet_read_async (GIOChannel* iochannel, 
				    gchar* buffer, guint length, guint timeout, 
				    gboolean read_one_byte_at_a_time, 
				    GNetReadAsyncCheckFunc check_func, gpointer check_user_data,
				    GNetReadAsyncFunc func, 	       gpointer user_data);

gint jm_gnet_readany_check_func (gchar* buffer, guint length, gpointer data);
gint jm_gnet_readline_check_func (gchar* buffer, guint length, gpointer data);

#define jm_gnet_readany_async(IO, BUF, LEN, TO, FUNC, UD)			\
  jm_gnet_read_async (IO, BUF, LEN, TO, FALSE, 					\
		      jm_gnet_readany_check_func, NULL, FUNC, UD)

#define jm_gnet_readline_async(IO, BUF, LEN, TO, FUNC, UD)			\
  jm_gnet_read_async (IO, BUF, LEN, TO, TRUE, 					\
		      jm_gnet_readline_check_func, NULL, FUNC, UD)

void jm_gnet_read_async_cancel(GNetReadAsyncID id);


/* **************************************** */

typedef enum {
  GNET_CONN_STATUS_CONNECT,
  GNET_CONN_STATUS_CLOSE,
  GNET_CONN_STATUS_READ,
  GNET_CONN_STATUS_WRITE,
  GNET_CONN_STATUS_TIMEOUT,
  GNET_CONN_STATUS_ERROR
} JMGConnStatus;

typedef struct _JMGConn JMGConn;

/* rv only affects reads... */
typedef gboolean (*JMGConnFunc)(JMGConn* conn,
				JMGConnStatus status, 
				gchar* buffer,		/* gnet owns on read, you do if write */
				gint length,
				gpointer user_data);


struct _JMGConn
{
  gchar*		hostname;
  gint			port;

  guint			ref_count;

  GTcpSocketConnectAsyncID connect_id;
  GTcpSocketNewAsyncID 	new_id;

  GTcpSocket* 		socket;
  GInetAddr*		inetaddr;
  GIOChannel* 		iochannel;

  guint			read_watch;
  guint			write_watch;
  guint			err_watch;

  GNetWriteAsyncID 	write_id;
  GList*		queued_writes;

  GNetReadAsyncID  	read_id;

  guint			timer;

  JMGConnFunc		func;
  gpointer		user_data;

};

JMGConn*   jm_conn_new (gchar* hostname, gint port, JMGConnFunc func, gpointer user_data);
JMGConn*   jm_conn_new_inetaddr (GInetAddr* inetaddr, JMGConnFunc func, gpointer user_data);
void 	   jm_conn_delete (JMGConn* conn, gboolean delete_buffers);
	      
void	   jm_conn_ref (JMGConn* conn);
void	   jm_conn_unref (JMGConn* conn, gboolean delete_buffers);
	      
void	   jm_conn_connect (JMGConn* conn, guint timeout);
void	   jm_conn_disconnect (JMGConn* conn, gboolean delete_buffers);
	      
gboolean   jm_conn_is_connected (JMGConn* conn);
	      
void	   jm_conn_watch_add_read (JMGConn* conn, GIOFunc func, gpointer user_data);
void	   jm_conn_watch_add_write (JMGConn* conn, GIOFunc func, gpointer user_data);
void	   jm_conn_watch_add_error (JMGConn* conn, GIOFunc func, gpointer user_data);

void	   jm_conn_watch_remove_read (JMGConn* conn);
void	   jm_conn_watch_remove_write (JMGConn* conn);
void	   jm_conn_watch_remove_error (JMGConn* conn);

void	   jm_conn_read (JMGConn* conn, gchar* buffer, guint length, guint timeout,
			 gboolean read_one_byte_at_a_time,
			 GNetReadAsyncCheckFunc check_func, gpointer check_user_data);
void	   jm_conn_readany (JMGConn* conn, gchar* buffer, guint length, guint timeout);
void	   jm_conn_readline (JMGConn* conn, gchar* buffer, guint length, guint timeout);
void	   jm_conn_write (JMGConn* conn, gchar* buffer, gint length, guint timeout);
void	   jm_conn_timeout (JMGConn* conn, guint timeout);


/* **************************************** */


typedef struct _JMGServer JMGServer;

typedef enum
{
  GNET_SERVER_STATUS_CONNECT,
  GNET_SERVER_STATUS_ERROR

} JMGServerStatus;

typedef void (*JMGServerFunc)(JMGServer* server,
			      JMGServerStatus status, 
			      JMGConn* conn,
			      gpointer user_data);


struct _JMGServer
{
  GInetAddr* 	interface;
  gint		port;

  GTcpSocket* 	socket;
  GIOChannel* 	iochannel;

  guint		watch_id;

  JMGServerFunc	func;
  gpointer	user_data;

};


JMGServer*  jm_server_new (GInetAddr* interface, gint port, gboolean force_port, 
			   JMGServerFunc func, gpointer user_data);
void 	    jm_server_delete (JMGServer* server);



/* **************************************** */
/* This isn't finished yet...		    */

typedef struct _JMGURL
{
  gchar* protocol;
  gchar* hostname;
  gint port;
  gchar* resource;

} JMGURL;


JMGURL*     jm_url_new (const gchar* url);
JMGURL*     jm_url_new_fields (const gchar* protocol, const gchar* hostname, 
			       const gint port, const gchar* resource);
JMGURL*     jm_url_clone (const JMGURL* url);
void 	    jm_url_delete (JMGURL* url);
	       
guint 	    jm_url_hash (const gpointer p);	/* untested */
gint 	    jm_url_equal (const gpointer p1, const gpointer p2); /* untested */

void 	    jm_url_set_protocol (JMGURL* url, gchar* protocol);
void 	    jm_url_set_hostname (JMGURL* url, gchar* hostname);
void 	    jm_url_set_port     (JMGURL* url, gint port);
void 	    jm_url_set_resource (JMGURL* url, gchar* resource);
	       
gchar* 	    jm_url_get_nice_string (JMGURL* url);


#if 0

equal
hash

void jm_gnet_url_get_canonical_string(JMGURL* url);

void jm_gnet_url_set_default_port (JMGURL* url, gint port);


#endif


#endif /* _JMGNET_H */
