/* Jungle Monkey Lite
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <dirent.h>

#ifdef JM_ENABLE_GNOME
#include <popt-gnome.h>
#else
#include <popt.h>
#endif

#include <pwd.h>
#include <grp.h>
#include <sys/types.h>

#include "jmintl.h"
#include "backend.h"
#include "util.h"
#include "globals.h"
#include "jmgnet.h"


/* ******************** */
/* Globals		*/

static int   		verbose = 0;
static JMFile* 		root_channel = NULL;


int main(int argc, char* argv[])
{
  gint is_channels_only = FALSE;
  gchar* root_name = NULL;
  gchar* url = NULL;
  GMainLoop* main_loop;

  gchar* hostname = NULL;
  gchar* hostname_name = NULL;

  GInetAddr* interface = NULL;
  gchar* interface_name = NULL;
  gboolean use_dns = FALSE;

  gint btp_port = 0;
  gboolean force_btp_port = FALSE;

  gint mtp_port = 0;
  gboolean force_mtp_port = FALSE;
  gboolean global_rendezvous = FALSE;

  struct poptOption jm_options[] = 
  {
    {"help", 'h', POPT_ARG_NONE, NULL, 1, 
     "Show this help message", NULL},
    {"usage", '\0', POPT_ARG_NONE, NULL, 2, 
     "Display brief usage message", NULL},
    {"version", 0, POPT_ARG_NONE, NULL, 3,
     "Output verson information and exit", NULL},
    {"verbose", 'v', POPT_ARG_NONE, &verbose, 0,
     "Print lots of stuff", NULL},

    {"debug-level", 'd', POPT_ARG_INT, &debug_level, 0, 
     "Set the debug level", "LEVEL"},

    {"hostname", '\0', POPT_ARG_STRING, &hostname_name, 0,
     "Hostname or address of this host", "HOSTNAME"},
    {"use-dns", '\0', POPT_ARG_NONE, &use_dns, 0,
     "Use DNS to figure out my hostname", NULL},

    {"interface", '\0', POPT_ARG_STRING, &interface_name, 0,
     "Interface address for listening socket", "ADDR"},

    {"btp-port", '\0', POPT_ARG_STRING, &btp_port, 0,
     "Port for BTP listening socket", "PORT"},
    {"force-btp-port", '\0', POPT_ARG_NONE, &force_btp_port, 0,
     "Fail if can't set MTP port", NULL},

    {"mtp-port", '\0', POPT_ARG_STRING, &mtp_port, 0,
     "Port for MTP listening socket", "PORT"},
    {"force-mtp-port", '\0', POPT_ARG_NONE, &force_mtp_port, 0,
     "Fail if can't set MTP port", NULL},

    {"rendezvous", '\0', POPT_ARG_NONE, &global_rendezvous, 0,
     "Allow others to rendezvous at MTP server", NULL},

    {"create-channel", 'r', POPT_ARG_STRING, &root_name, 0, 
     "Create a channel with the given name", "NAME"},
    {"channels-only", 'c', POPT_ARG_NONE, &is_channels_only, 0, 
     "Allow only channel advertisements in the created channel", NULL},

    {NULL, '\0', 0, NULL, 0, NULL, NULL}
  };

  poptContext ctx;
  int rc;
  int exit_status = EXIT_SUCCESS;


  /* ******************** */
  /* Initialize NLS 	  */
#ifdef ENABLE_NLS
  {
/*      gtk_set_locale(); */
    bindtextdomain (PACKAGE, LOCALEDIR);
    textdomain (PACKAGE);
  }
#endif

  /* ******************** */
  /* Parse command line	  */

  ctx = poptGetContext(NULL, argc, argv, jm_options, 0);
  poptSetOtherOptionHelp(ctx, _("[OPTION]... [URL]"));

  while ((rc = poptGetNextOpt(ctx)) > 0)
    {
      switch (rc)
	{
	case 1:	  goto help;	  break;
	case 2:	  goto usage;	  break;
	case 3:   goto version;	  break;
	}
    }

  url = poptGetArg(ctx);

  poptFreeContext(ctx);


  /* If an interface was specified, make sure it resoves to an address */
  if (interface_name)
    {
      /* Convert interface_name to InetAddr. */
      interface = gnet_inetaddr_new (interface_name, 0);
      if (!interface)
	my_error (_("Bad interface name: %s.\n"), interface_name);
    }

  /* Otherwise, auto-detect the interface */
  else
    {
      interface = jm_gnet_inetaddr_autodetect_interface ();
      if (!interface)
	my_error (_("Could not auto-detect an internet interface. "
		    "Use the --interface option to set it manually.\n"));
    }

  /* If a hostname was specified, use it */
  if (hostname_name)
    {
      hostname = g_strdup (hostname_name);
    }

  /* Autodetect the hostname */
  else
    {
      if (use_dns)
	hostname = gnet_inetaddr_get_name (interface);

      if (!hostname ||
	  !jm_gnet_inetaddr_is_internet_domainname (hostname))
	hostname = gnet_inetaddr_get_canonical_name (interface);

      g_assert (hostname);

      if (!jm_gnet_inetaddr_is_internet_domainname (hostname))
	my_error (_("Could not auto-detect hostname.  "
		    "Use the --hostname option to set it manually.\n"));
    }

  if (root_name == NULL && url == NULL)
    {
      url = JM_DEFAULT_CHANNEL;
    }

  /* Check arguments */
  if (root_name != NULL && url != NULL)
    {
      g_print ("*** You cannot be a root channel and join a root channel at the same time\n");
      exit_status = EXIT_FAILURE;
      goto usage;
    }


  /* ******************** */
  /* Set up backend	  */

  if (backend_init(hostname, interface, btp_port, force_btp_port, 
		   mtp_port, force_mtp_port, global_rendezvous))
    my_error (_("Could not start backend\n"));

  if (root_name)
    {
      root_channel = backend_channel_create (root_name, NULL, is_channels_only);
      if (root_channel)
	{
	  if (verbose)
	    g_print ("CREATE %s\n\turl: %s\n", 
		     root_channel->name, (root_channel->url? root_channel->url: "none"));
	}
      else
	g_warning("Couldn't create root channel \"%s\"", root_name);
    }
  else
    {
      root_channel = backend_channel_add (url, NULL);
      if (root_channel)
	{
	  backend_activate(root_channel);

	  if (verbose)
	    g_print ("JOIN %s\n\turl: %s\n", 
		     root_channel->name, (root_channel->url? root_channel->url: "none"));
	}
      else
	g_warning ("Couldn't join channel %s\n", url);
    }


  main_loop = g_main_new(FALSE);
  g_main_run(main_loop);

  exit(EXIT_SUCCESS);

 usage:
  poptPrintUsage(ctx, stdout, 0);
  exit(exit_status);

 help:
  g_print (_("`jmlite' - Light version of Jungle Monkey, a distributed file sharing\n"
	     "program.\n"));
  poptPrintHelp(ctx, stdout, 0);
  g_print (_("\n"
	     "How to:\n"
	     "  Join Monkey_Central, the default root channel\n"
	     "    jmlite\n"
	     "  Join some other channel\n"
	     "    jmlite jm://my-url\n"
	     "  Create my own root channel\n"
	     "    jmlite --create-root-channel \"my_root_channel\"\n"
	     "\nUse the --verbose flag if you want to see any output\n"
	     "\n"
	     "Report bugs to <jm-dev@umich.edu>\n"));
  exit(exit_status);

 version:
  g_print (_("jmlite %s\n"
	     "Written by David A. Helder\n"
	     "\n"
	     "Copyright (C) 2000 The Regents of the University of Michigan\n"
	     "This is free software; see the source for copying conditions.  There is NO\n"
	     "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"), 
	   VERSION);
  exit(exit_status);
}



void
gui_add(JMFile* jmfile)
{
  if (verbose)
    g_print (_("ADD %s (url %s)\n"), 
	     jmfile->name, 
	     (jmfile->url? jmfile->url: "none"));
}

void
gui_remove(JMFile* jmfile)
{
  if (verbose)
    g_print (_("REMOVE %s (url %s)\n"), 
	     jmfile->name, (jmfile->url? jmfile->url: "none"));
}

void
gui_update(JMFile* jmfile)
{
  if (verbose)
    g_print (_("UPDATE %s (url %s)\n"), 
	     jmfile->name, 
	     (jmfile->url? jmfile->url: "none"));
}


void gui_activate(JMFile* jmfile) {}
void gui_move(JMFile* jmfile, JMFile* new_parent, JMFile* old_parent) {}
void gui_freeze(void) {}
void gui_thaw(void) {}
