/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <ctype.h>
#include <string.h>

#include "jmp.h"



/* ************************************************************ */
/* Building							*/

JMPAnnouncement* 
jmp_announcement_new(gchar* name)
{
  JMPAnnouncement* announcement;

  announcement = g_new0(JMPAnnouncement, 1);
  if (name)
    announcement->name = g_strdup(name);

  return announcement;
}


void
jmp_announcement_delete(JMPAnnouncement* announcement)
{
  if (announcement != NULL)
    {
      GList* i;

      /* Check if we already have the attribute */
      for (i = announcement->attributes; i != NULL; i = i->next)
	{
	  JMPAttribute* attribute = (JMPAttribute*) i->data;
	  g_assert (attribute != NULL);
      
	  g_free(attribute->name);
	  g_free(attribute->value);
	  g_free(attribute);
	}

      g_list_free(announcement->attributes);
      g_free(announcement->name);
      g_free(announcement);
    }
}


void
jmp_announcement_delete_list(GList* list)
{
  GList* i;

  for (i = list; i != NULL; i = i->next)
    {
      JMPAnnouncement* announcement = (JMPAnnouncement*) i->data;
      g_assert (announcement != NULL);

      jmp_announcement_delete(announcement);
    }

  g_list_free(list);
}


void
jmp_announcement_set_attribute(JMPAnnouncement* announcement, gchar* name, gchar* value)
{
  JMPAttribute* attribute;
  GList* i;

  g_return_if_fail (announcement != NULL);
  g_return_if_fail (name != NULL);

  /* Check if we already have the attribute */
  for (i = announcement->attributes; i != NULL; i = i->next)
    {
      attribute = (JMPAttribute*) i->data;
      g_assert (attribute != NULL);
      
      if (!strcmp(attribute->name, name))
	{
	  g_free(attribute->value);
	  if (value)
	    attribute->value = g_strdup(value);
	  else
	    attribute->value = NULL;
	  return;
	}
    }

  /* Add the attribute */
  attribute = g_new0(JMPAttribute, 1);
  attribute->name = g_strdup(name);
  if (value)
    attribute->value = g_strdup(value);

  announcement->attributes = g_list_prepend(announcement->attributes, attribute);
}


gchar*
jmp_announcement_get_attribute(JMPAnnouncement* announcement, gchar* name)
{
  GList* i;

  g_return_val_if_fail (announcement != NULL, FALSE);
  g_return_val_if_fail (name != NULL, FALSE);

  for (i = announcement->attributes; i != NULL; i = i->next)
    {
      JMPAttribute* attribute = (JMPAttribute*) i->data;
      g_assert (attribute != NULL);
      
      if (!strcmp(attribute->name, name))
	return attribute->value;
    }

  return NULL;
}


gboolean
jmp_announcement_has_attribute(JMPAnnouncement* announcement, gchar* name)
{
  GList* i;

  g_return_val_if_fail (announcement != NULL, FALSE);
  g_return_val_if_fail (name != NULL, FALSE);

  for (i = announcement->attributes; i != NULL; i = i->next)
    {
      JMPAttribute* attribute = (JMPAttribute*) i->data;
      g_assert (attribute != NULL);
      
      if (!strcmp(attribute->name, name))
	return TRUE;
    }

  return FALSE;
}


void
jmp_announcement_remove_attribute(JMPAnnouncement* announcement, gchar* name)
{
  GList* i;

  g_return_if_fail (announcement != NULL);
  g_return_if_fail (name != NULL);

  /* TODO: Fix, this is inefficient */
  for (i = announcement->attributes; i != NULL; i = i->next)
    {
      JMPAttribute* attribute = (JMPAttribute*) i->data;
      g_assert (attribute != NULL);
      
      if (!strcmp(attribute->name, name))
	{
	  announcement->attributes = g_list_remove(announcement->attributes, attribute);
	  return;
	}
    }
}




/* ************************************************************ */
/* Reading							*/



#define SKIP_WHITESPACE	  { while (i < length && buffer[i] != '\0' && 			\
				   (isspace(buffer[i]) && buffer[i] != '\n')) 		\
                               ++i; }
#define CHECK_DONE	  {  if (i >= length || buffer[i] == '\0') goto done; }


static JMPAnnouncement* jmp_read_announcement(gchar** bufferp, gint* lengthp);
static JMPAttribute*	jmp_read_attribute(gchar** bufferp, gint* lengthp);


GList*
jmp_read(gchar* buffer, gint length)
{
  GList* announcements = NULL;

  while (length > 0 && buffer[0] != '\0')
    {
      JMPAnnouncement* announcement;

      announcement = jmp_read_announcement(&buffer, &length);
      if (announcement == NULL)
	break;

      announcements = g_list_prepend(announcements, announcement);
    }

  announcements = g_list_reverse(announcements);

  return announcements;
}


JMPAnnouncement*
jmp_read_announcement(gchar** bufferp, gint* lengthp)
{
  JMPAnnouncement* announcement = NULL;
  gchar* buffer = *bufferp;
  gint length = *lengthp;
  gint i = 0;
  gchar* p;

  /* Skip whitespace */
  SKIP_WHITESPACE;
  CHECK_DONE;

  /* Read in name */
  p = &buffer[i];
  while (i < length && buffer[i] != '\0' && !isspace(buffer[i])) ++i;
  if (p == &buffer[i]) goto done;
  announcement = g_new0(JMPAnnouncement, 1);
  announcement->name = g_strndup(p, &buffer[i] - p);
  CHECK_DONE;

  /* Skip whitespace */
  SKIP_WHITESPACE;
  CHECK_DONE;

  *bufferp += i;
  *lengthp -= i;
  i = 0;

  /* Read the attributes */
  while (*lengthp > 0 && (*bufferp)[0] != '\0')
    {
      JMPAttribute* attribute;

      attribute = jmp_read_attribute(bufferp, lengthp);
      if (attribute == NULL)
	break;
      
      announcement->attributes = g_list_prepend(announcement->attributes, attribute);
    }
  announcement->attributes = g_list_reverse(announcement->attributes);
  
  /* Skip whitespace and newline */
  SKIP_WHITESPACE;
  if (i < length && buffer[i] == '\n') ++i;

 done:
  *bufferp += i;
  *lengthp -= i;

  return announcement;
}



JMPAttribute*
jmp_read_attribute(gchar** bufferp, gint* lengthp)
{
  JMPAttribute* attribute = NULL;
  gchar* buffer = *bufferp;
  gint length = *lengthp;
  gint i = 0;
  gchar* p;
  gint j;

  /* Skip whitespace */
  SKIP_WHITESPACE;
  CHECK_DONE;
  if (buffer[i] == '\n')
    {
      ++i;
      goto done;
    }


  /* Read in the name */
  p = &buffer[i];
  while (i < length && buffer[i] != '\0' && !isspace(buffer[i]) && buffer[i] != '=') ++i;
  if (p == &buffer[i]) goto done;
  attribute = g_new0(JMPAttribute, 1);
  attribute->name = g_strndup(p, &buffer[i] - p);
  CHECK_DONE;

  /* Skip whitespace */
  SKIP_WHITESPACE;
  CHECK_DONE;

  /* Skip = */
  if (buffer[i] != '=') goto done;
  ++i;
  CHECK_DONE;

  /* Skip whitespace */
  SKIP_WHITESPACE;
  CHECK_DONE;

  /* Skip '"' */
  if (buffer[i] != '\"') goto done;
  ++i;
  CHECK_DONE;

  /* Read in the value */
  p = &buffer[i];
  while (i < length && buffer[i] != '\0')
    {
      /* Ignore \" */
      if (buffer[i] == '\\' && (i + 1 < length) && buffer[i + 1] == '\"')
	i += 2;
      else if (buffer[i] == '\"')
	break;

      ++i;
    }

  CHECK_DONE;
  if (buffer[i] != '\"') goto done;

  /* Save the attribute */
  attribute->value = g_new0(gchar, 1 + &buffer[i] - p);
  j = 0;
  while (p < &buffer[i])
    {
      /* Break if this is the last quote */
      if (*p == '\"')
	{
	  break;
	}
      /* Convert \" to " */
      else if (*p == '\\' && (&p[1] < &buffer[i]) && p[1] == '\"')
	{
	  attribute->value[j] = '\"';
	  p += 2;
	}
      /* Convert \\ to \ */
      else if (*p == '\\' && (&p[1] < &buffer[i]) && p[1] == '\\')
	{
	  attribute->value[j] = '\\';
	  p += 2;
	}
      /* Otherwise, it's just a normal letter */
      else
	{
	  attribute->value[j] = *p;
	  p++;
	}

      ++j;
    }
  attribute->value[j] = '\0';

  /* Move past the \ */
  ++i;

 done:
  *bufferp += i;
  *lengthp -= i;

  return attribute;
}


/* ************************************************************ */
/* Writing							*/

static void glue(gchar** dstp, gint* lenp, gchar* src1, gint length1, gchar* src2, gint length2);
static void jmp_write_attribute(JMPAttribute* attribute, gchar** bufferp, gint* lengthp);


static void
glue(gchar** dstp, gint* lenp, gchar* src1, gint length1, gchar* src2, gint length2)
{
  *dstp = g_new0(gchar, length1 + length2);
  *lenp = length1 + length2;

  memcpy(*dstp, src1, length1);
  memcpy(&(*dstp)[length1], src2, length2);

  g_assert (src1 != src2);

  if (src1) g_free(src1);
  if (src2) g_free(src2);
}


void
jmp_write_list(GList* announcements, gchar** bufferp, gint* lengthp)
{
  GList* i;

  gchar* buffer = NULL;
  gint length = 0;

  g_return_if_fail (bufferp != NULL);
  g_return_if_fail (lengthp != NULL);

  for (i = announcements; i != NULL; i = i->next)
    {
      JMPAnnouncement* announcement;
      gchar* new_buffer;
      gint  new_length;

      announcement = (JMPAnnouncement*) i->data;
      g_assert (announcement != NULL);

      jmp_write(announcement, &new_buffer, &new_length);
      glue(&buffer, &length, buffer, length, new_buffer, new_length);
    }

  *bufferp = buffer;
  *lengthp = length;
}


void
jmp_write(JMPAnnouncement* announcement, gchar** bufferp, gint* lengthp)
{
  GList* i;

  gchar* buffer;
  gint length;
  gchar* newline;

  g_return_if_fail (announcement != NULL);
  g_return_if_fail (bufferp != NULL);
  g_return_if_fail (lengthp != NULL);

  buffer = g_strdup_printf("%s ", announcement->name);
  length = strlen(buffer);

  for (i = announcement->attributes; i != NULL; i = i->next)
    {
      JMPAttribute* attribute;
      gchar* new_buffer;
      gint  new_length;

      attribute = (JMPAttribute*) i->data;
      g_assert (attribute != NULL);

      jmp_write_attribute(attribute, &new_buffer, &new_length);
      glue(&buffer, &length, buffer, length, new_buffer, new_length);
    }

  newline = g_strdup("\n");
  glue(bufferp, lengthp, buffer, length, newline, 1);
}


void
jmp_write_attribute(JMPAttribute* attribute, gchar** bufferp, gint* lengthp)
{
  gchar* buffer;

  g_return_if_fail (attribute != NULL);
  g_return_if_fail (attribute->name != NULL);
  g_return_if_fail (bufferp != NULL);
  g_return_if_fail (lengthp != NULL);

  if (attribute->value)
    {
      gchar* escaped_value;
      gint value_length;
      gint i, j;

      value_length = strlen(attribute->value);
      escaped_value = g_new0(gchar, 2 * value_length + 1);
      for (i = 0, j = 0; i < value_length; ++i)
	{	
	  if (attribute->value[i] == '\"')
	    {
	      escaped_value[j++] = '\\';
	      escaped_value[j++] = '\"';
	    }
	  else if (attribute->value[i] == '\\')
	    {
	      escaped_value[j++] = '\\';
	      escaped_value[j++] = '\\';
	    }
	  else
	    escaped_value[j++] = attribute->value[i];
	}

      buffer = g_strdup_printf("%s=\"%s\" ", attribute->name, escaped_value);
      g_free(escaped_value);
    }
  else
    {
      buffer = g_strdup_printf("%s ", attribute->name);
    }

  *bufferp = buffer;
  *lengthp = strlen(buffer);
}






/* ************************************************************ */

/*

  Test code.  To compile:

  Change #if 0

  gcc -g `glib-config --cflags --libs` jmp.c


*/
#if 0

#include <stdio.h>



gchar* tests[] = 
{
  /* Newlines */
  "\0",
  "FILE",
  "FILE\n",
  "FILE\nFILE",
  "FILE\nFILE\n",
  "\n\n",
  "  FILE\nFI\n",
  "FILE\n= = = \n",
  "FILE\n= \"= = \n",
  "FILE\n\"= \"= = \n",
  "FILE\n\"=\" \"=\\\" = \n",

  /* Quotes and spacing */
  "FILE name=\"",
  "FILE name=\"\"",
  "FILE name=\" ",

  "FILE name=\"value\"",
  "FILE name=value\"",
  "FILE name=\"value",
  "FILE name=value",

  "FILE name =\"value\"",
  "FILE name =value\"",
  "FILE name =\"value",
  "FILE name =value",

  "FILE name= \"value\"",
  "FILE name= value\"",
  "FILE name= \"value",
  "FILE name= value",

  "FILE name = \"value\"",
  "FILE name = value\"",
  "FILE name = \"value",
  "FILE name = value",

  /* EOF's */
  "FILE name=\"va\0lue\"",
  "FILE name=va\0lue\"",
  "FILE name=\"va\0lue",
  "FILE name=val\0ue",

  /* Newlines */
  "FILE name=\"val\nue\"",
  "FILE name=v\nalue\"",
  "FILE name=\"val\nue",
  "FILE name=va\nlue",

  /* Quotes and multiple lines */
  "FILE name=\"\nFILE",
  "FILE name=\"\"\nFILE",
  "FILE name=\" \nFILE",
  "FILE name=\"\nFILE\n",
  "FILE name=\"\"\nFILE\n",
  "FILE name=\" \nFILE\n",

  /* ******************** */

  /* Two attributes */

  "FILE name2=\"value\" name=\"value\"",
  "FILE name2=\"value\" name=value\"",
  "FILE name2=\"value\" name=\"value",
  "FILE name2=\"value\" name=value",

  "FILE name2=\"value\" name =\"value\"",
  "FILE name2= \"value\" name =value\"",
  "FILE name2 = \"value\" name =\"value",
  "FILE name2 = \"value\" name =value",

  "FILE name2=value name= \"value\"",
  "FILE name2=\"value name= value\"",
  "FILE name2=value\" name= \"value",
  "FILE name2=\"value\" name= value",

  /* Two attributes and two announcements */

  "FILE name2=\"value\" name=\"value\""		"\n"	  "FILE name2=\"value\" name=\"value\"",
  "FILE name2=\"value\" name=value\""		"\n"	  "FILE name2=\"value\" name=\"value\"",
  "FILE name2=\"value\" name=\"value"		"\n"	  "FILE name2=\"value\" name=\"value\"",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2=\"value\" name=\"value\"",

  "FILE name2=\"value\" name=\"value\""		"\n"	  "FILE name2=\"value\" name=value\"",
  "FILE name2=\"value\" name=value\""		"\n"	  "FILE name2=\"value\" name=value\"",
  "FILE name2=\"value\" name=\"value"		"\n"	  "FILE name2=\"value\" name=value\"",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2=\"value\" name=value\"",

  "FILE name2=\"value\" name=\"value\""		"\n"	  "FILE name2=\"value\" name=\"value",
  "FILE name2=\"value\" name=value\""		"\n"	  "FILE name2=\"value\" name=\"value",
  "FILE name2=\"value\" name=\"value"		"\n"	  "FILE name2=\"value\" name=\"value",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2=\"value\" name=\"value",

  "FILE name2=\"value\" name=\"value\""		"\n"	  "FILE name2=\"value\" name=value",
  "FILE name2=\"value\" name=value\""		"\n"	  "FILE name2=\"value\" name=value",
  "FILE name2=\"value\" name=\"value"		"\n"	  "FILE name2=\"value\" name=value",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2=\"value\" name=value",


  /* Two attributes and two announcements with early EOF's */

  "FILE name2=\"value\" name=\"value\""		"\n"	  "FILE name2=\"val\0ue\" name=\"value\"",
  "FILE name2=\"value\" name=value\""		"\n"	  "FILE name2=\"val\0ue\" name=\"value\"",
  "FILE name2=\"value\" name=\"value"		"\n"	  "FILE name2=\"val\0ue\" name=\"value\"",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2=\"val\0ue\" name=\"value\"",

  "FILE name2=\"value\" name=\"value\""		"\n"	  "FILE name2=\"val\0ue\" name=value\"",
  "FILE name2=\"value\" name=value\""		"\n"	  "FILE name2=\"val\0ue\" name=value\"",
  "FILE name2=\"value\" name=\"value"		"\n"	  "FILE name2=\"val\0ue\" name=value\"",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2=\"val\0ue\" name=value\"",

  "FILE name2=\"value\" name=\"value\""		"\n"	  "FILE name2=\"val\0ue\" name=\"value",
  "FILE name2=\"value\" name=value\""		"\n"	  "FILE name2=\"val\0ue\" name=\"value",
  "FILE name2=\"value\" name=\"value"		"\n"	  "FILE name2=\"val\0ue\" name=\"value",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2=\"val\0ue\" name=\"value",

  "FILE name2=\"value\" name=\"value\""		"\n"	  "FILE name2=\"val\0ue\" name=value",
  "FILE name2=\"value\" name=value\""		"\n"	  "FILE name2=\"val\0ue\" name=value",
  "FILE name2=\"value\" name=\"value"		"\n"	  "FILE name2=\"val\0ue\" name=value",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2=\"val\0ue\" name=value",

  /* EOFs at various boundries */
  "FILE\0",
  "FILE \0",
  "FILE name2\0", 
  "FILE name2=\0",
  "FILE name2=\"\0",
  "FILE name2=\"value\0",
  "FILE name2=\"value\"\0",
  "FILE name2=\"value\" \0",
  "FILE name2=\"value\" nam\0",
  "FILE name2=\"value\" name\0",
  "FILE name2=\"value\" name \0",
  "FILE name2=\"value\" name =\0",
  "FILE name2=\"value\" name=\0",
  "FILE name2=\"value\" name=value\0",
  "FILE name2=\"value\" name=value \0",
  "FILE name2=\"value\" name=value \n\0",
  "FILE name2=\"value\" name=value name2\0",
  "FILE name2=\"value\" name=value name2 \0",
  "FILE name2=\"value\" name=value name2=\0",
  "FILE name2=\"value\" name=value name2 =\0",
  "FILE name2=\"value\" name=value name2 = \0",
  "FILE name2=\"value\" name=value name2=va\0",
  "FILE name2=\"value\" name=value name2=value \0",
  "FILE name2=\"value\" name=value name2=value\0",
  "FILE name2=\"value\" name=value"		"\n"	  "FI\0",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE \0",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2\0",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2=\0",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2 =\0",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2 = \0",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2 = \"\0",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2=\"va\0",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2=\"value\0",
  "FILE name2=\"value\" name=value"		"\n"	  "FILE name2=\"value\"\0",

  /* Quotes */
  "FILE name2=\"va\\\"lue\"",
  "FILE name2=\"\\\"value\"",
  "FILE name2=\"value\\\"\"",
  "FILE name2=\"value\\\"",
  "FILE name2=va\\\"lue\"",
  "FILE name2=\\\"value\"",
  "FILE name2=value\" \\\"",

  NULL

};


/**

   JMP tests

 */
int main()
{
  int i;

  GList* list;
  gchar* buffer;
  gint length;

  for (i = 0; tests[i] != NULL; ++i)
    {
      list = jmp_read(tests[i], strlen(tests[i]) + 1);
      if (list)
	{
	  jmp_write_list(list, &buffer, &length);
	  if (buffer)
	    {
	      g_print ("%d: length (%d)\n", i, length);
	      fwrite(buffer, length, 1, stdout);
	      list = jmp_read(buffer, length);
	      if (!list)
		g_print ("WARNING: Could not reread %d\n", i);
	    }

	  jmp_announcement_delete_list(list);
	}
      else
	g_print ("%d: Could not read\n", i);
    }

  for (i = 0; tests[i] != NULL; ++i)
    {
      list = jmp_read(tests[i], strlen(tests[i]));
      if (list)
	{
	  jmp_write_list(list, &buffer, &length);
	  if (buffer)
	    {
	      g_print ("%d: length (%d)\n", i, length);
	      fwrite(buffer, length, 1, stdout);
	      list = jmp_read(buffer, length);
	      if (!list)
		g_print ("WARNING: Could not reread %d\n", i);
	    }
	  jmp_announcement_delete_list(list);
	}
      else
	g_print ("%d: Could not read\n", i);
    }

  return 0;
}



#endif

