/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>

#include "jmp.h"
#include "jmpio.h"
#include "gui.h"
#include "globals.h"


typedef enum {

  JMACTION_NONE, 
  JMACTION_UPDATE,
  JMACTION_REMOVE 

} JMFileAction;


static gboolean        jmpio_read_announcement(JMFile* channel, JMPAnnouncement* a);

static JMFileType      jmpio_read_type (gchar* typename);
static JMFileAction    jmpio_read_action (gchar* action_name);

static JMFile*	       jmpio_get_file (JMFile* channel, gchar* name, gchar* url,
				       JMFileType type, JMFileAction action,
				       JMFile** parentp, gboolean* did_create);

static JMPAnnouncement* jmpio_write_announcement (JMAnn* jmann, gchar* action);


static struct timeval timeofday;


/* ************************************************************ */
/* Reading functions						*/


gboolean
jmpio_read(JMFile* channel, gchar* buffer, guint length)
{
  GList* announcements;
  GList* i;
  gboolean rv = FALSE;

  gettimeofday(&timeofday, NULL);

  announcements = jmp_read(buffer, length);
  if (announcements == NULL)
    return TRUE;

  for (i = announcements; i != NULL; i = i->next)
    {
      JMPAnnouncement* announcement = (JMPAnnouncement*) i->data;
      g_assert (announcement != NULL);

      if (jmpio_read_announcement(channel, announcement))
	{
	  g_warning ("Read bad announcement on channel %s\n", channel->name);
	  rv = TRUE;
	}
    }

  jmp_announcement_delete_list(announcements);

  return FALSE;
}



gboolean
jmpio_read_announcement(JMFile* channel, JMPAnnouncement* a)
{
  JMFileAction action = JMACTION_UPDATE;
  gchar* type_str;
  JMFileType type;
  gchar* fullname;
  gchar* url;

  JMFile* jmfile = NULL;
  JMFile* parent = NULL;

  gboolean should_add_gui = FALSE;
  gboolean should_update_gui = FALSE;

  g_return_val_if_fail(channel != NULL, TRUE);
  g_return_val_if_fail(a != NULL, TRUE);

  /* Read the action */
  if (a->name == NULL)
    return TRUE;
  action = jmpio_read_action(a->name);
  if (action == JMACTION_NONE)
    action = JMACTION_UPDATE;	/* Assume update if action is weird */

  /* Read the type */
  type_str = jmp_announcement_get_attribute(a, "type");
  if (type_str != NULL)
    type = jmpio_read_type(type_str);
  if (type == JMFILE_NONE)
    return TRUE;

  /* Read the name */
  fullname = jmp_announcement_get_attribute(a, "name");
  if (fullname == NULL)
    return TRUE;

  /* Read the URL */
  url = jmp_announcement_get_attribute(a, "url");

  jmfile = jmpio_get_file (channel, fullname, url, type, action, &parent, &should_add_gui);

  /* Action is REMOVE */
  if (action == JMACTION_REMOVE)
    {
      DEBUGP ("REMOVE %s\n", fullname);

      /* We don't need to remove non-existant files */
      if (jmfile != NULL)
	{
	  /* Only remove if it is not local */
	  if (!jmfile->is_local)
	    jmfile_unlink (jmfile, parent);
	  else
	    g_warning ("Someone attempted to remove my file %s\n", jmfile->name);
	}
    }

  /* Action is ADD or UPDATE */
  else
    {
      DEBUGP ("UPDATE %s\n", fullname);

      /* We should have this file (either had or created) */
      if (jmfile == NULL)
	return TRUE;

      /* If it's not a directory, it should have a URL */
      if (type != JMFILE_DIR && url == NULL)
	{
	  g_warning ("Received announcment without URL\n");
	  return FALSE;
	}

      /* Ignore if the types don't match.  This can happen if A added
         directory "foo" at the same time B adds channel "foo".
         Really, we should change the name slightly.  (ie, append a
         number) */
      if (jmfile->type != type)
	{
	  g_warning ("JMFile name collision (name is %s, type is %d, other type is %d)\n",
		     jmfile->name, jmfile->type, type);
	  return FALSE;
	}

      /* If we're a channels-only channel and we get something that
         isn't a channel or a search, reject it. */
      if (channel->is_channels_only && 
	  (type != JMFILE_CHANNEL && type != JMFILE_SEARCH))
	return FALSE;

      /* Update timeofday */
      memcpy(&jmfile->last_update, &timeofday, sizeof(timeofday));

      /* Update the URL */
      if (url != NULL)
	{
	  if (jmfile->url == NULL || strcmp(jmfile->url, url) != 0)
	    {
	      g_free (jmfile->url);
	      jmfile->url = g_strdup(url);
	      should_update_gui = TRUE;

	      /* If the jmfile was marked as down, change to up -
                 something may have been fixed. */
	      if (jmfile->status == JMFILE_STATUS_DOWN)
		jmfile->status = JMFILE_STATUS_NONE;
	    }
	}

      if (jmfile->type == JMFILE_FILE)
	{
	  gchar* length_str;

	  /* Update length */
	  length_str = jmp_announcement_get_attribute(a, "length");
	  if (length_str != NULL)
	    {
	      guint length = atoi(length_str);
	      if (length != jmfile->length)
		{
		  jmfile->length = length;
		  should_update_gui = TRUE;
		}
	    }

	  /* Update SHA - we don't overwrite old SHA's (should we?) */
	  if (!jmfile->sha)
	    {
	      gchar* sha_str;

	      sha_str = jmp_announcement_get_attribute(a, "sha");
	      if (sha_str)
		{
		  jmfile->sha = gnet_sha_new_string (sha_str);
		  should_update_gui = TRUE;
		}
	    }
	}


      /* Update 'is_channels_only' */
      if (jmfile->type == JMFILE_CHANNEL && jmfile->is_channels_only == FALSE)
	{
	  gchar* ico;

	  ico = jmp_announcement_get_attribute(a, "channels_only");

	  /* TODO: Consider implications of it changing to false */
	  if ((ico != NULL) && strcmp(ico, "true") == 0)
	    {
	      jmfile->is_channels_only = TRUE;
	      should_update_gui = TRUE;
	    }
	}

      /* Update/add to GUI */
      if (should_add_gui)
	gui_add (jmfile);
      else if (should_update_gui)
	gui_update (jmfile);
    }

  return FALSE;
}



static JMFileAction
jmpio_read_action(gchar* action_name)
{
  g_return_val_if_fail (action_name != NULL, JMACTION_NONE);

  switch (action_name[0])
    {
    case 'a':
      if (strcmp(action_name, "add") == 0)
	return JMACTION_UPDATE;	/* TODO: phase out "add" */
      break;

    case 'r':
      if (strcmp(action_name, "remove") == 0)
	return JMACTION_REMOVE;
      break;

    case 'u':
      if (strcmp(action_name, "update") == 0)
	return JMACTION_UPDATE;
      break;
    }

  return JMACTION_NONE;
}


static JMFileType
jmpio_read_type(gchar* typename)
{
  g_return_val_if_fail (typename != NULL, JMFILE_NONE);

  switch (typename[0])
    {
    case 'c':
      if (!strcmp(typename, "channel"))
	return JMFILE_CHANNEL;
      break;

    case 'd':
      if (!strcmp(typename, "directory"))
	return JMFILE_DIR;
      break;

    case 'f':
      if (!strcmp(typename, "file"))
	return JMFILE_FILE;
      break;

    case 's':
      if (!strcmp(typename, "search"))
	return JMFILE_SEARCH;
      break;

    case 'u':
      if (!strcmp(typename, "url"))
	return JMFILE_URL;
      break;
    }

  return JMFILE_NONE;
}



/**

   jmpio_get_file

   Given the name and type of a file, create any missing ancestors of
   the file and possibly the file itself.  

   Caller is responsible for adding the file to the GUI.  This is so
   it can finish initializing it.  We set the did_create flag if we
   created the file.

 */
static JMFile*
jmpio_get_file (JMFile* channel, gchar* name, gchar* url, 
		JMFileType type, JMFileAction action, 
		JMFile** parentp, gboolean* did_create)
{
  JMFile* jmfile = NULL;
  JMFile* parent = NULL;
  gchar** tokens;
  guint i = 0;

  g_return_val_if_fail (channel != NULL, NULL);
  g_return_val_if_fail (name != NULL, NULL);
  g_return_val_if_fail (parentp != NULL, NULL);
  g_return_val_if_fail (did_create != NULL, NULL);

  *did_create = FALSE;
  *parentp = NULL;

  /* Skip the initial /, if there is one */
  if (name[0] == '/')
    name = &name[1];

  /* Break the name up into tokens */
  tokens = g_strsplit (name, "/", 0);

  if (tokens == NULL || tokens[0] == NULL)
    goto done;

  for (i = 0; tokens[i] != NULL; ++i)
    {
      JMFile* old_file;

      /* The first token refers to the channel */
      if (i == 0)
	old_file = channel;

      /* Otherwise, check if we already have this file */
      else
	old_file = jmfile_get_child_by_name (parent, tokens[i]);

      /* If we don't have this one, create it */
      if (old_file == NULL)
	{
	  /* Don't bother if we're going to remove the file */
	  if (action == JMACTION_REMOVE)
	    {
	      jmfile = NULL;
	      break;
	    }

	  /* Create a directory if there are more tokens */
	  if (tokens[i+1] != NULL)
	    {
	      jmfile = jmfile_new (tokens[i], JMFILE_DIR, NULL);
	    }

	  /* Else, this is the last token */
	  else
	    {
	      /* Non-directories must have URL's. */
	      if (type != JMFILE_DIR && url == NULL)
		{
		  jmfile = NULL;
		  break;
		}

	      /* Check if we already have it  */
	      jmfile = NULL;
	      if (url)
		jmfile = backend_get (url);

	      if (!jmfile)
		{
		  jmfile = jmfile_new (tokens[i], type, url);

		  /* Add to our hashtable */
		  if (url)
		    backend_put (jmfile);
		}
	    }

	  jmfile_link (jmfile, parent);

	  memcpy(&jmfile->last_update, &timeofday, sizeof(timeofday));

	  /* Add to GUI now if not the last file */
	  if (tokens[i+1] != NULL)
	    gui_add (jmfile);
	  else
	    *did_create = TRUE;
	}

      /* Otherwise, use the old one */
      else
	{
	  jmfile = old_file;
	}

      /* Break if we're done */
      if (tokens[i+1] == NULL)
	break;

      /* There are more tokens - save the parent */
      parent = jmfile;
    }

 done:
  g_free(tokens);

  *parentp = parent;
  
  return jmfile;
}



/* ************************************************************ */
/* Writing functions						*/


void
jmpio_write(JMAnn* jmann, gchar* action, gchar** bufferp, guint* lengthp)
{
  JMPAnnouncement* a;

  g_return_if_fail (jmann != NULL);
  g_return_if_fail (action != NULL);
  g_return_if_fail (bufferp != NULL);
  g_return_if_fail (lengthp != NULL);

  DEBUGP ("jmpip_write %s\n", jmann->fullname);

  a = jmpio_write_announcement (jmann, action);
  if (a != NULL)
    {
      jmp_write (a, bufferp, lengthp);
      jmp_announcement_delete (a);
    }
  else
    {
      *bufferp = NULL;
      *lengthp = 0;
    }
}



JMPAnnouncement*
jmpio_write_announcement(JMAnn* jmann, gchar* action)
{
  gchar* fullname = NULL;
  JMPAnnouncement* a;
  JMFile* jmfile;

  g_return_val_if_fail (jmann, NULL);
  g_return_val_if_fail (jmann->fullname, NULL);
  g_return_val_if_fail (jmann->file, NULL);
  g_return_val_if_fail (action, NULL);

  jmfile = jmann->file;

  DEBUGP("jmpio_write_announcement %s\n", jmann->fullname);

  /* Create a new announcement */
  a = jmp_announcement_new (action);

  /* Set the name */
  jmp_announcement_set_attribute (a, "name", jmann->fullname);
  g_free (fullname);

  /* Set the type and other type-specific attributes */
  switch (jmfile->type)
    {
    case JMFILE_FILE:
      {
	gchar length_buffer[32];

	g_return_val_if_fail (jmfile->url, NULL);

	jmp_announcement_set_attribute(a, "type", "file");
	jmp_announcement_set_attribute(a, "url", jmfile->url);
	
	snprintf(length_buffer, sizeof(length_buffer), "%d", jmfile->length);
	jmp_announcement_set_attribute(a, "length", length_buffer);

	if (jmfile->sha)
	  {
	    guchar sha[2 * GNET_SHA_HASH_LENGTH + 1];

	    gnet_sha_copy_string (jmfile->sha, sha);
	    sha[2 * GNET_SHA_HASH_LENGTH] = '\0';

	    jmp_announcement_set_attribute(a, "sha", sha);
	  }

	break;
      }

    case JMFILE_CHANNEL:
      {
	g_return_val_if_fail (jmfile->url, NULL);

	jmp_announcement_set_attribute(a, "type", "channel");
	jmp_announcement_set_attribute(a, "url", jmfile->url);

	if (jmfile->is_channels_only)
	  jmp_announcement_set_attribute(a, "channels_only", "true");
	else
	  jmp_announcement_set_attribute(a, "channels_only", "false");

	break;
      }

    case JMFILE_DIR:
      {
	jmp_announcement_set_attribute(a, "type", "directory");
	break;
      }

    case JMFILE_SEARCH:
      {
	g_return_val_if_fail (jmfile->url, NULL);

	jmp_announcement_set_attribute(a, "type", "search");
	jmp_announcement_set_attribute(a, "url", jmfile->url);

	break;
      }

    case JMFILE_URL:
      {
	g_return_val_if_fail (jmfile->url, NULL);

	jmp_announcement_set_attribute(a, "type", "url");
	jmp_announcement_set_attribute(a, "url", jmfile->url);

	break;
      }

    default:
      g_assert_not_reached();

    }

  return a;
}
