/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */



#include <glib.h>
#include <gnet/gnet.h>
#include <string.h>
#include <stdlib.h>

#include "jmsearch.h"
#include "jmp.h"
#include "jmgnet.h"
#include "util.h"


#define SEARCH_TIMEOUT	(2 * 60 * 1000)	/* Searches timeout in 2 minutes */
#define MAX_BUFFER_IN	4096		/* A line shouldn't be longer than this */


typedef struct _JMSearchState
{
  JMGURL* url;
  gchar* query;
  JMGConn* conn;
  
  GList* results;

  JMSearchFunc func;

} JMSearchState;
  

static gboolean jmsearch_conn_func (JMGConn* conn, JMGConnStatus status, 
				    gchar* buffer, gint length, gpointer user_data);

static void jmsearchstate_delete (JMSearchState* state);



JMSearchID
jmsearch_search (JMGURL* url, gchar* query, JMSearchFunc func)
{
  JMSearchState* state = NULL;
  JMPAnnouncement* a;
  gchar* buffer;
  guint length;

  g_return_val_if_fail (url, NULL);
  g_return_val_if_fail (query, NULL);
  g_return_val_if_fail (query[0], NULL);
  g_return_val_if_fail (func, NULL);
  
  state = g_new0(JMSearchState, 1);
  state->url = jm_url_clone (url);
  state->query = g_strdup(query);
  state->func = func;

  state->conn = jm_conn_new (url->hostname, url->port ? url->port : SEARCH_SERVER_PORT, 
			     jmsearch_conn_func, state);
  g_return_val_if_fail (state->conn, NULL);

  jm_conn_connect (state->conn, SEARCH_TIMEOUT);

  a = jmp_announcement_new ("SEARCH");
  jmp_announcement_set_attribute (a, "query", state->query);

  jmp_write (a, &buffer, &length);
  g_return_val_if_fail (buffer, NULL);

  jm_conn_write (state->conn, buffer, length, SEARCH_TIMEOUT);

  jmp_announcement_delete (a);

  return state;
}


static gboolean
jmsearch_conn_func (JMGConn* conn, JMGConnStatus status, gchar* buffer, gint length, gpointer user_data)
{
  JMSearchState* state = (JMSearchState*) user_data;

  g_return_val_if_fail (state, FALSE);

  switch (status)
    {
    case GNET_CONN_STATUS_CONNECT:
      {
	/* Read reponse */
	jm_conn_readline (conn, NULL, MAX_BUFFER_IN, SEARCH_TIMEOUT);
	break;
      }

    case GNET_CONN_STATUS_READ:
      {
	GList* as;
	JMPAnnouncement* a;

	as = jmp_read (buffer, length);
	if (as == NULL)
	  {
	    state->func (JMSEARCH_ERROR, state->url, state->query, NULL);
	    jmsearchstate_delete (state);
	    return FALSE;
	  }

	a = (JMPAnnouncement*) as->data;
	g_return_val_if_fail (a, FALSE);

	if (a->name != NULL && !strcmp(a->name, "MATCH"))
	  {
	    gchar* name;
	    gchar* url;

	    name = jmp_announcement_get_attribute (a, "name");
	    url = jmp_announcement_get_attribute (a, "url");

	    if (name && url)
	      {
		JMGURL* gurl;

		gurl = jm_url_new (url);
		if (gurl)
		  {
		    gchar* length;
		    JMSearchResult* result;

		    result = g_new0 (JMSearchResult, 1);
		    result->name = g_strdup(name);
		    result->url = gurl;

		    length = jmp_announcement_get_attribute (a, "length");
		    if (length)
		      result->length = atoi (length);

		    state->results = g_list_prepend (state->results, result);
		  }
	      }
	    /* The announcement owns name and url */
	  }

	jmp_announcement_delete_list(as);

	return TRUE;
	break;
      }

    case GNET_CONN_STATUS_WRITE:
      {
	g_free (buffer);
	break;
      }

    case GNET_CONN_STATUS_CLOSE:
      {
	state->results = g_list_reverse (state->results);
	state->func (JMSEARCH_OK, state->url, state->query, state->results);
	jmsearchstate_delete (state);

	break;
      }

    case GNET_CONN_STATUS_TIMEOUT:
    case GNET_CONN_STATUS_ERROR:
      {
	state->func (JMSEARCH_ERROR, state->url, state->query, NULL);
	jmsearchstate_delete (state);

	break;
      }
    }

  return FALSE;
}



static void
jmsearchstate_delete(JMSearchState* state)
{
  GList* i;

  g_return_if_fail (state != NULL);

  jm_url_delete (state->url);
  g_free (state->query);
  jm_conn_delete (state->conn, TRUE);

  for (i = state->results; i != NULL; i = i->next)
    {
      JMSearchResult* result = (JMSearchResult*) i->data;

      g_free(result->name);
      g_free(result->url);
      g_free(result);
    }
  g_list_free (state->results);

  g_free(state);
}



void
jmsearch_cancel(JMSearchID id)
{
  JMSearchState* state = (JMSearchState*) id;

  g_return_if_fail (state);

  jmsearchstate_delete (state);
}
