/* Jungle Monkey Search Server
 * Copyright (C) 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>

#ifdef JM_ENABLE_GNOME
#include <popt-gnome.h>
#else
#include <popt.h>
#endif

#include "jmintl.h"
#include "util.h"
#include "globals.h"
#include "jmp.h"
#include "jmgnet.h"

#include "jmsearch.h"


static void jmsearch_cb(JMSearchStatus status, JMGURL* url, 
			gchar* str, /* const */ GList* results);


/* ******************** */
/* Globals		*/

int   			debug_level = 0;
static int   		verbose = 0;



int main (int argc, char* argv[])
{
  gchar* query = NULL;
  gchar* url = NULL;
  JMGURL* gurl;
  GMainLoop* main_loop;
  JMSearchID search_id;
  

  struct poptOption jm_options[] = 
  {
    {"help", 'h', POPT_ARG_NONE, NULL, 1, 
     "Show this help message", NULL},
    {"usage", '\0', POPT_ARG_NONE, NULL, 2, 
     "Display brief usage message", NULL},
    {"version", 0, POPT_ARG_NONE, NULL, 3,
     "Output verson information and exit", NULL},
    {"verbose", 'v', POPT_ARG_NONE, &verbose, 0,
     "Print lots of stuff", NULL},

    {"debug-level", 'd', POPT_ARG_INT, &debug_level, 0, 
     "Set the debug level", "LEVEL"},

    {NULL, '\0', 0, NULL, 0, NULL, NULL}
  };

  poptContext ctx;
  int rc;
  int exit_status = EXIT_SUCCESS;


  /* ******************** */
  /* Initialize NLS 	  */
#ifdef ENABLE_NLS
  {
/*      gtk_set_locale(); */
    bindtextdomain (PACKAGE, LOCALEDIR);
    textdomain (PACKAGE);
  }
#endif

  /* ******************** */
  /* Parse command line	  */

  ctx = poptGetContext (NULL, argc, argv, jm_options, 0);
  poptSetOtherOptionHelp (ctx, _("[OPTION]... [QUERY] [URL]"));

  while ((rc = poptGetNextOpt(ctx)) > 0)
    {
      switch (rc)
	{
	case 1:	  goto help;	  break;
	case 2:	  goto usage;	  break;
	case 3:   goto version;	  break;
	}
    }

  query = poptGetArg (ctx);
  url = poptGetArg (ctx);

  poptFreeContext (ctx);

  if (!query)
    my_error (_("You must specify a query string.\n"));

  if (!url)
    my_error (_("You must specify a URL.\n"));

  gurl = jm_url_new (url);
  if (!gurl)
    my_error (_("Malformed URL: %s\n"), url);


  /* ******************** */
  /* Do whatever	  */

  search_id = jmsearch_search (gurl, query, jmsearch_cb);
  g_assert (search_id);


  /* ******************** */
  /* Start the main loop  */

  main_loop = g_main_new (FALSE);
  g_main_run (main_loop);

  exit (EXIT_SUCCESS);

 usage:
  poptPrintUsage (ctx, stdout, 0);
  exit (exit_status);

 help:
  g_print (_("`jmsearchclient' - JM Search Client\n"));
  poptPrintHelp (ctx, stdout, 0);
  g_print (_("\n"
	     "Report bugs to <jm-dev@umich.edu>\n"));
  exit (exit_status);

 version:
  g_print (_("jmsearchclient %s\n"
	     "Written by David A. Helder\n"
	     "\n"
	     "Copyright (C) 2000 The Regents of the University of Michigan\n"
	     "This is free software; see the source for copying conditions.  There is NO\n"
	     "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
	     VERSION);
  exit (exit_status);
}



static void
jmsearch_cb(JMSearchStatus status, JMGURL* url, gchar* str, /* const */ GList* results)
{
  if (status != JMSEARCH_OK)
    {
      g_print ("Search error.\n");
    }

  else if (results == NULL)
    {
      g_print ("No matches for %s.\n", str);
    }

  else
    {
      GList* i;

      for (i = results; i != NULL; i = i->next)
	{
	  JMSearchResult* result;
	  gchar* url_str;
	  
	  result = (JMSearchResult*) i->data;
	  g_return_if_fail (result->url);
	  g_return_if_fail (result->name);

	  url_str = jm_url_get_nice_string (result->url);
	  g_assert (url_str);

	  g_print ("%s\t%s\t%d\n", result->name, url_str, result->length);
	  g_free (url_str);
	}
    }

  exit (0);
}
