/* Jungle Monkey Search Server
 * Copyright (C) 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>

#ifdef JM_ENABLE_GNOME
#include <popt-gnome.h>
#else
#include <popt.h>
#endif

#include "jmintl.h"
#include "backend.h"
#include "util.h"
#include "globals.h"
#include "jmp.h"
#include "jmgnet.h"
#include "jmsearch.h"
#include "btp.h"
#include "b_conn_accept.h"


#define MAX_RESULTS 		200
#define PRINT_TIME_TIMEOUT	5 * 60 * 1000	/* 5 minutes */

#define CONNECTION_TIMEOUT	2 * 60 * 1000	/* 2 minutes */
#define MAX_BUFFER_IN		2048		/* Don't read more than 2048 bytes */


#define ANY_IO_CONDITION  (G_IO_IN|G_IO_OUT|G_IO_PRI|G_IO_ERR|G_IO_HUP|G_IO_NVAL)

static void server_start(void);


/* ******************** */
/* Globals		*/

static int   		verbose = 0;
static JMFile* 		root_channel = NULL;
static gchar* 		server_name = NULL;

static GList* 		files = NULL;



int main(int argc, char* argv[])
{
  gchar* url = NULL;
  GMainLoop* main_loop;

  gchar* hostname = NULL;
  gchar* hostname_name = NULL;

  GInetAddr* interface = NULL;
  gchar* interface_name = NULL;
  gboolean use_dns = FALSE;

  gint btp_port = 0;
  gboolean force_btp_port = FALSE;

  gint mtp_port = 0;
  gboolean force_mtp_port = FALSE;

  struct poptOption jm_options[] = 
  {
    {"help", 'h', POPT_ARG_NONE, NULL, 1, 
     "Show this help message", NULL},
    {"usage", '\0', POPT_ARG_NONE, NULL, 2, 
     "Display brief usage message", NULL},
    {"version", 0, POPT_ARG_NONE, NULL, 3,
     "Output verson information and exit", NULL},
    {"verbose", 'v', POPT_ARG_NONE, &verbose, 0,
     "Print lots of stuff", NULL},

    {"hostname", '\0', POPT_ARG_STRING, &hostname_name, 0,
     "Hostname or address of this host", "HOSTNAME"},
    {"use-dns", '\0', POPT_ARG_NONE, &use_dns, 0,
     "Use DNS to figure out my hostname", NULL},

    {"interface", '\0', POPT_ARG_STRING, &interface_name, 0,
     "Interface address for listening socket", "ADDR"},

    {"btp-port", '\0', POPT_ARG_STRING, &btp_port, 0,
     "Port for BTP listening socket", "PORT"},
    {"force-btp-port", '\0', POPT_ARG_NONE, &force_btp_port, 0,
     "Fail if can't set MTP port", NULL},

    {"mtp-port", '\0', POPT_ARG_STRING, &mtp_port, 0,
     "Port for MTP listening socket", "PORT"},
    {"force-mtp-port", '\0', POPT_ARG_NONE, &force_mtp_port, 0,
     "Fail if can't set MTP port", NULL},

    {"name", 0, POPT_ARG_STRING, &server_name, 0, 
     "Name of the search server (ex: `mofo - MI, USA')", "NAME"},

    {"debug-level", 'd', POPT_ARG_INT, &debug_level, 0, 
     "Set the debug level", "LEVEL"},

    {NULL, '\0', 0, NULL, 0, NULL, NULL}
  };

  poptContext ctx;
  int rc;
  int exit_status = EXIT_SUCCESS;


  /* ******************** */
  /* Initialize NLS 	  */
#ifdef ENABLE_NLS
  {
/*      gtk_set_locale(); */
    bindtextdomain (PACKAGE, LOCALEDIR);
    textdomain (PACKAGE);
  }
#endif

  /* ******************** */
  /* Parse command line	  */

  ctx = poptGetContext(NULL, argc, argv, jm_options, 0);
  poptSetOtherOptionHelp(ctx, _("[OPTION]... [URL]"));

  while ((rc = poptGetNextOpt(ctx)) > 0)
    {
      switch (rc)
	{
	case 1:	  goto help;	  break;
	case 2:	  goto usage;	  break;
	case 3:   goto version;	  break;
	}
    }

  url = poptGetArg(ctx);

  poptFreeContext(ctx);

  /* If an interface was specified, make sure it resoves to an address */
  if (interface_name)
    {
      /* Convert interface_name to InetAddr. */
      interface = gnet_inetaddr_new (interface_name, 0);
      if (!interface)
	my_error (_("Bad interface name: %s.\n"), interface_name);
    }

  /* Otherwise, auto-detect the interface */
  else
    {
      interface = jm_gnet_inetaddr_autodetect_interface ();
      if (!interface)
	my_error (_("Could not auto-detect an internet interface. "
		    "Use the --interface option to set it manually.\n"));
    }

  /* If a hostname was specified, use it */
  if (hostname_name)
    {
      hostname = g_strdup (hostname_name);
    }

  /* Autodetect the hostname */
  else
    {
      if (use_dns)
	hostname = gnet_inetaddr_get_name (interface);

      if (!hostname ||
	  !jm_gnet_inetaddr_is_internet_domainname (hostname))
	hostname = gnet_inetaddr_get_canonical_name (interface);

      g_assert (hostname);

      if (!jm_gnet_inetaddr_is_internet_domainname (hostname))
	my_error (_("Could not auto-detect hostname.  "
		    "Use the --hostname option to set it manually.\n"));
    }

  if (server_name == NULL)
    {
      server_name = gnet_inetaddr_gethostname();
      g_assert (server_name);
    }

  if (url == NULL)
    url = JM_DEFAULT_CHANNEL;


  /* ******************** */
  /* Set up backend	  */

  if (backend_init(hostname, interface, btp_port, force_btp_port, 
		   mtp_port, force_mtp_port, FALSE))
    my_error (_("Could not start backend\n"));

  root_channel = backend_channel_add (url, NULL);
  if (root_channel)
    {
      backend_activate(root_channel);

      if (verbose)
	g_print (_("JOIN %s\n\turl: %s\n"), 
		 root_channel->name, (root_channel->url? root_channel->url: "none"));
    }
  else
    my_error (_("Could not join channel %s\n"), url);


  /* ******************** */
  /* Start up the server  */

  server_start();


  /* ******************** */
  /* Start the main loop  */

  main_loop = g_main_new(FALSE);
  g_main_run(main_loop);

  exit(EXIT_SUCCESS);

 usage:
  poptPrintUsage(ctx, stdout, 0);
  exit(exit_status);

 help:
  g_print (_("`jmsearchserver' - Jungle Monkey Search Server\n"));
  poptPrintHelp(ctx, stdout, 0);
  g_print (_("\n"
	     "Report bugs to <jm-dev@umich.edu>\n"));
  exit(exit_status);

 version:
  g_print (_("jmsearchserver %s\n"
	     "Written by David A. Helder\n"
	     "\n"
	     "Copyright (C) 2000 The Regents of the University of Michigan\n"
	     "This is free software; see the source for copying conditions.  There is NO\n"
	     "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
	     VERSION);
  exit(exit_status);
}


/* **************************************** */

static gboolean print_time_cb (gpointer data);
static void server_func (JMGServer* server, JMGServerStatus status, 
			 JMGConn* conn, gpointer user_data);
static gboolean conn_func (JMGConn* conn, JMGConnStatus status, 
			   gchar* buffer, gint length, gpointer user_data);
static GList* find_matches(gchar* word);




static void
server_start(void)
{
  JMGServer* server;
  gchar* url;

  g_assert (backend_hostname != NULL);

  server = jm_server_new (backend_interface, SEARCH_SERVER_PORT, FALSE, server_func, NULL);
  if (!server)
    my_error (_("Could not start search server.\n"));

  url = g_strdup_printf("jmsearch://%s:%d", backend_hostname, server->port);

  if (verbose)
    g_print (_("search server started: %s\n"), url);

  if (verbose)
    {
      g_timeout_add(PRINT_TIME_TIMEOUT, print_time_cb, NULL);
      print_time_cb(NULL);
    }

  if (backend_search_create (server_name, url, root_channel) == NULL)
    g_warning (_("Could not add search to root channel\n"));

  g_free(url);	
}


static gboolean
print_time_cb (gpointer data)
{
  time_t t;
  char* time_str;

  t = time(NULL);
  t = mktime(gmtime(&t));
  time_str = ctime(&t);		/* don't free string */
  time_str[strlen(time_str) - 1] = '\0';

  g_print ("time: %s UTC\n", time_str);

  return TRUE;
}


static void
server_func (JMGServer* server, JMGServerStatus status, JMGConn* conn, gpointer user_data)
{
  g_return_if_fail (status == GNET_SERVER_STATUS_CONNECT);

  conn->func = conn_func;
  jm_conn_readline (conn, NULL, MAX_BUFFER_IN, CONNECTION_TIMEOUT);
}


static gboolean
conn_func (JMGConn* conn, JMGConnStatus status, gchar* buffer, gint length, gpointer user_data)
{
  switch (status)
    {
    case GNET_CONN_STATUS_READ:
      {
	GList* as_in = NULL;
	JMPAnnouncement* a;
	gchar* query = NULL;

	GList* matches = NULL;
	GList* i;
	GList* as_out = NULL;

	gchar* buffer_out = NULL;
	guint length_out;


	/* Read the search requests */
	as_in = jmp_read (buffer, length);
	if (as_in == NULL)
	  goto error;

	/* For now, we just assume there's one */
	a = (JMPAnnouncement*) as_in->data;

	/* It should be a search */
	if (a->name == NULL || strcmp(a->name, "SEARCH"))
	  goto error;

	/* It should have a query (do not delete the query!) */
	query = jmp_announcement_get_attribute (a, "query");
	if (query == NULL)
	  goto error;

	if (verbose)
	  g_print ("SEARCH \"%s\" (%s, %d)\n", query, conn->hostname, conn->port);

	matches = find_matches (query);
	for (i = matches; i != NULL; i = i->next)
	  {
	    JMPAnnouncement* new_a;
	    JMFile* file;
	    gchar length[20];

	    file = (JMFile*) i->data;
	  
	    new_a = jmp_announcement_new ("MATCH");
	    jmp_announcement_set_attribute (new_a, "name",  file->name);
	    jmp_announcement_set_attribute (new_a, "url",  file->url);
	    g_snprintf (length, sizeof(length), "%d", file->length);
	    jmp_announcement_set_attribute (new_a, "length",  length);

	    as_out = g_list_prepend (as_out, new_a);
	  }
	g_list_free (matches);

	as_out = g_list_reverse (as_out);
	if (!as_out)
	  goto error;

	jmp_write_list(as_out, &buffer_out, &length_out);

	jm_conn_write (conn, buffer_out, length_out, CONNECTION_TIMEOUT);
  
      done:
	if (as_in)
	  jmp_announcement_delete_list(as_in);

	if (as_out)
	  jmp_announcement_delete_list(as_out);

	/* For now, we don't try to read another line */
	return FALSE;
	break;

      error:
	jm_conn_delete (conn, TRUE);
	goto done;

	break;
      }

    case GNET_CONN_STATUS_WRITE:
      {
	g_free (buffer);
	/* fall through */
      }

    case GNET_CONN_STATUS_CONNECT:
    case GNET_CONN_STATUS_CLOSE:
    case GNET_CONN_STATUS_TIMEOUT:
    case GNET_CONN_STATUS_ERROR:
      {
	jm_conn_delete (conn, TRUE);
      }
    }

  return FALSE;
}



GList*
find_matches(gchar* word)
{
  GList* i;
  GList* matches = NULL;

  int num_results = 0;

  if (word == NULL || word[0] == '\0')
    return NULL;

  for (num_results = 0, i = files; 
       num_results < MAX_RESULTS && i != NULL; 
       i = i->next)
    {
      JMFile* file = (JMFile*) i->data;

      if (file->name == NULL || file->url == NULL)
	continue;

      if (strcasestr (file->name, word))
	{
	  matches = g_list_prepend (matches, file);
	  ++num_results;
	}
    }

  return matches;
}





/* **************************************** */


void
gui_add(JMFile* jmfile)
{
  if (verbose)
    g_print (_("ADD %s (%s)\n"), jmfile->name, jmfile->url);

  /* Join the channel */
  if (jmfile->type == JMFILE_CHANNEL && jmfile->btsmp == NULL)
    {
      backend_activate(jmfile);

      if (verbose)
	g_print (_("JOIN %s\n"), jmfile->name);
    }

  /* Add to list (files only) */
  if (jmfile->type == JMFILE_FILE)
    files = g_list_prepend(files, jmfile);
}


void
gui_remove(JMFile* jmfile)
{
  if (verbose)
    g_print (_("REMOVE %s (%s)\n"), jmfile->name, jmfile->url);

  /* Remove from list */
  files = g_list_remove(files, jmfile);
}


void gui_activate(JMFile* jmfile) {}
void gui_update(JMFile* jmfile) {}
void gui_move(JMFile* jmfile, JMFile* new_parent, JMFile* old_parent) {}
void gui_freeze(void) {}
void gui_thaw(void) {}

