/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "mtp_types.h"


/* Create a new MTP Troop. */
MtpTroop*
mtp_troop_new (gchar* name)
{
  MtpTroop* troop;

  g_return_val_if_fail (name, NULL);

  troop = g_new0(MtpTroop, 1);

  troop->name = g_strdup (name);
  troop->mirrors = g_ptr_array_new();

  return troop;
}


/* Update an MTP Troop given a mirror's hostname and port.  If the
   mirror does not exist, the mirror is added. */
void
mtp_troop_update (MtpTroop* troop, gchar* hostname, gint port, gboolean is_local)
{
  guint i;
  MtpMirror* mirror = NULL;

  g_return_if_fail (troop);
  g_return_if_fail (hostname);

  /* Find the mirror */
  for (i = 0; i < troop->mirrors->len; ++i)
    {
      MtpMirror* m = (MtpMirror*) g_ptr_array_index (troop->mirrors, i);

      if ((port == m->port) && !strcmp (hostname, m->hostname))
	{
	  mirror = m;
	  break;
	}
    }

  /* If we could not find the mirror, create a new one. */
  if (!mirror)
    {
      mirror = g_new0 (MtpMirror, 1);
      mirror->hostname = g_strdup (hostname);
      mirror->port = port;
      mirror->is_local = is_local;

      g_ptr_array_add (troop->mirrors, mirror);
    }

  /* Update the time the mirror was last updated. */
  if (!is_local)
    g_assert (gettimeofday(&mirror->last_update, NULL) == 0);
}


/* Remove a mirror from an MTP Troop. */
void
mtp_troop_remove (MtpTroop* troop, gchar* hostname, gint port)
{
  guint i;

  g_return_if_fail (troop);
  g_return_if_fail (hostname);

  /* Find the mirror */
  for (i = 0; i < troop->mirrors->len; ++i)
    {
      MtpMirror* m = (MtpMirror*) g_ptr_array_index (troop->mirrors, i);

      if ((port == m->port) && !strcmp (hostname, m->hostname))
	{
	  /* Delete the mirror */
	  g_ptr_array_remove_index_fast (troop->mirrors, i);
	  mtp_mirror_delete (m);

	  /* Adjust the next mirror pointer if necessary. */
	  if (troop->next_mirror == troop->mirrors->len)
	    troop->next_mirror = 0;

	  return;
	}
    }
}


/* Delete a MTP Troop.  We delete the MTP Mirrors within the troop. */
void
mtp_troop_delete (MtpTroop* troop)
{
  if (troop)
    {
      guint i;

      g_free (troop->name);
      
      for (i = 0; i < troop->mirrors->len; ++i)
	mtp_mirror_delete ((MtpMirror*) g_ptr_array_index (troop->mirrors, i));
      g_ptr_array_free (troop->mirrors, TRUE);

      memset (troop, 0, sizeof(*troop));
      g_free (troop);
    }
}


void
mtp_troop_delete_hfunc (gpointer key, gpointer value, gpointer user_data)
{
  mtp_troop_delete ((MtpTroop*) value);
}


void
mtp_troop_print (MtpTroop* troop, FILE* file)
{
  guint i;

  g_return_if_fail (troop);
  g_return_if_fail (file);

  fprintf (file, "%s (%d GET_MIRRORS):\n", 
	   troop->name, troop->num_get_mirrors);
      
  for (i = 0; i < troop->mirrors->len; ++i)
    mtp_mirror_print ((MtpMirror*) g_ptr_array_index(troop->mirrors, i), file);
}


void 
mtp_troop_print_hfunc (gpointer key, gpointer value, gpointer user_data)
{
  mtp_troop_print ((MtpTroop*) value, (FILE*) user_data);
}



/* ******************** */

void
mtp_mirror_delete (MtpMirror* mirror)
{
  if (mirror)
    {
      g_free (mirror->hostname);
      memset (mirror, 0, sizeof(*mirror));
      g_free (mirror);
    }
}


void
mtp_mirror_print (MtpMirror* mirror, FILE* file)
{
  fprintf (file, "\t%s:%d, %d redirects\n",
	   mirror->hostname, mirror->port, 
	   mirror->num_redirects);
}


/* ******************** */

/* Delete a MtpLocalMirror.  Remove the mirror from any rendezvous. */
void
mtp_local_mirror_delete (MtpLocalMirror* lm)
{
  if (lm)
    {
      GList* i;

      g_free (lm->name);

      for (i = lm->rendezvous; i != NULL; i = i->next)
	{
	  MtpRendezvous* r;

	  r = (MtpRendezvous*) i->data;
	  r->local_mirrors = g_list_remove (r->local_mirrors, lm);
	}
      g_list_free (lm->rendezvous);
      
      memset (lm, 0, sizeof(*lm));
      g_free (lm);
    }
}


void
mtp_local_mirror_delete_hfunc (gpointer key, gpointer value, gpointer user_data)
{
  mtp_local_mirror_delete ((MtpLocalMirror*) value);
}


void
mtp_local_mirror_print (MtpLocalMirror* lm, FILE* file)
{
  GList* i;

  g_return_if_fail (lm);
  g_return_if_fail (lm);

  fprintf (file, "%s (%d downloads)\n",
	   lm->name, lm->num_downloads);

  for (i = lm->rendezvous; i != NULL; i = i->next)
    {
      MtpRendezvous* r = (MtpRendezvous*) i->data;
      g_return_if_fail (r && r->conn);

      fprintf (file, "\trendezvous: %s, %d\n", 
	       r->conn->hostname, r->conn->port);
    }
}


void
mtp_local_mirror_print_hfunc (gpointer key, gpointer value, gpointer user_data)
{
  mtp_local_mirror_print ((MtpLocalMirror*) value, (FILE*) user_data);
}


/* ******************** */

/* Delete a MtpRendezvous.  Remove the rendezvous from any mirror. */
void
mtp_rendezvous_delete (MtpRendezvous* r)
{
  if (r)
    {
      GList* i;

      g_free (r->rendezvous_name);

      for (i = r->local_mirrors; i != NULL; i = i->next)
	{
	  MtpLocalMirror* lm;

	  lm = (MtpLocalMirror*) i->data;
	  lm->rendezvous = g_list_remove (lm->rendezvous, r);
	}
      g_list_free (r->local_mirrors);

      jm_conn_delete (r->conn, TRUE);

      if (r->update_timer)
	g_source_remove (r->update_timer);

      memset (r, 0, sizeof(*r));
      g_free (r);
    }
}


void
mtp_rendezvous_delete_hfunc (gpointer key, gpointer value, gpointer user_data)
{
  mtp_rendezvous_delete ((MtpRendezvous*) value);
}



void
mtp_rendezvous_print (MtpRendezvous* r, FILE* file)
{
  GList* i;

  g_return_if_fail (r);
  g_return_if_fail (r->conn);
  g_return_if_fail (file);

  fprintf (file, "Rendezvous %s:%d\n", r->conn->hostname, r->conn->port);
      
  for (i = r->local_mirrors; i != NULL; i = i->next)
    {
      MtpLocalMirror* lm = (MtpLocalMirror*) i->data;

      fprintf (file, "\t%s\n", lm->name);
    }
}


void 
mtp_rendezvous_hfunc (gpointer key, gpointer value, gpointer user_data)
{
  mtp_rendezvous_print ((MtpRendezvous*) value, (FILE*) user_data);
}


