/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef _MTP_TYPES_H
#define _MTP_TYPES_H


/* Default MTP server port */
#define MTP_SERVER_PORT	    5775	

/* Period mirrors should send update to rendezvous */
#define MTP_UPDATE_TIME     (5 * 60000)	   

/* Time mirrors expire if no update received by rendezvous */
#define MTP_EXPIRE_TIME	    (MTP_UPDATE_TIME * 3)

/* Max time allowed for read or write to finish */
#define MTP_SERVER_TIMEOUT  30000

/* Number of mirrors to return in GET_MIRRORS query */
#define MTP_NUM_MIRRORS	    4


#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

#include <glib.h>
#include <gnet/gnet.h>
#include "jmgnet.h"
#include "sha.h"



/* MTP Server Type.  A server can be a rendezvous, a mirror, or
   both. */
typedef enum
{
  MTP_SERVER_TYPE_RENDEZVOUS	 	= 1,
  MTP_SERVER_TYPE_MIRROR	 	= 2,
  MTP_SERVER_TYPE_BOTH			= 3

} MtpServerType;


/* MTP Server. */
typedef struct _MtpServer
{
  gchar*	hostname;		/* Hostname of server 		*/
  JMGServer*	server;			/* Low-level listening socket	*/
  MtpServerType type;			/* Server type			*/
  gboolean	is_global; 		/* Does the rendezvous allow new
					   files to be created (ie, used
					   globally?			*/
  gboolean	verbose;		/* Print lots of messages	*/

  GList*	mtp_conns;		/* List of current connections	*/

  /* Rendezvous server members */
  GHashTable*	name_to_troop;		/* Map resource name to MtpTroop */
  guint		gc_timer;		/* Garbage collection timer for
					   removing old MtpMirrors	*/

  /* Mirror server members */
  GHashTable*	name_to_local_mirror;	/* Map resource name to MtpLocalMirror */
  GHashTable*	name_to_rendezvous;	/* Map rendezvous name to MtpRendezvous */

} MtpServer;



/* MTP Troop.  A troop if a set of mirrors for a resource.  A MtpTroop
   contains MtpMirror's. */
typedef struct _MtpTroop
{
  gchar* 	name;			/* Name of resource		*/
  GPtrArray*	mirrors;		/* Array of MtpMirrors		*/
  guint		next_mirror;		/* Next mirror to send after
					   GET_MIRRORS request		*/

  guint		num_get_mirrors;	/* Number of GET_MIRRORS received
					   for this mirror.		*/

} MtpTroop;



/* MTP Mirror.  A mirror is a host that provides some resource.  Each
   mirror is part of a troop.  A mirror can be local or remote.  A
   MtpMirror is contained by a MtpTroop. */
typedef struct _MtpMirror { 

  gchar* 	hostname; 	/* Hostname of mirror 			*/
  gint 		port;		/* Port of mirror			*/

  gboolean	is_local;	/* Is the mirror local?  If so, it
				   cannot time out.  			*/

  /* gint 	bps; */		/* BPS (not used)			*/

  gint 		num_redirects;	/* Number of times mirror has appeared
				   in MIRRORS response. 		*/
  struct timeval last_update;	/* Last time an update received from
                                   mirror by rendezvous.  		*/

} MtpMirror;


struct _MtpLocalMirror;

/* Upcall for reading data from a file for downloads of file. */
typedef gint (*MtpLocalMirrorReadFunc)(struct _MtpLocalMirror* mirror, 
				       gint offset, gchar** bufferp, gint length,
				       gpointer user_data);

/* MTP Local Mirror.  A mirror of a file on a the local host.  */
typedef struct _MtpLocalMirror
{
  gchar* 		name;	       /* Name of the file		 */
  MtpServer*		server;	       /* Server associated with mirror	 */

  GList*		rendezvous;    /* Rendezvouses this file is on
					  (should only be one...)	 */

  gint			length;	       /* Length of the file		 */
  MtpLocalMirrorReadFunc read_func;    /* Upcall for reading file	 */
  gpointer		user_data;     /* Passed to read_func		 */

  guint			num_downloads; /* Number of downloads		 */

} MtpLocalMirror;



/* MTP Rendezvous.  A remote rendezvous which a local mirror, or group
   of local mirrors, is associated with.  This structure allows us to
   aggregate rendezvous updates.  */
typedef struct _MtpRendezvous
{
  gchar* 	rendezvous_name;   /* Name of the rendezvous (hostname:port) */
  MtpServer*	server;		   /* Server associated with rendezvous [1]  */

  GList*	local_mirrors;	   /* Local mirrors using rendezous	*/
  gboolean	is_local;	   /* Is the rendezvous on this host	*/

  JMGConn*	conn;		   /* Connection to rendezvous		*/
  guint		update_timer;	   /* Timer for updating rendezvous	*/
  gint		failures;	   /* Number of failures connecting	  */
  gint		requests_pending;  /* Number of requests (writes) pending */


} MtpRendezvous;




MtpTroop* mtp_troop_new (gchar* name);
void mtp_troop_update (MtpTroop* troop, gchar* hostname, gint port, gboolean is_local);
void mtp_troop_remove (MtpTroop* troop, gchar* hostname, gint port);
void mtp_troop_delete (MtpTroop* troop);
void mtp_troop_delete_hfunc (gpointer key, gpointer value, gpointer user_data);
void mtp_troop_print (MtpTroop* troop, FILE* file);
void mtp_troop_print_hfunc (gpointer key, gpointer value, gpointer user_data);

void mtp_mirror_delete (MtpMirror* mirror);
void mtp_mirror_print (MtpMirror* mirror, FILE* file);

void mtp_local_mirror_delete (MtpLocalMirror* lm);
void mtp_local_mirror_delete_hfunc (gpointer key, gpointer value, gpointer user_data);
void mtp_local_mirror_print (MtpLocalMirror* lm, FILE* file);
void mtp_local_mirror_print_hfunc (gpointer key, gpointer value, gpointer user_data);

void mtp_rendezvous_delete (MtpRendezvous* r);
void mtp_rendezvous_delete_hfunc (gpointer key, gpointer value, gpointer user_data);


#endif /* _MTP_TYPES_H */
