/* Jungle Monkey Search Server
 * Copyright (C) 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>

#ifdef JM_ENABLE_GNOME
#include <popt-gnome.h>
#else
#include <popt.h>
#endif

#include "jmintl.h"
#include "util.h"
#include "globals.h"
#include "jmp.h"
#include "jmgnet.h"

#include "mtp_client.h"



static void mirrors_func (MtpClient* client, MtpClientStatus status, 
			  GList* conns, gpointer user_data);
static void mirror_func (MtpClient* client, MtpClientStatus status, 
			 JMGConn* conn, gpointer user_data);
static void info_func (MtpClient* client, MtpClientStatus status, 
		       gint length, gpointer user_data);
static void file_func (MtpClient* client, MtpClientStatus status, 
		       guint8* buffer, guint length, 
		       gpointer user_data);

/* ******************** */
/* Globals		*/

int   			debug_level = 0;
static int   		verbose = 0;



int main (int argc, char* argv[])
{
  gchar* url = NULL;
  JMGURL* gurl;
  GMainLoop* main_loop;

  gboolean get_mirrors = FALSE;
  gboolean get_mirror = FALSE;
  gboolean get_info = FALSE;
  gboolean get_file_from_mirror = FALSE;
  gboolean get_file_from_rendezvous = FALSE;
  MtpClient* client;


  struct poptOption jm_options[] = 
  {
    {"help", 'h', POPT_ARG_NONE, NULL, 1, 
     "Show this help message", NULL},
    {"usage", '\0', POPT_ARG_NONE, NULL, 2, 
     "Display brief usage message", NULL},
    {"version", 0, POPT_ARG_NONE, NULL, 3,
     "Output verson information and exit", NULL},
    {"verbose", 'v', POPT_ARG_NONE, &verbose, 0,
     "Print lots of stuff", NULL},

    {"debug-level", 'd', POPT_ARG_INT, &debug_level, 0, 
     "Set the debug level", "LEVEL"},

    {"get-mirrors", '\0', POPT_ARG_NONE, &get_mirrors, 0, 
     "Get a list of mirrors from rendezvous", NULL},
    {"get-mirror", '\0', POPT_ARG_NONE, &get_mirror, 0, 
     "Get the best mirror", NULL},
    {"get-info", '\0', POPT_ARG_NONE, &get_info, 0, 
     "Get info from mirror about file", NULL},
    {"get-file-from-mirror", '\0', POPT_ARG_NONE, &get_file_from_mirror, 0, 
     "Request file from mirror", NULL},
    {"get-file", '\0', POPT_ARG_NONE, &get_file_from_rendezvous, 0, 
     "Request file via rendezvous", NULL},

    {NULL, '\0', 0, NULL, 0, NULL, NULL}
  };

  poptContext ctx;
  int rc;
  int exit_status = EXIT_SUCCESS;


  /* ******************** */
  /* Initialize NLS 	  */
#ifdef ENABLE_NLS
  {
/*      gtk_set_locale(); */
    bindtextdomain (PACKAGE, LOCALEDIR);
    textdomain (PACKAGE);
  }
#endif

  /* ******************** */
  /* Parse command line	  */

  ctx = poptGetContext (NULL, argc, argv, jm_options, 0);
  poptSetOtherOptionHelp (ctx, _("[OPTION]... [URL]"));

  while ((rc = poptGetNextOpt(ctx)) > 0)
    {
      switch (rc)
	{
	case 1:	  goto help;	  break;
	case 2:	  goto usage;	  break;
	case 3:   goto version;	  break;
	}
    }

  url = poptGetArg (ctx);

  poptFreeContext (ctx);

  if (!url)
    my_error (_("You must specify a URL.\n"));

  gurl = jm_url_new (url);
  if (!gurl)
    my_error (_("Malformed URL: %s\n"), url);


  /* ******************** */
  /* Do whatever	  */

  if (get_mirrors)
    {
      client = mtp_client_get_mirrors (gurl, mirrors_func, NULL);
      if (!client)
	g_warning (_("Unable to get mirrors\n"));
    }

  if (get_mirror)
    {
      client = mtp_client_get_mirror (gurl, mirror_func, NULL);
      if (!client)
	g_warning (_("Unable to get best mirror\n"));
    }

  if (get_info)
    {
      client = mtp_client_get_info (gurl, info_func, NULL);
      if (!client)
	g_warning (_("Unable to request info\n"));
    }

  if (get_file_from_mirror)
    {
      client = mtp_client_get_file_mirror (gurl, file_func, NULL);
      if (!client)
	g_warning (_("Unable to get file from mirror\n"));
    }

  if (get_file_from_rendezvous)
    {
      client = mtp_client_get_file_rendezvous (gurl, file_func, NULL);
      if (!client)
	g_warning (_("Unable to get file from rendzvous\n"));
    }


  /* ******************** */
  /* Start the main loop  */

  main_loop = g_main_new (FALSE);
  g_main_run (main_loop);

  exit (EXIT_SUCCESS);

 usage:
  poptPrintUsage (ctx, stdout, 0);
  exit (exit_status);

 help:
  g_print (_("`mtpclient' - MTP Client\n"));
  poptPrintHelp (ctx, stdout, 0);
  g_print (_("\n"
	     "Report bugs to <jm-dev@umich.edu>\n"));
  exit (exit_status);

 version:
  g_print (_("mtpclient %s\n"
	     "Written by David A. Helder\n"
	     "\n"
	     "Copyright (C) 2000 The Regents of the University of Michigan\n"
	     "This is free software; see the source for copying conditions.  There is NO\n"
	     "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n"),
	     VERSION);
  exit (exit_status);
}


static void
mirrors_func (MtpClient* client, MtpClientStatus status, 
	      GList* conns, gpointer user_data)
{
  if (status == MTP_CLIENT_STATUS_OK)
    {
      GList* i;

      g_print (_("Found %d mirrors\n"), g_list_length (conns));
      
      for (i = conns; i != NULL; i = i->next)
	{
	  JMGConn* conn;

	  conn = (JMGConn*) i->data;

	  g_print (_("\t%s:%d\n"), conn->hostname, conn->port);
	  jm_conn_unref (conn, TRUE);
	}
    }
  else
    {
      g_warning (_("Failed to get mirrors.\n"));
    }

  mtp_client_delete (client);
}


static void
mirror_func (MtpClient* client, MtpClientStatus status, 
	     JMGConn* conn, gpointer user_data)
{
  if (status == MTP_CLIENT_STATUS_OK)
    {
      g_print (_("Best mirror is %s:%d\n"), conn->hostname, conn->port);
      jm_conn_unref (conn, TRUE);
    }
  else
    g_warning (_("Failed to get mirror.\n"));

  mtp_client_delete (client);
}


static void
info_func (MtpClient* client, MtpClientStatus status, gint length, gpointer user_data)
{
  if (status == MTP_CLIENT_STATUS_OK)
    g_print (_("Got info: length = %d\n"), length);
  else
    g_warning (_("Failed to get info.\n"));

  mtp_client_delete (client);
}


static void
file_func (MtpClient* client, MtpClientStatus status, guint8* buffer, guint length, 
	   gpointer user_data)
{
  if (status == MTP_CLIENT_STATUS_OK)
    {
      if (!buffer)
	g_print (_("Transfering file %s (length %d) from %s:%d\n"), 
		 client->url->resource, client->length, 
		 client->conn->hostname, client->conn->port);
      else
	fwrite (buffer, length, 1, stdout);

      if (client->length == (client->offset + length))
	mtp_client_delete (client);
    }
  else
    {
      g_warning (_("Failed to get file.\n"));
    }
}


