/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>

/* These are here so MIN/MAX don't get redefined */
#include "util.h"
#include <glib.h>


static gboolean is_blankpathname(gchar* path, gboolean (*func)(gchar*));


/* **************************************** */


time_t 
get_time_gmt(void)
{
  time_t current_time;

  time(&current_time);
  g_assert(gmtime(&current_time) != NULL);

  return current_time;
}


int 
ssleep(int sec)
{
  struct timeval tv;
  
  tv.tv_sec = sec;
  tv.tv_usec = 0;

  return select(0, NULL, NULL, NULL, &tv);
}


int 
mssleep(int ms)
{
  struct timeval tv;
  
  tv.tv_sec = 0;
  tv.tv_usec = ms * 1000;

  return select(0, NULL, NULL, NULL, &tv);
}


int 
ussleep(int us)
{
  struct timeval tv;
  
  tv.tv_sec = 0;
  tv.tv_usec = us;

  return select(0, NULL, NULL, NULL, &tv);
}


/* **************************************** */

gchar* 
strcasestr(const gchar* haystack, const gchar* needle)
{
  guint i, j;

  g_return_val_if_fail (haystack != NULL, NULL);
  g_return_val_if_fail (needle   != NULL, NULL);

  if (needle[0] == '\0')
    return (gchar*) haystack;

  for (i = 0; haystack[i] != '\0'; ++i)
    {
      for (j = 0; needle[j] != '\0'; ++j)
	{
	  if (tolower(haystack[i + j]) != tolower(needle[j]))
	    goto next;
	}
      return (gchar*) &haystack[i];

    next:
      ;
    }

  return NULL;

}




gchar*
strtokq(gchar* s, gchar* delim, gchar** next)
{
  gchar* start = NULL;
  gchar last_char = 0;
  gboolean is_quote = 0;
  gint i;

  /* Set s if necessary */
  if (s == NULL)
    s = *next;

  /* Find the beginning of the next token */
  for (;;)
    {
      if (*s == '\0')
	break;

      /* Try to find a skipable character */
      for (i = 0; delim[i] != *s && delim[i] != '\0'; ++i);

      /* We found non-skippable character - break */
      if (delim[i] == '\0')
	break;

      /* Check if it's a quote */
      if (delim[i] == '\"')
	{
	  /* If it's not part of a \", break */
	  if (last_char != '\\')
	    {
	      is_quote = TRUE;
	      ++s;
	      break;
	    }
	}

      /* Skip it */
      last_char = *s++;
    }


  /* Return NULL if we're done */
  if (*s == '\0')
    {
      *next = s;
      return NULL;
    }

  /* Read the token */
  start = s;
  last_char = '\0';

  for (;;)
    {
      if (*s == '\0')
	break;

      /* Try to find a skipable character */
      for (i = 0; delim[i] != *s && delim[i] != '\0'; ++i);

      /* Check if we found a quote (endquote or otherwise, it doesn't matter) */
      if (delim[i] == '\"' && last_char != '\\')
	break;

      /* Check if we found a skippable character */
      else if (is_quote == FALSE && delim[i] != '\0')
	break;

      /* This character is part of the token, continue */
      last_char = *s++;
    }

  /* If we didn't hit the end, write '\0' and update next.
     Otherwise, set next to the end so that the next call returns NULL */
  if (*s != '\0')
    {
      *s++ = '\0';
      *next = s;
    }
  else
    *next = s;		/* This causes next call to return NULL */

  return start;
}


gchar*
strescape(gchar* s)
{
  gchar escextra;
  gchar* r;
  gchar* start;

  int size;

  size = 1 + strlen(s) + strchrcount(s, "\a\b\f\n\r\t\v\\\'\"");
  size += strchrcount(s, "&") * 3;

  start = r = g_new(gchar, size);
  
  for (;;)
    {
      switch (*s)
	{
	case '\a':      	{ escextra = 'a';  goto esc_extra; }
	case '\b':		{ escextra = 'b';  goto esc_extra; }
	case '\f':		{ escextra = 'f';  goto esc_extra; }
	case '\n':	      	{ escextra = 'n';  goto esc_extra; }
	case '\r':      	{ escextra = 'r';  goto esc_extra; }
	case '\t':	      	{ escextra = 't';  goto esc_extra; }
	case '\v':	      	{ escextra = 'v';  goto esc_extra; }
	case '\\' :       	{ escextra = '\\'; goto esc_extra; }
	case '\'' :       	{ escextra = '\''; goto esc_extra; }
	case '\"' :       	{ escextra = '\"'; goto esc_extra; }
	  {
	  esc_extra:
	    *r++ = '\\';
	    *r++ = escextra;
	    ++s;
	    break;
	  }
	case '\0':	  	{ *r = *s;	    return start; }
	case '&': 		{ *r++ = *s++; 
	                          *r++ = 'a';
				  *r++ = 'm';
				  *r++ = 'p';			  }
	default:                { *r++ = *s++;			  }
	}
    }

  g_assert_not_reached();
  return NULL;
}


gchar*
strunescape(gchar* s)
{
  gchar* r;
  gchar* start;
  int size;

  size = 1 + strlen(s);		/* This is an overestimate */
  start = r = g_new(gchar, size);
  
  for (;;)
    {
      switch (*s)
	{
	case '\\':
	  {
	    switch (*++s)
	      {
	      case 'a':      	{ *r++ = '\a';	break;	}
	      case 'b':		{ *r++ = '\b';  break; 	}
	      case 'f':		{ *r++ = '\f';  break; 	}
	      case 'n':	      	{ *r++ = '\n';  break; 	}
	      case 'r':      	{ *r++ = '\r';  break; 	}
	      case 't':	      	{ *r++ = '\t';  break; 	}
	      case 'v':	      	{ *r++ = '\v';  break; 	}
	      case '\\' :       { *r++ = '\\'; 	break;	}
	      case '\'' :       { *r++ = '\''; 	break; 	}
	      case '\"' :       { *r++ = '\"'; 	break; 	}
	      case '\0':	{ *r   = '\0';	return start; }
	      default:         	{ *r++ = '\\';  *r++ = *s;    }
	      }
	    ++s;
	    break;
	  }
	case '&':
	  {
	    *r++ = *s++;

	    if (strncmp("amp", s, 3) == 0)
	      s += 3;
	  }
	case '\0': 		{ *r   = '\0';	return start; 	}
	default: 		{ *r++ = *s++; 		}
	}
    }

  g_assert_not_reached();
  return NULL;
}


gint
strchrcount(char* s, char* chars)
{
  int i;
  int count = 0;


  for (; *s != '\0'; ++s)
    {
      for (i = 0; chars[i] != '\0'; ++i)
	{
	  if (chars[i] == *s)
	    {
	      ++count;
	      goto next;
	    }
	}
    next:;
    }

  return count;
}


gint
strsuffixcmp(gchar* s1, gchar* s2)
{
  int l_s1 = strlen(s1);
  int l_s2 = strlen(s2);

  if (l_s1 > l_s2)
    return strcmp(&s1[l_s1 - l_s2], s2);
  else
    return strcmp(s1, &s2[l_s2 - l_s1]);

  g_assert_not_reached();
}



/* **************************************** */

int 
file_size(FILE* f)
{
  int rv;
  struct stat filestat;

  rv = fstat(fileno(f), &filestat);

  if (rv == 0)
    return filestat.st_size;

  return rv;
}


int
filename_size (gchar* filename)
{
  FILE* file;
  int rv;

  if ((file = fopen (filename, "r")) == NULL)
    return -1;

  rv = file_size (file);

  fclose (file);

  return rv;
}


gchar* 
filesize_to_string(gint size)
{
  gchar* str;

  g_return_val_if_fail (size >= 0, NULL);

  if (size < 1024)
    str = g_strdup_printf("%d bytes", size);
  else if (size < (1024 * 1024))
    str = g_strdup_printf("%.1fk", (double) size / 1024.0);
  else
    str = g_strdup_printf("%.1fM", (double) size / 1048576.0);

  return str;
}


gchar* 
filesize_fraction_to_string(gint offset, gint size)
{
  gchar* str;

  g_return_val_if_fail (offset >= 0, NULL);
  g_return_val_if_fail (size >= 0, NULL);

  if (size < 1024)
    str = g_strdup_printf("%d/%d bytes", offset, size);
  else if (size < (1024 * 1024))
    str = g_strdup_printf("%.1f/%.1fk", 
			  (double) offset / 1024.0,
			  (double) size / 1024.0);
  else
    str = g_strdup_printf("%.1f/%.1fM", 
			  (double) offset / 1048576.0,
			  (double) size / 1048576.0);

  return str;
}


gboolean
file_exists(gchar* path)
{
  FILE* f;

  g_return_val_if_fail(path != NULL, FALSE);

  f = fopen(path, "r");
  if (f == NULL)
    return FALSE;

  fclose(f);

  return TRUE;
}


gboolean
is_dir(gchar* path)
{
  struct stat filestat;

  if (stat(path, &filestat) == 0)
    return (S_ISDIR(filestat.st_mode) != 0);

  return FALSE;
}


gboolean
mkdirr(gchar* path, gint mode)
{
  gchar* parent;
  gboolean rv;

  /* Try to make the directory as given */
  if (mkdir(path, mode) == 0 || errno == EEXIST)
    return TRUE;

  /* Failed - try to make the parent directory */
  parent = g_dirname(path);
  rv = mkdirr(parent, mode);
  g_free(parent);

  /* Try to make the directory again if we created the parent.  We
     check EEXIST because the original path may have ended with a
     directory separator. */
  if (rv)
    return (mkdir(path, mode) == 0 || errno == EEXIST);

  return FALSE;
}


gboolean
is_okfilename(gchar* name)
{
  g_return_val_if_fail (name != NULL, FALSE);

  /* Skip any whitespace */
  while (isspace(*name))
    ++name;

  /* Can't be blank */
  if (name[0] == '\0')
    return FALSE;

  /* The name can't be . or .. */
  if (name[0] == '.')
    {
      if (name[1] == '\0')
	return FALSE;
      else if (name[1] == '.' && name[2] == '\0')
	return FALSE;
    }

  /* The name can't have / in it */
  while (*name)
    {
      if (*name == '/')
	return FALSE;

      ++name;
    }

  return TRUE;
}


gboolean
is_okpathname(gchar* path)
{
  return is_blankpathname(path, is_okfilename);
}



gboolean
is_goodfilename(gchar* name)
{
  int i = 0;

  g_return_val_if_fail (name != NULL, FALSE);

  /* Skip any whitespace */
  while (isspace(*name))
    ++name;

  /* Good file names begin with an alphanumerical character */
  if (name[0] == '\0' || !isalnum(name[i]))
    return FALSE;
      

  /* Good names have alpha-num characters and certain punctuation only */
  /* We start at 0 to weed out names that begin with weird punctuation (like /) */
  for (i = 0; name[i] != '\0'; ++i)
    {
      /* Don't allow /,\, or : because the are directory seperators on
         some systems. */

      if (isalnum(name[i]) || 
	  name[i] == '.' || name[i] == '-' || 
	  name[i] == '_' || name[i] == ' ' ||
	  name[i] == '!' || name[i] == '\'' ||
	  name[i] == '(' || name[i] == ')' ||
	  name[i] == '&' || name[i] == ',' ||
	  name[i] == '\"' || name[i] == '+')
	continue;

      return FALSE;
    }

  return TRUE;
}


gboolean
is_goodpathname(gchar* path)
{
  return is_blankpathname(path, is_goodfilename);
}


gboolean
is_blankpathname(gchar* path, gboolean (*func)(gchar*))
{
  gchar** tokens;
  guint i;
  gboolean rv = FALSE;

  g_return_val_if_fail(path != NULL, FALSE);
  g_return_val_if_fail(func != NULL, FALSE);

  /* Skip leading '/' */
  if (path[0] == G_DIR_SEPARATOR)
    path = &path[1];

  tokens = g_strsplit(path, G_DIR_SEPARATOR_S, 0);

  if (tokens == NULL || tokens[0] == NULL)
    goto done;

  for (i = 0; tokens[i] != NULL; ++i)
    if (!func(tokens[i]))
      goto done;

  rv = TRUE;

 done:
  g_free(tokens);
  return rv;
  

}



gint
launch(gchar* program_name, gchar* argv[], gchar* envp[])
{
  pid_t pid = -1;
  gchar* path;

  path = search_path (program_name);

  if (path != NULL)
    {
     if ((pid = fork()) == 0)
       {
	 extern char** environ;

	 /* Execute the program */
	 execve (path, argv, (envp != NULL)?(envp):(environ));

	 fprintf (stderr, "*** Failed to launch %s", path);
	 perror(" ");

	 /* This point should not be reached. */
	 exit (EXIT_FAILURE);
       }
     else if (pid > 0)
       {
	 g_free (path);
       }
     else
       {
	 fprintf (stderr, "*** Error execve %s\n", path);
	 perror(" ");
       }
    }
  else 
    fprintf (stderr, "*** Can't find %s\n", program_name);

  return pid;
}



gchar*
search_path(gchar* filename)
{
  /* Find the full path */
  if (strchr(filename, '/'))		/* Check if absolute or relative */
    return g_strdup(filename);

  else					/* Look in path */
    {
      char* pathenv;
      char* next;
      char* save;

      /* Get the PATH environment variable */
      pathenv = g_strdup(getenv("PATH"));

      next = strtok_r (pathenv, ":", &save);
      while (next != NULL)
	{
	  struct stat statbuf;
	  gchar path[MAXPATHLEN];
	  int length = strlen(next);

	  /* Copy the path into pathname */
	  strncpy (path, next, length);

	  /* Add a / if necessary */
	  if (path[length - 1] != '/')
	    path[length++] = '/';

	  /* Copy the filename into pathname */
	  strcpy (path + length, filename);

	  fprintf (stderr, "search_path: Looking for %s\n", path);

	  /* Stop if the file exists */
	  if (stat(path, &statbuf) == 0)
	    {
	      g_free (path);
	      return g_strdup (path);
	    }

	  next = strtok_r(NULL, ":", &save);
	}

      /* File doesn't exist */
      g_free (pathenv);
      return NULL;
    }

  return NULL;	/* not reached */
}



void
parse_url(const gchar* url, gchar** proto, gchar** server, gint* port, gchar** filename)
{
  const gchar* p;
  const gchar* temp;

  g_return_if_fail(url != NULL);

  if (proto != NULL)	*proto = NULL;
  if (server != NULL)  	*server = NULL;
  if (port != NULL)	*port = 0;
  if (filename != NULL)	*filename = NULL;

  /* Skip initial whitespace */
  while (*url != 0 && isspace(*url))
    ++url;
  p = url;

  /* Read in protocol (protocol is optional) */
  temp = p;
  while (*p != 0 && *p != ':' && *p != '/')
    ++p;
  if (*p == ':')
    {
      if (proto != NULL)
	*proto = g_strndup(url, p - url);
      while (*p != 0 && (*p == ':' || *p == '/')) ++p;
    }
  else
    p = temp;

  /* Read in the server */
  temp = p;
  while (*p != 0 && *p != '/' && *p != ':') ++p;
  if ((p - temp) == 0) return;
  if (server != NULL)
    *server = g_strndup(temp, p - temp);

  /* Read in the port */
  if (*p == ':')
    {
      temp = ++p;
      while (*p != 0 && *p != '/') ++p;
      if ((p - temp) != 0 && port != NULL) 
	{
	  gchar* port_str = g_strndup(temp, p - temp);
	  *port = atoi(port_str);
	  g_free(port_str);
	}
    }

  /* Read in the file */
  if (filename != NULL)
    {
      if (*p == 0)
	*filename = g_strdup("/");
      else
	*filename = g_strdup(p);
    }
}



gchar* 		
make_url(const gchar* url, gchar* should_be, gchar* can_be, gint default_port)
{
  gchar* proto;
  gchar* server;
  gint port;
  gchar* file;
  gchar num[16];
  gchar* rv = NULL;

  /* Parse the URL */
  parse_url(url, &proto, &server, &port, &file);

  if (server == NULL)
    goto done;

  if (proto != NULL)
    {
      if (strcmp(proto, should_be) != 0)
	{
	  if (!can_be || strcmp(proto, can_be) != 0)
	    goto done;
	}
    }

  if (port == 0)
    port = default_port;

  if (port)
    {
      g_snprintf(num, sizeof(num), "%d", port);

      if (file)
	rv = g_strconcat(should_be, "://", server, ":", num, /*"/",*/ file, NULL);
      else
	rv = g_strconcat(should_be, "://", server, ":", num, NULL);
    }
  else
    {
      if (file)
	rv = g_strconcat(should_be, "://", server, /*"/",*/ file, NULL);
      else
	rv = g_strconcat(should_be, "://", server, NULL);
    }


 done:
  g_free(proto);
  g_free(server);
  g_free(file);

  return rv;
}

void
my_error (const gchar* format, ...)
{
  va_list args;
  
  va_start (args, format);
  g_logv (G_LOG_DOMAIN, G_LOG_LEVEL_CRITICAL, format, args);
  va_end (args);

  exit (EXIT_FAILURE);
}

