/* Jungle Monkey
 * Copyright (C) 1999, 2000  The Regents of the University of Michigan
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <glib.h>
#include <stdio.h>
#include <time.h>

#ifndef _UTIL_H
#define _UTIL_H



/* ******************************************** */
/* General functions/macros			*/

/**

   Round n up to the nearest base.

*/
#define ROUNDUP(n, base)   ( (((n) % (base)) == 0) ? \
			     (n) :\
			     ((n) + ((base) - ((n) % (base)))) )




/* ********************************************	*/
/* Time related macros and functions.		*/

#define timernormalize(tvp)						\
       do {								\
           (tvp)->tv_sec  += (tvp)->tv_usec / 1000000; 			\
	   (tvp)->tv_usec %= 1000000;					\
       } while (0)

#define timeraddusec(tvp, usec)					\
       do {								\
           (tvp)->tv_usec += usec;					\
           if ((tvp)->tv_usec < 0)					\
           {								\
             (tvp)->tv_sec -= (-((tvp)->tv_usec / 1000000) + 1);	\
             (tvp)->tv_usec = (1000000 + ((tvp)->tv_usec % 1000000));	\
           }								\
           timernormalize(tvp);					\
       } while (0)

#define timerusec(tvp) (((tvp)->tv_sec * 1000000) + (tvp)->tv_usec)

/* if negative, round up to zero */
#define timermakepos(tvp) (((tvp)->tv_sec < 0))				\
                              ?((tvp)->tv_sec = 0, (tvp)->tv_usec = 0)	\
			      :(0)


/**

   Get the time in GMT.

*/
time_t get_time_gmt(void);


/**

   Sleep a given number of seconds.  

   Note that the accuracy of this is limited by the OS.  This function
   uses select, so it should be thread safe and portable.  Returns the
   return value of select, which is 0 on success.

*/
int ssleep(int ms);


/**

   Sleep a given number of milliseconds.  

   Note that the accuracy of this is limited by the OS.  This function
   uses select, so it should be thread safe and portable.  Returns the
   return value of select, which is 0 on success.

*/
int mssleep(int ms);


/**

   Sleep a given number of microseconds.  

   Note that the accuracy of this is limited by the OS.  This function
   uses select, so it should be thread safe and portable.  Returns the
   return value of select, which is 0 on success.

*/
int ussleep(int us);



/* ******************************************** */
/* String functions				*/

gchar* strcasestr(const gchar* haystack, const gchar* needle);



/**

   Thread safe string tokenizer with double quote's.

   It works like strtok_r.  Arguments are string to tokenize, the
   delimiters, and the next pointer.  If you want double quotes, put a
   double quote in the delim's.  It doesn't count \"'s as quotes.

   TODO: Add single quote's?

*/
gchar* strtokq(gchar* s, gchar* delim, gchar** next);


/**

   Add quotes around the string s.

*/
#define strquote(s)   (g_strconcat("\"", (s), "\"", NULL))


/**

   Convert special characters to escape sequences.  This includes
   converting &'s to &amp's.

*/
gchar* strescape(gchar* s);


/**

   Convert escape sequences to special characters.  This includes
   converting &amp's to &'s.

*/
gchar* strunescape(gchar* s);


/**

   Count the number of occurences of any of the characters in chars in
   s.

*/
gint strchrcount(gchar* s, gchar* chars);


/**

   Do a strncmp on the suffixes of the strings.

 */
gint strsuffixcmp(gchar* s1, gchar* s2);



/* ******************************************** */
/* File/program execution functions		*/


/**

   Get the size of a file.  Returns negative number if there was an
   error.

*/
int file_size (FILE* f);

int filename_size (gchar* filename);


gchar* filesize_to_string(gint size);
gchar* filesize_fraction_to_string(gint offset, gint size);


/**

   Check if the file exists and is readable.

*/
gboolean file_exists(gchar* path);


/**

   Check if the path is a directory.  Returns TRUE if the path is a
   directory, FALSE otherwise.

*/
gboolean is_dir(gchar* path);


/**

   Make the directory recursively.  Returns TRUE on success.

   This attempts to create the path given, if this cannot be done, it
   attempts to mkdir the parent directory and so on.  When this is
   done, it tries again.

 */
gboolean mkdirr(gchar* path, gint mode);





/**

   Check if the name is just an ok file name.  An ok file name is any
   name that isn't just whitespace, isn't '.' or '..', and doesn't
   have '/' in it.  Anything else is allowed.

 */
gboolean is_okfilename(gchar* name);



/**

   Check if the path is an ok path name.  A good path is made of up ok
   file names.  This just strtoks the path on '/' and checks to see if
   each token is an ok file name.  

*/
gboolean is_okpathname(gchar* path);




/**

   Check if the name a good file name.  A good file name is
   alphanumeric characters with some punctuation in the middle.  A bad
   file name has weird punctuation in it.  We don't allow files that
   begin with a '.'.

   Punctuation we allow: period, dash, underscore, space, exclamation
   point, single quote, double quote, parenthesis, comma, and
   ampersand.
   
   Note if you exec something, you may want to check for ampersand.

 */
gboolean is_goodfilename(gchar* name);



/**

   Check if the path is a good path name.  A good path is made of up
   good file names.  This just strtoks the path on '/' and checks to
   see if each token is a good file name.
*/
gboolean is_goodpathname(gchar* path);


/**
   
   Launch the program specified by program name with the given
   arguments and environment variables.  The arguments and environment
   variables should be NULL terminated.  If envp is NULL, the global
   environ is used.  By convention, the first argument should be the
   program name.

   Returns the number of the process lauched on success, -1 otherwise.

 */
gint launch(gchar* program_name, gchar* argv[], gchar* envp[]);



/**

   Search the environment variable PATH for the specified file, if it
   is not absolute or relative.  The function will not work correctly
   for malformed PATH variables - don't rely on this for security.

   Returns the full path name on success, NULL otherwise.  The path
   name can be deleted.

 */
gchar* search_path(char* filename);





/**
   
   Parse the URL and put the result in the given pointers.  Delete the
   pointers when you're done.

   Allowed formats:
     [proto[/:]*]]hostname[:port][/filename]


 */
void
parse_url(const gchar* url, gchar** proto, gchar** server, gint* port, gchar** filename);

/* ugly function...  I really need a URL class */
gchar*
make_url(const gchar* url, gchar* should_be, gchar* can_be, gint default_port);


/**

   Log error as critical and then call exit(FAIL).  Glib logs as error
   then calls abort, which dumps core.

 */
void
my_error (const gchar* format, ...);


#endif /* _UTIL_H */
